/**
 * 
 */
package cnh.crb;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;

import javax.xml.stream.XMLStreamException;

import org.apache.commons.collections15.Predicate;

import repast.simphony.PropertiesWriter;
import repast.simphony.context.Context;
import repast.simphony.context.space.graph.NetworkBuilder;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.parameter.ParameterConstants;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;
import simphony.util.messages.MessageCenter;
import cern.jet.random.Uniform;
import cern.jet.random.engine.MersenneTwister;
import cnh.crb.TimeKeeper.Day;
import cnh.crb.TimeKeeper.TimeKeeperAction;
import cnh.crb.agent.AbstractWaterManager;
import cnh.crb.agent.DataDrivenDemandWaterManager;
import cnh.crb.agent.ResidentialWaterManager;
import cnh.crb.climate.Climate;
import cnh.crb.climate.ClimateImpacts;
import cnh.crb.climate.ClimateLoader;
import cnh.crb.climate.TemperatureStation;
import cnh.crb.econ.BillAction;
import cnh.crb.household.DefaultHouseholdInitializer;
import cnh.crb.household.Household;
import cnh.crb.household.Household.HouseholdData;
import cnh.crb.household.HouseholdLoader;
import cnh.crb.household.InfluenceSetCreator;
import cnh.crb.household.KENetworkGenerator;
import cnh.crb.hydro.FlowAction;
import cnh.crb.hydro.FlowFrequency;
import cnh.crb.hydro.FlowHydrograph;
import cnh.crb.hydro.FlowManager;
import cnh.crb.hydro.FlowNode;
import cnh.crb.hydro.FlowRate;
import cnh.crb.hydro.FlowRate.FlowRateUnits;
import cnh.crb.hydro.NaturalWaterSource;
import cnh.crb.hydro.SourceFlowNode;
import cnh.crb.hydro.WaterSink;
import cnh.crb.messaging.HouseholdMessager;
import cnh.crb.parameter.DistributionParameter;
import cnh.crb.regional.CAPSubcontractor;
import cnh.crb.regional.CAWCD;
import cnh.crb.stats.HistogramDataRecorder;
import cnh.crb.wbm.WBMCell;
import cnh.crb.wbm.WBMCrop;
import cnh.crb.wbm.WBMNaturalWaterSource;
import cnh.crb.wbm.WBMProxy;
import cnh.crb.wbm.WBMWaterSourceCollection;
import cnh.crb.wbm.WBMWaterSourceManager;

/**
 * Builds the context by creating households, water managers etc.
 * 
 * @author NC
 */
public class ContextCreator implements ContextBuilder<Object> {

	static MessageCenter msg = MessageCenter.getMessageCenter(ContextCreator.class);

	// Switches
	
	// Context Builder
    boolean recordHistograms               = false;
	boolean useTucsonHouseholds            = false;
	boolean useSocialNetwork               = false  && useTucsonHouseholds;
	boolean useMessaging                   = false  && useTucsonHouseholds; // Can't use messaging without households
	boolean useCAPCanal                    = true;

	// Create CAP Canal
	// Assumes that there will be an offtake at Lake Havasu;
	// other components are optional
	boolean useCAPNetwork                  = true;
	boolean useCAPSubcontractors           = true && useCAPNetwork;
	boolean useDataDriven                  = true && useCAPNetwork; // Dependent on CAWCD
	boolean useDataDrivenAG                = true && useDataDriven;
	boolean useDataDrivenMI                = true && useDataDriven;
	boolean useDataDrivenRX                = true && useDataDriven;

	// WBM Water Sources
	boolean useCAPCustomerWaterSources     = true;
	boolean useCAPGridWaterSources         = true;
	boolean allowSourcesToManageIrrigation = true;
	boolean cropsReturnZeroInAllCases      = false;
	
	public static Map<WBMNaturalWaterSource, List<WBMNaturalWaterSource>> mapOfDuplicatesRemoved = new HashMap<WBMNaturalWaterSource, List<WBMNaturalWaterSource>>();

	
	private static final class ProxyRunner implements Runnable {

		WBMProxy proxy;

		ProxyRunner(WBMWaterSourceCollection collection) {
			proxy = new WBMProxy(collection);
		}

		public void run() {
			while (true)
				proxy.run();
		}

	}

	private static class AlwaysPassFilter implements Predicate<Household.HouseholdData> {
		@Override
		public boolean evaluate(HouseholdData arg0) {
			return true;
		}
	}

	// Use this to limit the number of households created, regardless of the
	// data
	private static class LimitFilter implements Predicate<Household.HouseholdData> {

		int count = 0;
		int limit;

		public LimitFilter(int limit) {
			this.limit = limit;
		}

		public boolean evaluate(HouseholdData arg0) {
			count++;
			return (count < limit);
		}
	}

	private static class SampleFilter implements Predicate<Household.HouseholdData> {
		double prob;
		Uniform sampleDist = RandomHelper.createUniform(0, 1);

		public SampleFilter(double probability) {
			prob = probability;
		}

		public boolean evaluate(HouseholdData arg0) {
			return sampleDist.nextDouble() < prob;
		}
	}

	private static class ProgressIndicator implements IAction {
		Calendar calendar;
		MessageCenter msg;
		
		public ProgressIndicator(Calendar cal, MessageCenter messageCenter){
			calendar = cal;
			msg = messageCenter;
		}
		
		public void execute(){
			if(calendar.get(Calendar.DAY_OF_YEAR) == 1) msg.info("Run: " + RunState.getInstance().getRunInfo().getRunNumber() + " " + calendar.getTime() + " tick: " + RunEnvironment.getInstance().getCurrentSchedule().getTickCount() + "/" + RunEnvironment.getInstance().getParameters().getDouble("stopAt"));
		}
	}
	
	private static class UpdateSeasonalData implements IAction {

		private Calendar cal;

		public UpdateSeasonalData(Calendar calendar) {
			this.cal = calendar;
		}

		public void execute() {
			SeasonalWaterUseInstance.getWaterUse().resetMonthlyScaleFactor(cal);
		}
	}
	
	private static class RecordClimateData implements IAction{
		private Context<Object> context;
		private String filename;
		private TimeKeeper tk;
		boolean firstTime = true;
		Vector<String> stations = new Vector<String>();

		public RecordClimateData(Context<Object> context, String file, TimeKeeper tk){
			this.context = context;
			this.filename = file;
			this.tk = tk;
		}
		
		public void execute(){
			System.out.println("Recording climate data...");
			String dateToWrite = tk.getCal().get(Calendar.YEAR) + "-";
			int month = tk.getCal().get(Calendar.MONTH) + 1;
			dateToWrite += (month < 10 ? "0" : "") + month + "-00"; // Monthly data ends in 00
//			int day = tk.getCal().get(Calendar.DAY_OF_MONTH);
//			dateToRead += (day < 10 ? "0" : "") + day;
			System.out.println("WRITING DAY: " + dateToWrite);
			String header = "Date";
			String dataLine  =  dateToWrite; 
			
			// Update all duplicates... This won't be done automatically because they've been removed...
			for(Map.Entry<WBMNaturalWaterSource, List<WBMNaturalWaterSource>> entry: ContextCreator.mapOfDuplicatesRemoved.entrySet()){
				double temp = ((TemperatureStation)(entry.getKey())).getTemperature();
				for(WBMNaturalWaterSource otherDup: entry.getValue()) ((TemperatureStation)otherDup).setTemperature(temp);
			}
			
			
			Map<String, Double> stationTemps = new TreeMap<String, Double>();
			for (Object obj : context.getObjects(TemperatureStation.class)) {
				TemperatureStation ts = (TemperatureStation)obj;
				String tsID = ts.getTemperatureStationId();
				stationTemps.put(tsID, ts.getTemperature());
			}
			if(firstTime){
				// Initialize the canonical order of stations
				for(Entry<String, Double> entry: stationTemps.entrySet()){
					stations.add(entry.getKey());
					header += "," + entry.getKey();
					dataLine += "," + entry.getValue();
				}
			}
			else{
				for(String stationID: stations)	dataLine += "," + stationTemps.get(stationID);
			}
			try{
				Writer writer = new FileWriter(filename, !firstTime);
				if(firstTime) writer.write(header + System.getProperty("line.separator"));
				writer.write(dataLine + System.getProperty("line.separator"));
				writer.close();
			}
			catch(Exception e){
				System.err.println("Could not write Climate Data to file");
			}
			firstTime = false;
		}
	}
	

	private static class UpdateClimateData implements IAction{
		
		private Context<Object> context;
		private String filename;
		private TimeKeeper tk;
		private String[] stations = null;
		BufferedReader bufferedReader = null;
		
		public UpdateClimateData(Context<Object> context, String file, TimeKeeper tk){
			this.context = context;
			this.filename = file;
			this.tk = tk;
		}
		public void execute(){
			// Determine the data
			String dateToRead = tk.getCal().get(Calendar.YEAR) + "-";
			int month = tk.getCal().get(Calendar.MONTH) + 1;
			dateToRead += (month < 10 ? "0" : "") + month + "-00"; // Monthly data ends in 00
//			int day = tk.getCal().get(Calendar.DAY_OF_MONTH);
//			dateToRead += (day < 10 ? "0" : "") + day;
			msg.info("Reading climate data for day: " + dateToRead);
			// Read the file
			

			try {

				// First, set all temperature stations from the file
				String line;
				if(bufferedReader == null){
					bufferedReader = new BufferedReader(new FileReader(filename));
					String firstLine = bufferedReader.readLine();
					stations = firstLine.split(",");
				}
				Map<String, Double> entryMap = new HashMap<String, Double>();
				while ((line = bufferedReader.readLine()) != null) {
					// Should only read one line...
					String[] entries = line.split(",");
					if(dateToRead.compareTo(entries[0]) == 0){
						for(int i = 1; i < stations.length; i++){
//							System.out.println("Putting station: " + stations[i] + " setting to " + Double.parseDouble(entries[i]));
							entryMap.put(stations[i], Double.parseDouble(entries[i]));
						}
						break; // Break out of loop
					}
				}
				if(entryMap.size() > 0){
					for (Object obj : context.getObjects(TemperatureStation.class)) {
						TemperatureStation ts = (TemperatureStation)obj;
						String tsID = ts.getTemperatureStationId();
						if(entryMap.containsKey(tsID)){
							double temperature = entryMap.get(tsID);
							ts.setTemperature(temperature);							
						}
						else msg.warn("TEMPERATURE STATION " + tsID + " NOT FOUND!"); // NOTE: This could be very bad, if a station is not present in the input file; throw an exception
					}
				}
				if(line == null){
					bufferedReader.close(); // Should close at end...
				}
				
				// Next, if specified by parameters, reset to all CAP Subcontractors to default temp values
				if(RunEnvironment.getInstance().getParameters().getBoolean("UseDefaultTemperature")){
					msg.info("Re-Setting CAP Subcontractors to default temperatures:");
					for(Object obj: context.getObjects(CAPSubcontractor.class)){
						CAPSubcontractor subcontractor = (CAPSubcontractor)obj;
						((TemperatureStation)(subcontractor.getSink())).setTemperature(CAPSubcontractor.getDefaultTemp(subcontractor.getId(), month - 1));
					}
				}
				
				// If specified by parameters, reset all to PHX's values...
				if(RunEnvironment.getInstance().getParameters().getBoolean("UseSingleTemperature")){
					String canonicalCity = RunEnvironment.getInstance().getParameters().getString("UseSingleTemperatureCity");
					msg.info("Canonical City is" + canonicalCity);
					double temp = -100;
					for(Object obj: context.getObjects(CAPSubcontractor.class)){
						CAPSubcontractor subcontractor = (CAPSubcontractor)obj;
						if(subcontractor.getId().compareTo(canonicalCity) == 0){
							temp = ((TemperatureStation)subcontractor.getSink()).getTemperature();
							break;
						}
					}
					if(temp != -100){
						msg.info("Re-Setting CAP Subcontractors' temperatures to Phoenix's temperature (" + temp + ")");
						for(Object obj: context.getObjects(CAPSubcontractor.class)){
							((TemperatureStation)(((CAPSubcontractor)obj).getSink())).setTemperature(temp);
						}
					}
					else{
						msg.warn("Could not find Phoenix to re-set temperature data.");
					}
				}
				

			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	// Gets the schedules for the next year
	private static class CAPPlanning implements IAction {

		private Context<Object> context;
		private TimeKeeper tk;
		
		public CAPPlanning(Context<Object> context, TimeKeeper tk) {
			this.context = context;
			this.tk      = tk;
		}

		public void execute() {
			for (Object obj : context.getObjects(CAWCD.class)) ((CAWCD)obj).getSchedules(); // Should only be one
			// Reschedule this for next year on the same day (presumably Oct 1)
			tk.registerSpecificDateAction(
					new TimeKeeper.Day(tk.getCal().get(Calendar.DAY_OF_MONTH),
							           tk.getCal().get(Calendar.MONTH),
							           tk.getCal().get(Calendar.YEAR) + 1), new TimeKeeperAction(this, CNHConstants.CAWCD_PLANNING_PRIORITY));
		}
	}

	// Monthly CAP Actions
	private static class CAPBookkeeping implements IAction {
		
		private Context<Object> context;
		private Calendar cal;

		public CAPBookkeeping(Context<Object> context, Calendar cal) {
			this.context = context;
			this.cal = cal;
		}

		public void execute() {
			Parameters params = RunEnvironment.getInstance().getParameters();
			if(params.getBoolean(CNHConstants.ALLOW_CAP_RETURN)){
				int month = cal.get(Calendar.MONTH);
				if((month > 0) && (month < 12)){
					for (Object obj : context.getObjects(CAWCD.class)) {
						((CAWCD) obj).pollForAdjustments(cal);
					}
				}
			}			
		}
	}

	
	private static class CAPSubcontractorResetMonth implements IAction {
		
		private Context<Object> context;

		public CAPSubcontractorResetMonth(Context<Object> context) {
			this.context = context;
		}

		public void execute() {
			for (Object obj : context.getObjects(CAPSubcontractor.class)) {
				((CAPSubcontractor) obj).resetMonth();
			}
		}
	}
	
	private static class CAPRollover implements IAction {

		private Context<Object> context;
		private boolean firstTime;

		public CAPRollover(Context<Object> context) {
			this.context = context;
			firstTime = true;
		}

		public void execute(){
			if(firstTime){
				firstTime = false;
				return;
			}
			for (Object obj : context.getObjects(CAWCD.class)) {
				((CAWCD) obj).reschedule();
			}
			CAPSubcontractor.resetGlobalNoise();
			for (Object obj : context.getObjects(CAPSubcontractor.class)) {
				((CAPSubcontractor) obj).newYearRollover();
			}
		}
	}	
	
	// Schedulable Action for updating the households
	// social norms via their influence sets (i.e. via their
	// geographic neighbors). This is an action rather than
	// a schedulable method so we can schedule its start and
	// interval via a parameter
	private static class UpdateSocialFactorFromNghAction implements IAction {

		Context<Object> context;

		public UpdateSocialFactorFromNghAction(Context<Object> context) {
			this.context = context;
		}

		public void execute() {
			// we don't need to randomize here as the
			// operation is unaffected by order
			for (Object obj : context.getObjects(Household.class)) {
				((Household) obj).updateSocialFactorsFromNgh();
			}
		}
	}

	private static class UpdateNoiseAction implements IAction {

		Context<?> context;
		Uniform noiseDist;

		public UpdateNoiseAction(Context<?> context, double noise) {
			this.context = context;
			noiseDist = RandomHelper.createUniform(-noise, noise);
		}

		@Override
		public void execute() {
			for (Object o : context.getObjects(Household.class)) {
				((Household) o).setWaterUsageNoise(noiseDist.nextDouble());
			}

		}

	}

	private static class UpdateSocialFactorsFromPriorUsageAction implements IAction {
		Context context;

		public UpdateSocialFactorsFromPriorUsageAction(Context context) {
			this.context = context;
		}

		@Override
		public void execute() {
			for (Object o : context.getObjects(Household.class)) {

				((Household) o).updateSocialFactorsFromUsage();
			}
		}
	}

	private static class DistributeLowFlowFixturesAction implements IAction {

		Context context;
		double  retrofitFraction;
		double  newConstFraction;
		double  retrofitReduction;
		boolean skip = false;
		Uniform sampleDist = RandomHelper.createUniform(0, 1);

		public DistributeLowFlowFixturesAction(Context context, double retrofitFraction, double newConstructionFraction, double reduction) {
			this.context = context;
			this.retrofitFraction = retrofitFraction/12d;        // Fraction is annual, applied monthly
			this.newConstFraction = newConstructionFraction/12d; // Fraction is annual, applied monthly
			this.retrofitReduction = reduction;
		}

		@Override
		public void execute() {
		  if(!skip && ((newConstFraction != 0) || (retrofitFraction != 0))){
			// Determine the fraction to be changed
			// This will do the correct number of households (if possible)
			// But will OMIT households that can't be changed 
			// (because they have already been set to the max)
			double countAll  = 0;
			double countZero = 0;
			msg.debug("COUNTING...");
            for(Object o : context.getObjects(Household.class)){
			  countAll++; 	
		      if(((Household) o).getLowFlowFixtureReduction() < 1) countZero++;
	        }
			if(countZero == 0) skip = true; // Once they have all been modified, nothing more to do
			else{
			  double totalFraction = (retrofitFraction + newConstFraction) * (countAll / countZero);
			  double retroFraction = retrofitFraction * (countAll / countZero);
			  for(Object o : context.getObjects(Household.class)){
				Household household = ((Household) o);
			    if(household.getLowFlowFixtureReduction() < 1){
			      double prob = sampleDist.nextDouble();
			      if(prob < totalFraction){
			    	  if(prob < retroFraction) household.incrementLowFlowFixtureReduction(retrofitReduction);
			    	  else                     household.incrementLowFlowFixtureReduction(1);
			      }
			    }
			  }
			}
		  }
		}
	}
	
	private static class HistogramDataRecorderAction implements IAction {

		private Context<Object> context;
		private HistogramDataRecorder recorder;
		//private boolean first = true;

		public HistogramDataRecorderAction(Context<Object> context, HistogramDataRecorder recorder) {
			this.context = context;
			this.recorder = recorder;
		}

		@Override
		public void execute() {
			// skip the first record as that occurs at the 
			// start of the simulation
			//if (!first) {
				for (Object o : context.getObjects(Household.class)) {
					recorder.record((Household) o);
				}
				try {
					recorder.writeAndReset();
				} catch (IOException ex) {
					ex.printStackTrace();
					msg.error("Error while write histogram data.", ex);
				}
			//}
			//first = false;
		}
	}

	private static class UpdateWBMWaterSources implements IAction {

		private WBMWaterSourceCollection coll;
		private Calendar cal;
		private Context<Object> context;

		public UpdateWBMWaterSources(Context<Object> context, Calendar calendar, WBMWaterSourceCollection collection) {
			cal = calendar;
			coll = collection;
			this.context = context;
		}

		@Override
		public void execute() {
			msg.debug("UPDATING WBM WATER SOURCES...");
			Date today = cal.getTime();
			Calendar previousDay = (Calendar) (cal.clone());
			previousDay.add(Calendar.DAY_OF_YEAR, -1);
			Date yesterday = previousDay.getTime();

			Calendar nextDay = (Calendar) (cal.clone());
			nextDay.add(Calendar.DAY_OF_YEAR, 1);
			Date tomorrow = nextDay.getTime();

			// Must notify WBM of the amount that was taken yesterday
			coll.notifyWBM(yesterday, today);

			// WBM returns the amount available today(NOTE: [today, tomorrow) )
			coll.notifySourcesOfAvailableWater(today, tomorrow);

			// Must reconcile any problems: if the ABM thinks there are multiple
			// sources that the WBM would consider to fall within a single
			// grid cell, check and repair here. Should only execute on first
			// run through, unless sources are added.
			if(coll.hasDuplicates()){
			  System.out.println("WBM Returned duplicate grid cells that will be reconciled.");
		      List<WBMWaterSourceManager> wbmManagers       = new ArrayList<WBMWaterSourceManager>();
		      Set<AbstractWaterManager>  dependentManagers  = new HashSet<AbstractWaterManager>(); 
				
		      // Get a list of the WBM Managers and a list of the Managers that depend on them
			  for(Object mgr: context.getObjects(AbstractWaterManager.class)){
			    if(mgr instanceof WBMWaterSourceManager) wbmManagers.add((WBMWaterSourceManager)mgr);
				else{
				  AbstractWaterManager awm = (AbstractWaterManager)mgr;
			      List<SourceFlowNode> sources = awm.getSources();
			      for(SourceFlowNode source: sources) if(source instanceof WBMWaterSourceManager) dependentManagers.add(awm); 
			      WaterSink sink = awm.getSink();
			      if(sink != null && sink instanceof WBMWaterSourceManager)	dependentManagers.add(awm);
				}
			  }
			  
			  // There is a map of sources that will be merged; need a map
			  // of sources and the managers that connect to them
			  // Can assume each source has a unique manager at this point
			  Map<WBMNaturalWaterSource, WBMWaterSourceManager> sourcesAndManagers = new HashMap<WBMNaturalWaterSource, WBMWaterSourceManager>();
			  for(WBMWaterSourceManager wbmManager: wbmManagers) sourcesAndManagers.put(wbmManager.getSource(), wbmManager);
			  
			  // Now use this map to create another map, this one of the original
			  // manager and the new manager that should be used to replace it 
			  Map<WBMWaterSourceManager, WBMWaterSourceManager> managerMap = new HashMap<WBMWaterSourceManager, WBMWaterSourceManager>();
			  for(WBMWaterSourceManager wbmManager: wbmManagers){
				  WBMWaterSourceManager replacement = sourcesAndManagers.get(coll.getDuplicate(wbmManager.getSource()));
				  if(!replacement.equals(wbmManager)) replacement.addWaterSinkAmounts(wbmManager);
				  managerMap.put(wbmManager, replacement);
			  }
			  
			  
			  // Now loop through the dependent managers and replace the old managers with the new ones
			  for(AbstractWaterManager dependentMgr: dependentManagers){
				  List<SourceFlowNode> sources = dependentMgr.getSources();
				  for(int i = 0; i < sources.size(); i++){
					  SourceFlowNode source = sources.get(i);
					  if(source instanceof WBMWaterSourceManager){
						  sources.remove(i);
						  sources.add(i, managerMap.get((WBMWaterSourceManager)source));
					  }
				  }
				  
				  WaterSink sink = dependentMgr.getSink();
				  if(sink != null && sink instanceof WBMWaterSourceManager) dependentMgr.setSink(managerMap.get((WBMWaterSourceManager)sink));
				  
			  }
			  
			  // Must also re-set the Households
			  for(Object obj: context.getObjects(Household.class)){
				  Household h = (Household)obj;
				  WaterSink sink = h.getSink();
				  if(sink != null && sink instanceof WBMWaterSourceManager) h.setSink(managerMap.get((WBMWaterSourceManager)sink));
			  }
			  
			  // At the end of this, clean up the collection
			  Map<WBMNaturalWaterSource, WBMNaturalWaterSource> sourceMap = coll.removeDuplicates();
			  for(Map.Entry<WBMNaturalWaterSource, WBMNaturalWaterSource> entry: sourceMap.entrySet()){
				  WBMNaturalWaterSource duplicateSource = entry.getKey();
				  WBMNaturalWaterSource originalSource  = entry.getValue();
				  if(!ContextCreator.mapOfDuplicatesRemoved.containsKey(originalSource)){
					  List<WBMNaturalWaterSource> newList = new ArrayList<WBMNaturalWaterSource>();
					  newList.add(duplicateSource);
					  ContextCreator.mapOfDuplicatesRemoved.put(originalSource, newList);
				  }
				  else{
					  ContextCreator.mapOfDuplicatesRemoved.get(originalSource).add(duplicateSource);
				  }
			  }
			  System.out.println("LISTING DUPLICATES:");
			  for(Map.Entry<WBMNaturalWaterSource, List<WBMNaturalWaterSource>> entry: mapOfDuplicatesRemoved.entrySet()){
				  System.out.println("ORIGINAL: " + ((TemperatureStation)entry.getKey()).getTemperatureStationId());
				  for(WBMNaturalWaterSource nws: entry.getValue()) System.out.println("    DUP: " + ((TemperatureStation)nws).getTemperatureStationId());
			  }
			}			
		}
	}

	private static class EndWBMWaterSources implements IAction {
		private WBMWaterSourceCollection coll;

		public EndWBMWaterSources(WBMWaterSourceCollection collection) {
			coll = collection;
		}

		@Override
		public void execute() {
			coll.end();
		}

	}

    private static class WriteSubcontractorStrategies implements IAction{
		
		Context context;
		
		public WriteSubcontractorStrategies(Context context){
			this.context = context; 
		}
		
		@Override
		public void execute(){
			System.out.println("WRITING SUBCONTRACTOR STRATEGIES...");
			double tick = RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			int runNumber = RunState.getInstance().getRunInfo().getRunNumber(); // Only works with batch...?
			System.out.println("RUN NUMBER: " + runNumber + " is batch: " + RunEnvironment.getInstance().isBatch() + " " + PropertiesWriter.lastRunNumberSaved);
			if(!RunEnvironment.getInstance().isBatch()){
				int pwRunNumber = PropertiesWriter.lastRunNumberSaved;
				if(pwRunNumber != PropertiesWriter.NO_RUN_NUMBER_ESTABLISHED) runNumber = pwRunNumber;
			}
			try{
				File file = new File("Strategies.csv");
				if(!file.exists()){
					Writer writer = new FileWriter("Strategies.csv", true);
					writer.write("RunNumber,Tick,ID,Pop,GPCD,AmtToObtain,AmtToReturn,OverageThresh,ShortageThresh,PriorYearScheduled,PriorYearDelivered" + System.getProperty("line.separator"));
					writer.close();
				}
				Writer writer = new FileWriter("Strategies.csv", true);
				for(Object mgr: context.getObjects(CAPSubcontractor.class)){
					CAPSubcontractor subcontractor = (CAPSubcontractor)mgr;
					writer.write(runNumber + "," + 
					             tick + "," + 
					             subcontractor.getId() + "," +
					             subcontractor.getPopulation() + "," +
					             subcontractor.getGPCD() + "," +
					             subcontractor.getAmountToObtain() + "," + 
					             subcontractor.getAmountToReturn() + "," + 
					             subcontractor.getOverageThreshold() + "," + 
					             subcontractor.getShortageThreshold() + "," +
					             subcontractor.getLastYearScheduledAF() + "," + 
					             subcontractor.getLastYearDeliveredAF() + 
					             System.getProperty("line.separator"));
				}
			
			// Write to a file along with run number
				writer.close();
			}
			catch(Exception e){
				System.err.println("Could not write Global Totals to file");
			}
		}
	}

	
	private static class WriteGlobalTotals implements IAction{
		
		Context context;
		
		public WriteGlobalTotals(Context context){
			this.context = context; 
		}
		
		@Override
		public void execute(){
			System.out.println("GETTING GLOBAL TOTALS...");
			double totalShortage = 0;
			double totalPerCapitaShortage = 0;
			double maxPerCapitaShortage = 0;
			for(Object mgr: context.getObjects(CAPSubcontractor.class)){
				CAPSubcontractor subcontractor = (CAPSubcontractor)mgr;
				totalShortage += subcontractor.getFullSimShortageYTDPerYear();
				totalPerCapitaShortage += subcontractor.getFullSimShortageYTDPerCapitaPerYear();
				double pct = subcontractor.getSimToDateShortageYTDPerCapitaFractionGPCDPerYear();
				if(pct < maxPerCapitaShortage) maxPerCapitaShortage = pct;
			}
			// Write to a file along with run number
			int runNumber = RunState.getInstance().getRunInfo().getRunNumber(); // Only works with batch...?
			if(!RunEnvironment.getInstance().isBatch()){
				int pwRunNumber = PropertiesWriter.lastRunNumberSaved;
				if(pwRunNumber != PropertiesWriter.NO_RUN_NUMBER_ESTABLISHED) runNumber = pwRunNumber;
			}
			try{
				Writer writer = new FileWriter("FinalGlobalTotals.txt", true);
				writer.write(runNumber + "," + totalShortage + "," + totalPerCapitaShortage + "," + maxPerCapitaShortage + System.getProperty("line.separator"));
				writer.close();
			}
			catch(Exception e){
				System.err.println("Could not write Global Totals to file");
			}
		}
	}
	

	private void initializeRandomDistributions() {
		Parameters p = RunEnvironment.getInstance().getParameters();
		// initialize the distribution parameters.
		MersenneTwister rndEngine = new MersenneTwister(
				p.getInteger(ParameterConstants.DEFAULT_RANDOM_SEED_USAGE_NAME));
		for (String name : p.getSchema().parameterNames()) {
			if (p.getSchema().getDetails(name).getType().equals(DistributionParameter.class)) {
				((DistributionParameter) p.getValue(name)).initialize(rndEngine);
			}
		}
	}
	
	private TimeKeeper initializeTimeKeeper() {
		Parameters p = RunEnvironment.getInstance().getParameters();
		int startYear = p.getInteger("startYear");
		Calendar calendar = GregorianCalendar.getInstance();
		calendar.set(startYear, 0, 1, 0, 0); // Note: first call to TimeKeeper does _not_ increment
		int baseCalendarIncrement = 1;
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		TimeKeeper tk = new TimeKeeper(calendar, baseCalendarIncrement);
		schedule.schedule(ScheduleParameters.createRepeating(1, 1, CNHConstants.CAL_UPDATE_PRIORITY), tk);
		return tk;
	}

	private void initializeSeasonalData(Calendar calendar) {
		Parameters p = RunEnvironment.getInstance().getParameters();
		String file = p.getString("waterUseInput");
		int endYear = p.getInteger("endYear");
		SeasonalWaterUse waterUse = null;
		if (file.trim().length() > 0) {
			SeasonalWaterUseLoader loader = new SeasonalWaterUseLoader();
			try {
				waterUse = loader.load(file, endYear, true);
				SeasonalWaterUseInstance.initialize(waterUse);
				waterUse.resetMonthlyScaleFactor(calendar);
			} catch (IOException | XMLStreamException e) {
				msg.fatal("Error while loading seasonal water use data from " + file, e);
				return;
			}
		}
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		schedule.schedule(ScheduleParameters.createRepeating(1, 1,
				CNHConstants.SEASONAL_DATA_UPDATE_PRIORITY), new UpdateSeasonalData(calendar));
	}

	private void initializeHistogramRecording(TimeKeeper tk, Context<Object> context) {
		try {
			final HistogramDataRecorder recorder = new HistogramDataRecorder();
			HistogramDataRecorderAction action = new HistogramDataRecorderAction(context, recorder);
			tk.registerMonthlyAction(new TimeKeeper.TimeKeeperAction(action,
					CNHConstants.HISTOGRAM_RECORD_PRIORITY));
			RunEnvironment.getInstance().getCurrentSchedule()
					.schedule(ScheduleParameters.createAtEnd(0), new IAction() {

						@Override
						public void execute() {
							try {
								recorder.close();
							} catch (IOException ex) {
								ex.printStackTrace();
								msg.error("Error closing histogram recorder", ex);
							}
						}
					});

		} catch (IOException ex) {
			ex.printStackTrace();
			msg.error("Error initializing histogram data recorder", ex);
		}
	}
	
	private void initializeFlowManager(Context<Object> context, Calendar cal) {
		// Create the flow manager:
		FlowManager fm = new FlowManager();
		for (Object node : context)
			fm.addFlowNode((FlowNode) node);
		
		context.add(fm);

		// Schedule the flow as an action at an interval
		IAction flowAction = new FlowAction(fm, cal, FlowFrequency.DAILY);
		RunEnvironment
				.getInstance()
				.getCurrentSchedule()
				.schedule(ScheduleParameters.createRepeating(1, 1, CNHConstants.FLOW_PRIORITY),
						flowAction);
	}
	

	/**
	 * Builds the context.
	 */
	@Override
	public Context<Object> build(Context<Object> context) {
		msg.info("Building simulation...");
		initializeRandomDistributions();
		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 1: " + RandomHelper.nextDouble());
		WaterSourceLoader.init();
		
		Parameters params = RunEnvironment.getInstance().getParameters();
		System.out.println("Writing properties; run number assigned is: " + PropertiesWriter.write("PROPS.txt", params));
		
		TimeKeeper tk = initializeTimeKeeper();
		initializeSeasonalData(tk.getCal());
		tk.registerYearlyAction(new TimeKeeper.TimeKeeperAction(new ProgressIndicator(tk.getCal(), msg),1));
		if(recordHistograms) initializeHistogramRecording(tk, context);
		if(useTucsonHouseholds){
			try {
				msg.info("Creating Households");
				createTucsonHouseholds(context, tk.getCal());
			} catch (IOException ex) {
				msg.fatal("Error while creating households", ex);
			}
		}
		
		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 2: " + RandomHelper.nextDouble());
		
		if(useSocialNetwork){
			// this will create a network out of whatever is in the context
			// so its important that only the households are in the context
			// at this point
			msg.info("Creating Household Network");
			createTucsonHouseholdNetwork(context);
			
			// these must be created after the households
			msg.info("Creating Water Managers");
			NetworkBuilder<Object> netBuilder = new NetworkBuilder<Object>("flowNetwork", context, true);
			netBuilder.buildNetwork();
		}

		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 3: " + RandomHelper.nextDouble());
		if(useCAPCanal) createCAPCanal(context, tk);
		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 4: " + RandomHelper.nextDouble());

		if(useTucsonHouseholds){
			// Subscribe residences to residential water manager:
			// Also add sink
			WaterSink tucsonOutflow = null; 
			String tucsonName = WaterSourceLoader.nameMap.get("Tucson, City of");
			System.out.println("Trying to find Tucson: " + tucsonName);
			for (Object obj : context){
				if(obj instanceof WaterSink){
					WaterSink sink = (WaterSink)obj;
					if(sink.getId().compareTo(tucsonName) == 0) tucsonOutflow = sink;
				}
			}
			ResidentialWaterManager mgr = (ResidentialWaterManager)context.getObjects(ResidentialWaterManager.class).get(0);
			for (Object agent : context){
				if (agent instanceof Household){
					Household household = (Household)agent;
					msg.info("Adding household " + household.getId() + " to manager " + mgr.getId());
					mgr.addWaterRecipient(household);
					household.setSink(tucsonOutflow); // Doesn't matter if it's null
				}
			}
			// Set up billing (no free lunches!)
			BillAction billAction = new BillAction(tk.getCal());
			billAction.addManager(mgr);
			tk.registerMonthlyAction(new TimeKeeper.TimeKeeperAction(billAction, CNHConstants.BILL_PRIORITY));
			
			IAction usf = new UpdateSocialFactorsFromPriorUsageAction(context);
			TimeKeeper.TimeKeeperAction tka3 = new TimeKeeper.TimeKeeperAction(usf,
					CNHConstants.UPDATE_FROM_PRIOR_USAGE_PRIORITY);
			tk.registerYearlyAction(tka3);
			
			// schedule social norm update for each household
			ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
			double interval = RunEnvironment.getInstance().getParameters()
					.getDouble(CNHConstants.NORM_UPDATE_INTERVAL);
			schedule.schedule(ScheduleParameters.createRepeating(interval, interval,
					CNHConstants.UPDATE_FROM_NGH_PRIORITY),
					new UpdateSocialFactorFromNghAction(context));
			
			
		}
		
		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 5: " + RandomHelper.nextDouble());
		if(useMessaging){
			// schedule messaging events
			msg.info("Scheduling Messaging Events");
			scheduleMessagingEvents(context, tk);

		}

		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 6: " + RandomHelper.nextDouble());
		
		double noise = params.getDouble(CNHConstants.WATER_USAGE_NOISE);
		IAction updateNoise = new UpdateNoiseAction(context, noise);
		TimeKeeper.TimeKeeperAction tka4 = new TimeKeeper.TimeKeeperAction(updateNoise,
				CNHConstants.UPDATE_NOISE_PRIORITY);
		tk.registerMonthlyAction(tka4);

		IAction dlff = new DistributeLowFlowFixturesAction(context, 
				params.getDouble(CNHConstants.LOW_FLOW_ANNUAL_FRACTION),
				params.getDouble(CNHConstants.NEW_CONST_ANNUAL_FRACTION),
				params.getDouble(CNHConstants.LOW_FLOW_INCREMENT));
		TimeKeeper.TimeKeeperAction tka5 = new TimeKeeper.TimeKeeperAction(dlff,
				CNHConstants.DISTRIBUTE_LOW_FLOW_PRIORITY);
		tk.registerMonthlyAction(tka5);


		
		double stopAt = params.getDouble(CNHConstants.STOP_AT);
		if (stopAt > 0) {
			RunEnvironment.getInstance().endAt(stopAt);
		}
		
		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 7: " + RandomHelper.nextDouble());
		// Write the CAP Subcontractor strategies
		tk.registerYearlyAction(new TimeKeeper.TimeKeeperAction(new WriteSubcontractorStrategies(context), CNHConstants.CAP_SUB_STRATEGY_PRIORITY));
		
//		IAction updateTemps = new UpdateClimateData(context, "MonthlyAvgTempRecord_NORMAL_12YR_FINAL.csv", tk);
//		IAction updateTemps = new UpdateClimateData(context, "MonthlyAvgTempRecord_NORMAL_TEST.csv", tk);
//		IAction updateTemps = new UpdateClimateData(context, "MonthlyAvgTempRecord_NORMAL_12YR_REV_EXTENDED.csv", tk);
//		TimeKeeper.TimeKeeperAction tka6 = new TimeKeeper.TimeKeeperAction(updateTemps, CNHConstants.UPDATE_CLIMATE_PRIORITY);
//		tk.registerMonthlyAction(tka6);
		
//		IAction recordTemps = new RecordClimateData(context, "MonthlyAvgTempRecord_NORMAL_12YR_REV.csv", tk);
//		TimeKeeper.TimeKeeperAction tka7 = new TimeKeeper.TimeKeeperAction(recordTemps, CNHConstants.RECORD_CLIMATE_PRIORITY);
//		tk.registerMonthlyAction(tka7);
		
		
		// Schedule the final write for the simulation end
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		schedule.schedule(ScheduleParameters.createAtEnd(0.0), new WriteGlobalTotals(context));
		
		msg.info("GENERATING RANDOM NUMBER JUST FOR TEST AFTER INITIALIZE 8: " + RandomHelper.nextDouble());
		return context;
	}


	private Map<String, NaturalWaterSource> createWaterSources(Context<Object> context, Calendar cal, Calendar calendarStart, Calendar calendarEnd){ // For now assume just one...
		msg.info("CREATING WATER SOURCES: " + cal.getTime());
		// Get a parameter indicating whether the WBM will be used, and if so at what interval

		Parameters p = RunEnvironment.getInstance().getParameters();
		int i = p.getInteger(CNHConstants.WBM_INTERVAL);

		Map<String, NaturalWaterSource> waterSources = new HashMap<String, NaturalWaterSource>();

		if(i == -1){ // Basic flow with no WBM Interaction
			msg.info("Establishing basic water source...");

			Date start = calendarStart.getTime();
			Date end = calendarEnd.getTime();

			FlowHydrograph basicNaturalSource = new FlowHydrograph(start, end, 5200000,	FlowRateUnits.GPD);

			// Some more water in the summers
			Calendar summerStart = (Calendar) calendarStart.clone();
			summerStart.add(Calendar.MONTH, 6);
			Calendar summerEnd = (Calendar) summerStart.clone();
			summerEnd.add(Calendar.MONTH, 3);

			while (summerStart.getTime().compareTo(calendarEnd.getTime()) < 0) {
				basicNaturalSource.add(new FlowHydrograph(summerStart.getTime(), summerEnd
						.getTime(), 600, FlowRateUnits.GPD));
				summerStart.add(Calendar.YEAR, 1);
				summerEnd.add(Calendar.YEAR, 1);
			}

			NaturalWaterSource nws = new NaturalWaterSource("CAP_Start", basicNaturalSource);
			waterSources.put("CAP_Start", nws);

		}
		else{
			WBMCell.reinitialize();
			msg.info("Establishing WBM water sources...");
			WBMWaterSourceCollection wsColl = new WBMWaterSourceCollection(p.getString(CNHConstants.WBM_URL));
			
			WaterSourceLoader.load("./data/CAPStartWaterSource.xml", context, wsColl);
			if(useCAPCustomerWaterSources) WaterSourceLoader.load("./data/CAPCustomerWaterSources.xml", context, wsColl, allowSourcesToManageIrrigation);
			if(useCAPGridWaterSources){
				WaterSourceLoader.create(34.00, -114.0, 33.5, -113.3, .1, .1, context, wsColl, allowSourcesToManageIrrigation);
				WaterSourceLoader.create(34.00, -113.3, 32.3, -111.1, .1, .1, context, wsColl, allowSourcesToManageIrrigation);
				
				
//				WaterSourceLoader.create(34.75, -114.4, 33.75, -113.9, .1, .1, context, wsColl, allowSourcesToManageIrrigation);
//				WaterSourceLoader.create(34.75, -113.9, 32.75, -111.9, .1, .1, context, wsColl, allowSourcesToManageIrrigation);
//				WaterSourceLoader.create(34.00, -111.9, 31.25, -110.1, .1, .1, context, wsColl, allowSourcesToManageIrrigation);
			}
			
			WBMCrop.useWBMDemand = !cropsReturnZeroInAllCases;
						
			// Schedule the WBM Updating
			if(p.getInteger(CNHConstants.WBM_HOST) != -2){ // Use -2 for reading temperature data...
				IAction wbmAction = new UpdateWBMWaterSources(context, cal, wsColl);
				RunEnvironment
						.getInstance()
						.getCurrentSchedule()
						.schedule(ScheduleParameters.createRepeating(1, 1, CNHConstants.WBM_PRIORITY),
								wbmAction);
				IAction wbmEnd = new EndWBMWaterSources(wsColl);
				RunEnvironment.getInstance().getCurrentSchedule()
						.schedule(ScheduleParameters.createOneTime(ScheduleParameters.END), wbmEnd);
				int wbm = p.getInteger(CNHConstants.WBM_HOST);
				switch(wbm){
				  case 1:{
					  msg.info("USING WBM ONLINE AT " + p.getString(CNHConstants.WBM_URL) );
					  // Do nothing: This means there will be a real WBM Server out there, at the specified URL
					  break;
				  }
				  case 0:
				  default:{
					  msg.info("USING WBM PROXY AT " + p.getString(CNHConstants.WBM_URL));
	//				  ((WBMNaturalWaterSource)waterSources.get("CAP_Start_Source")).setDefaultFlow(1000);
	//				  ((WBMNaturalWaterSource)waterSources.get("TempeReturnPoint")).setDefaultFlow(0);
					  ProxyRunner pRunner = new ProxyRunner(wsColl);
					  new Thread(pRunner).start();
				  }
				}
			}
		}
		
		msg.info("Adding to context: " + waterSources.size());
		for(NaturalWaterSource source: waterSources.values()){
			context.add(source);
		}
		
		return waterSources;
	}

	private void createCAPCanal(Context<Object> context, TimeKeeper tk){
		Calendar calendarStart = (Calendar) tk.getCal().clone();
		Calendar calendarEnd = (Calendar) calendarStart.clone();
		calendarEnd.add(Calendar.YEAR, 5);
		
		Parameters params = RunEnvironment.getInstance().getParameters();
		
		// Create the water sources
		createWaterSources(context, tk.getCal(), calendarStart, calendarEnd);

		// Load the network of water managers
		ManagerNetworkLoader.load("./data/LakeHavasuManager.xml", context, tk);
		if(useCAPNetwork){
			ManagerNetworkLoader.load("./data/CAPNetwork.xml", context, tk);
			SourceFlowNode cawcd = null;
			for(Object obj: context.getObjects(SourceFlowNode.class)){
				SourceFlowNode node = (SourceFlowNode)obj;
				if(node.getId().compareTo("CAWCD") == 0){
					cawcd = node;
					break;
				}
			}
            if(cawcd==null) msg.debug("Could not find CAWCD in context.");
            if(useCAPSubcontractors){
				CAPSubcontractor.globalNoise = 0; // For reinitialize
				String subcontractorNetworkFile = params.getString("CAPSubcontractorFile");
				ManagerNetworkLoader.load("./data/" + subcontractorNetworkFile, context, tk);
			    // On Oct 1, the CAWCD should poll all Subcontractors for next year's schedules
			    IAction cap_plan = new CAPPlanning(context, tk);
			    TimeKeeper.TimeKeeperAction tka1 = new TimeKeeper.TimeKeeperAction(cap_plan, CNHConstants.CAWCD_PLANNING_PRIORITY);
			    int firstYear = tk.getCal().get(Calendar.YEAR) + ((tk.getCal().get(Calendar.MONTH) < 9) ? 0 : 1);
			    tk.registerSpecificDateAction(new TimeKeeper.Day(1, 9, firstYear), tka1);
			    
				// On the first of each month, the CAP Subcontractors need to update their noise values
				// and do other bookkeeping
				IAction cap_book = new CAPBookkeeping(context, tk.getCal());
				TimeKeeper.TimeKeeperAction tka2 = new TimeKeeper.TimeKeeperAction(cap_book,
						CNHConstants.CAWCD_BOOKKEEPING_PRIORITY);
				tk.registerMonthlyAction(tka2);
				
				IAction cap_sub_month_reset = new CAPSubcontractorResetMonth(context);
				TimeKeeper.TimeKeeperAction tka_sub_reset = new TimeKeeper.TimeKeeperAction(cap_sub_month_reset,
						CNHConstants.CAP_SUBCONTRACTOR_MONTH_RESET_PRIORITY);
				tk.registerMonthlyAction(tka_sub_reset);
				
				
				// On Jan 1, the CAWCD should perform its rollover, and all subcontractors should do theirs
				IAction cap_roll = new CAPRollover(context);
				TimeKeeper.TimeKeeperAction tka3 = new TimeKeeper.TimeKeeperAction(cap_roll,
						CNHConstants.CAWCD_ROLLOVER_PRIORITY);
				tk.registerYearlyAction(tka3);
			}
		
		
			// Hack: All managers draw directly from Havasu
			if(useDataDriven){
				msg.info("Using data driven");
				if(this.useCAPSubcontractors){
					DataDrivenDemandWaterManager.idsToSkip.add("Avondale, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Peoria, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Glendale, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Surprise, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Chandler, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Chandler, City Of"); // Fix in original spreadsheet!
					DataDrivenDemandWaterManager.idsToSkip.add("Scottsdale, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Cave Creek, Town of");
					DataDrivenDemandWaterManager.idsToSkip.add("Goodyear, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Arizona Water Company");
					DataDrivenDemandWaterManager.idsToSkip.add("Apache Junction Water Utilities Community Facilities District");
					DataDrivenDemandWaterManager.idsToSkip.add("Phoenix, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Mesa, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Gilbert, Town of");
					DataDrivenDemandWaterManager.idsToSkip.add("Tempe, City of");
					DataDrivenDemandWaterManager.idsToSkip.add("Carefree Water Co");
				}
	
				DataDrivenDemandWaterManager.initialize("./data/CAP_DELIVERIESvr.xls", context);
				for(DataDrivenDemandWaterManager mgr: DataDrivenDemandWaterManager.instances){
					mgr.addWaterSource(cawcd);
				}
				DataDrivenDemandWaterManager.includeAgriculturalDemand = useDataDrivenAG;
				DataDrivenDemandWaterManager.includeMunicipalDemand    = useDataDrivenMI;
				DataDrivenDemandWaterManager.includeRechargeDemand     = useDataDrivenRX;
			}
			
		}
		// Initialize the flow manager with all flow nodes now in place
		initializeFlowManager(context, tk.getCal());
	}	

	private void createTucsonHouseholds(Context<Object> context, Calendar cal) throws IOException {
		Climate climate;
		try {
			climate = new ClimateLoader().load("data/tucson_climate.xml");
			climate.setCalendar(cal);
		} catch (XMLStreamException xmlStreamException) {
			climate = new Climate();
		}

		Parameters p = RunEnvironment.getInstance().getParameters();

		ClimateImpacts climateImpacts = new ClimateImpacts(
				p.getDouble(CNHConstants.CLIMATE_TEMP_IMPACT),
				p.getDouble(CNHConstants.CLIMATE_PRECIP_IMPACT));

		String file = p.getString(CNHConstants.HOUSEHOLD_INPUT_FILE);
		HouseholdLoader creator = new HouseholdLoader(file, new SampleFilter(
				p.getDouble(CNHConstants.HOUSEHOLD_SAMPLE_PROBABILITY)), // new LimitFilter(20),  new AlwaysPassFilter(),
				new DefaultHouseholdInitializer(p), cal, climate, climateImpacts);

		InfluenceSetCreator infCreator = new InfluenceSetCreator();
		int count = 0;
		while (creator.hasNext()) {
			Household h = creator.nextHousehold();
			context.add(h);
			count++;
			infCreator.addHousehold(h);
		}
		msg.info("Created " + count + " households");

		double socWeight = p.getDouble(CNHConstants.SOCIAL_WEIGHT_PARAMETER);
		if (socWeight > 0) {
			double distance = p.getDouble(CNHConstants.HOUSEHOLD_NGH_DISTANCE);
			infCreator.run(distance);
		}
		creator.close();
	}

	private void createTucsonHouseholdNetwork(Context<Object> context) {
		Parameters p = RunEnvironment.getInstance().getParameters();
		double socWeight = p.getDouble(CNHConstants.SOCIAL_WEIGHT_PARAMETER);

		if (socWeight > 0) {
			NetworkBuilder<Object> builder = new NetworkBuilder<Object>(
					CNHConstants.HOUSEHOLD_NETWORK_ID, context, true);
			int m = p.getInteger(CNHConstants.KE_NETWORK_M);
			double alpha = p.getDouble("waterUseAlpha");
			double lambda = p.getDouble("waterUseLambda");

			builder.setGenerator(new KENetworkGenerator(m, alpha, lambda));
			builder.buildNetwork();
		}
	}

	
	
	private void scheduleMessagingEvents(Context<Object> context, TimeKeeper tk) {
		// Read relevant parameters
		Parameters params = RunEnvironment.getInstance().getParameters();
		int beginYear = params.getInteger("messagingBeginYear");
		int beginMonth = params.getInteger("messagingBeginMonth");
		
		int interval = params.getInteger("messagingInterval");
		int recurrence = params.getInteger("messagingRecurrence");
		
		double waterUseTop = params.getDouble("waterUseTop");
		double waterUseBottom = params.getDouble("waterUseBottom");
		double messagingIntensity = params.getDouble("messagingIntensity");
		double messagingEffectiveness = params.getDouble("messagingEffectiveness");
		double messagingSubsampling = params.getDouble("messagingSubsampling");

		boolean messagingSameHouseholds = params.getBoolean("messagingSameHouseholds");

		final HouseholdMessager hm = new HouseholdMessager(waterUseTop, waterUseBottom,
				messagingIntensity, messagingEffectiveness, messagingSubsampling,
				messagingSameHouseholds);
		hm.initialize(context);

		IAction messagingIAction = new IAction() {

			@Override
			public void execute() {
				hm.message();

			}
		};
		
		TimeKeeper.TimeKeeperAction tka = new TimeKeeper.TimeKeeperAction(messagingIAction, CNHConstants.UPDATE_FROM_MESSAGING_PRIORITY);
		
		Calendar endC = GregorianCalendar.getInstance();
		endC.set(params.getInteger("startYear"), 0, 1);
		endC.add(Calendar.DAY_OF_YEAR, params.getDouble("stopAt").intValue() - 1);
		CalendarEndChecker cec = new CalendarEndChecker(endC);
		

		Calendar calendar = GregorianCalendar.getInstance();
		calendar.set(beginYear, beginMonth, 1, 0, 0);
		for (int i = 0; i < recurrence; i++) {
			int date = calendar.get(Calendar.DATE);
			int month = calendar.get(Calendar.MONTH);
			int year = calendar.get(Calendar.YEAR);
			Day day = new Day(date, month, year);
			if (cec.pastOrEqualsEnd(calendar)){
				break;
			}
			msg.debug("Scheduling messaging for day: " + day);
			tk.registerSpecificDateAction(day, tka);
			calendar.add(Calendar.MONTH, interval);
		}

	}

}
