package cnh.crb;

import java.io.File;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.stream.XMLStreamException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.Network;
import simphony.util.messages.MessageCenter;
import cnh.crb.agent.AbstractWaterManager;
import cnh.crb.agent.BasicWaterManager;
import cnh.crb.agent.ConstantDemandWaterManager;
import cnh.crb.agent.DataDrivenDemandWaterManager;
import cnh.crb.agent.ResidentialWaterManager;
import cnh.crb.econ.HouseholdBillCalculator;
import cnh.crb.econ.io.RateLoader;
import cnh.crb.econ.io.SimpleCostLoader;
import cnh.crb.hydro.CanalSegmentManager;
import cnh.crb.hydro.CylindricalStorage;
import cnh.crb.hydro.FlowRate;
import cnh.crb.hydro.Reservoir;
import cnh.crb.hydro.SourceFlowNode;
import cnh.crb.hydro.Storage;
import cnh.crb.hydro.TrapezoidalCanalSegment;
import cnh.crb.hydro.WaterSink;
import cnh.crb.regional.CAPSubcontractor;
import cnh.crb.regional.CAWCD;
import cnh.crb.regional.tucson.TucsonHouseholdSewerChargeCalculatorLoader;
import cnh.crb.regional.tucson.TucsonSummerSurchargeLoader;

public class ManagerNetworkLoader {

	static MessageCenter msg = MessageCenter.getMessageCenter(ManagerNetworkLoader.class);
	
	private static HouseholdBillCalculator createCalculator(Parameters params) {
		HouseholdBillCalculator calculator = new HouseholdBillCalculator();
		try {
			calculator.addCalculator(new RateLoader().load(params.getString(CNHConstants.RATES_FILE)));
			calculator.addCalculator(new TucsonSummerSurchargeLoader(10).load(params.getString(CNHConstants.SURCHARGE_FILE)));
			calculator.addCalculator(new SimpleCostLoader().load(params.getString(CNHConstants.MONTHLY_CHARGES_FILE)));
			calculator.addCalculator(new SimpleCostLoader().load(params.getString(CNHConstants.CAP_CHARGES_FILE)));
			calculator.addCalculator(new SimpleCostLoader().load(params.getString(CNHConstants.CONSERVE_CHARGES_FILE)));
			calculator.addCalculator(new SimpleCostLoader().load(params.getString(CNHConstants.SEWER_RATE_FILE)));
			calculator.addCalculator(new TucsonHouseholdSewerChargeCalculatorLoader().load(params.getString(CNHConstants.SEWER_MONTHLY_CHARGES_FILE)));
		} catch (IOException | XMLStreamException ex) {
			msg.error("Error creating household calculator", ex);
		}

		return calculator;
	}

	
	public static Storage getStorage(Date start, Element element){
		NodeList canalList = element.getElementsByTagName("canal");
		if(canalList.getLength() > 0){
			Element canalNode = (Element)(canalList.item(0));
			String canalCrossSection = canalNode.getAttribute("crossSection");
			  switch(canalCrossSection){
			    case "trapezoidal":
			    default:
			    	double topWidth = Double.valueOf(canalNode.getAttribute("topWidth"));
			    	double bottomWidth = Double.valueOf(canalNode.getAttribute("bottomWidth"));
			    	double channelHeight = Double.valueOf(canalNode.getAttribute("channelHeight"));
			    	double channelLength = Double.valueOf(canalNode.getAttribute("channelLength"));
			    	String id            = canalNode.getAttribute("name");
			    	TrapezoidalCanalSegment seg = new TrapezoidalCanalSegment(start, topWidth, bottomWidth, channelHeight, channelLength, id);
			    	seg.setVolumeStored(seg.getMaxVolume() * Double.valueOf(canalNode.getAttribute("initialPctFull")));
			    	return seg;
			  }
		}
		NodeList reservoirList = element.getElementsByTagName("reservoir");
		if(reservoirList.getLength() > 0){
			Element reservoirNode = (Element)(reservoirList.item(0));
			String reservoirShape = reservoirNode.getAttribute("shape");
			String id             = reservoirNode.getAttribute("name");
			switch(reservoirShape){
			  case "cylindrical":
			  default:
				  double altitudeAtBottom = Double.valueOf(reservoirNode.getAttribute("altitudeAtBottom"));
				  double radius = Double.valueOf(reservoirNode.getAttribute("radius"));
				  double depth = Double.valueOf(reservoirNode.getAttribute("depth"));
				  Reservoir reservoir = new Reservoir(altitudeAtBottom, new CylindricalStorage(start, radius, depth, id));
				  reservoir.setVolumeStored(reservoir.getMaxVolume() * Double.valueOf(reservoirNode.getAttribute("initialPctFull")));
			      return reservoir;
			}
		}
		return null;
	}
	
	public static void load(String filename, Context<Object> context, TimeKeeper tk){ 
		msg.info("LOADING " + filename);
		Map<String, WaterSink> sinks = new HashMap<String, WaterSink>();
	    for(Object sink: context.getObjects(WaterSink.class)){
		  WaterSink s = (WaterSink)sink;
		  sinks.put(s.getId(), s);
	    }
		
		Parameters params = RunEnvironment.getInstance().getParameters();
		
		Set<String> subcontractorsToScale = new HashSet<String>();
		String subsToScaleList = params.getString("CAPSubcontractorsToScale");
		if(subsToScaleList.length() > 0){
			String[] subsToScale = params.getString("CAPSubcontractorsToScale").split(",");
			for(String subToScale: subsToScale){
				msg.info("Will rescale " + subToScale);
				subcontractorsToScale.add(subToScale);
			}
		}
		double subScaleValue = params.getDouble("CAPSubcontractorsScaleValue");
		
		
		
		try {
			Calendar calendarStart = (Calendar) tk.getCal().clone();
			Calendar calendarEnd = (Calendar) calendarStart.clone();
			calendarEnd.add(Calendar.YEAR, 5);
			Date start = calendarStart.getTime();

			File inputFile = new File(filename);
			DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
			Document doc = docBuilder.parse(inputFile);
		 
			doc.getDocumentElement().normalize();
			Node root = doc.getElementsByTagName("NodeCollection").item(0);
			NodeList nodeList = root.getChildNodes();
			
			
			
			for(int i = 0; i < nodeList.getLength(); i++){
				Node node = nodeList.item(i);
				if(node.getNodeType() == Node.ELEMENT_NODE){
					Element element = (Element) node;
					String type = element.getAttribute("type");
					// Only two types: definitions of managers and redirects to sub-files:
					switch(type){
					  case "manager":
						AbstractWaterManager mgr = null;
						// Managers have storage; create the storage here
						Storage storage = getStorage(start, element); 
											  
						String managerType = element.getAttribute("managerType");
						String managerId   = element.getAttribute("name");
						msg.info("Loading manager type: {}", managerType);
						switch(managerType){
						  case "canal":
							  Element canalNode = (Element)(element.getElementsByTagName("canal").item(0));
							  String canalCrossSection = canalNode.getAttribute("crossSection");
							  switch(canalCrossSection){
							    case "trapezoidal":
							    default:
							    	mgr = new CanalSegmentManager((TrapezoidalCanalSegment)storage, managerId);
							  }
							  break;
						  case "residential":
							  msg.debug("CREATING A RESIDENTIAL WATER MANAGER " + managerId);
 							  HouseholdBillCalculator calc = createCalculator(params);
 							  mgr = new ResidentialWaterManager((Reservoir)storage, calc, calendarStart, managerId);
							  break;
						  case "constant":
						      double demand = Double.valueOf(element.getAttribute("constantRate"));
						      mgr = new ConstantDemandWaterManager(storage, managerId, demand, FlowRate.FlowRateUnits.AFY);
						      break;
						  case "dataDriven":
							  mgr = new DataDrivenDemandWaterManager(storage, managerId);
							  break;
						  case "CAWCD":
							  mgr = new CAWCD(storage, managerId);
							  break;
						  case "CAPSubcontractor":
							  msg.debug("IN CAP Subcontractor");
							  CAPSubcontractor.globalNoiseRange = params.getDouble("waterUsageNoise");
							  msg.debug("globalNoiseRange = " + CAPSubcontractor.globalNoiseRange);
							  CAPSubcontractor.initGlobalNoise(); // This gets called multiple times but 

							  double pop       = Double.parseDouble(element.getAttribute("pop"));
							  
							  if(subcontractorsToScale.contains(managerId) || subcontractorsToScale.contains("ALL")){
								  msg.info("Rescaling " + managerId + " from " + pop + " by " + subScaleValue + " to: " + pop * subScaleValue);
								  pop *= subScaleValue;
							  }
							  
							  double popGrowth = Double.parseDouble(element.getAttribute("popGrowth")); 
							  double GPCD      = Double.parseDouble(element.getAttribute("GPCD"));
							  double GPCDTrend = Double.parseDouble(element.getAttribute("GPCDTrend"));
							  double noisiness = Double.parseDouble(element.getAttribute("noiseRange"));
							  // Can be overridden by parameters
							  if(params.getSchema().contains(managerId + "_noisiness"))          noisiness = params.getDouble(managerId + "_noisiness");
							  
							  String intercepts = element.getAttribute("intercepts");
							  String slopes     = element.getAttribute("slopes");
							  mgr = new CAPSubcontractor(storage, managerId, pop, popGrowth, GPCD, GPCDTrend, noisiness, tk.getCal(), intercepts, slopes);
							  msg.debug("Created water manager {}", mgr.getId());

							  CAPSubcontractor capMgr = (CAPSubcontractor)mgr;
							  
							  String capSubStrat = params.getString("CAPSubsStrategies");
							  switch(capSubStrat){
								  case("PARAMS"):{
									  capMgr.setOverageThreshold(params.getDouble(managerId + "_overageThreshold"));
									  capMgr.setShortageThreshold(params.getDouble(managerId + "_shortageThreshold"));
									  capMgr.setAmountToObtain(params.getDouble(managerId + "_amtToObtain"));
									  capMgr.setAmountToReturn(params.getDouble(managerId + "_amtToReturn"));
									  break;
								  }
								  case("CLASS"):{
									  String stratClass = element.getAttribute("stratClass"); // Note: will fail if this is not in the CAPSubcontractor xml file used
									  capMgr.setAmountToObtain(params.getDouble(stratClass + "_amtToObtain"));
									  capMgr.setAmountToReturn(params.getDouble(stratClass + "_amtToReturn"));
									  msg.info("SETTING " + mgr.getId() + " OF STRAT CLASS " + stratClass + " to " +params.getDouble(stratClass + "_amtToObtain") + "," + params.getDouble(stratClass + "_amtToReturn"));
									  // Note: these are kept constant
									  capMgr.setOverageThreshold(params.getDouble("CAPSubcontractorOverageThreshold"));
									  capMgr.setShortageThreshold(params.getDouble("CAPSubcontractorShortageThreshold"));
									  break;
								  }
								  case("USEVAR"):{
									  capMgr.setOverageThreshold(params.getDouble("CAPSubcontractorOverageThreshold"));
									  capMgr.setShortageThreshold(params.getDouble("CAPSubcontractorShortageThreshold"));
									  double AmtObtainAdj = params.getDouble("CAPSubAmtObtainVar");
									  if(AmtObtainAdj > 0) AmtObtainAdj = AmtObtainAdj * (RandomHelper.nextDouble() - .5);
									  capMgr.setAmountToObtain(Math.max(params.getDouble("CAPSubcontractorAmtToObtain") + AmtObtainAdj, .03));
									  msg.debug("AmountToObtain Adjust = " + AmtObtainAdj + " final value will be: " + Math.max(params.getDouble("CAPSubcontractorAmtToObtain") + AmtObtainAdj, .03));  
									  double AmtReturnAdj = params.getDouble("CAPSubAmtReturnVar");
									  if(AmtReturnAdj > 0) AmtReturnAdj = AmtReturnAdj * (RandomHelper.nextDouble() -.5);
									  msg.debug("AmountToReturn Adjust = " + AmtReturnAdj + " final value will be: " + Math.max(params.getDouble("CAPSubcontractorAmtToReturn") + AmtReturnAdj, .03));
									  capMgr.setAmountToReturn(Math.max(params.getDouble("CAPSubcontractorAmtToReturn") + AmtReturnAdj, .03));
									  break;
								  }
								  case("EQUAL"):
								  default:{
									  capMgr.setOverageThreshold(params.getDouble("CAPSubcontractorOverageThreshold"));
									  capMgr.setShortageThreshold(params.getDouble("CAPSubcontractorShortageThreshold"));
									  capMgr.setAmountToObtain(params.getDouble("CAPSubcontractorAmtToObtain"));
									  capMgr.setAmountToReturn(params.getDouble("CAPSubcontractorAmtToReturn"));
									  break;
								  }
							  }
							  							  
							  String sinkName  = element.getAttribute("sink");
							  msg.debug("TRYING TO FIND SINK: " + sinkName);
							  if(sinkName != null){
								  sinkName = WaterSourceLoader.nameMap.get(sinkName); // Look up 
								  WaterSink sink = sinks.get(sinkName);
								  if(sink != null){
									  msg.debug("Setting sink for " + mgr.getId() + " to " + sink.getId());
									  mgr.setSink(sink);
								  }  
							  }
							  break;
						  case "basic":
						  default:
						    mgr = new BasicWaterManager(storage, managerId);
						}
						
						if(mgr != null){
						  context.add(mgr);
						  String sourceNodeId = element.getAttribute("sourceId");
						  if(sourceNodeId != null && sourceNodeId.length() > 0){
						    for(Object sfnode: context.getObjects(SourceFlowNode.class)){
						      SourceFlowNode source = (SourceFlowNode)sfnode;
							  String nodeID = source.getId();
							  if(nodeID.compareTo(sourceNodeId) == 0){
							    mgr.addWaterSource(source);
							    // Can set a network connection here
							    Network<Object> net = (Network<Object>)context.getProjection("flowNetwork");
							    if(net != null){
							    	net.addEdge(source,  mgr);
							    }
							    break;
							  }
						    }
						  }
						  

						  // If there is a sink:
						  String sinkId = element.getAttribute("sinkId");
						  if(sinkId != null && sinkId.length() > 0){
							for(Object sinkObj: context.getObjects(WaterSink.class)){
						      WaterSink sink = (WaterSink)sinkObj;
							  String nodeID = sink.getId();
							  if(nodeID.compareTo(sinkId) == 0){
							    mgr.setSink(sink);
							    break;
							  }
						    }
						  }
						  
						}
						break; // End 'manager' switch
					  case "file":
					  default:
						String fileToLoad = element.getAttribute("name");
						load(fileToLoad, context, tk);
					}
				}
			}
			msg.info("DONE CREATING WATER MANAGERS from " + filename);
		} catch (Exception e) {
			msg.error("ERROR OCCURRED IN CREATEWATERMANAGERS for " + filename, e);
		}
		

	}
	
}
