package cnh.crb;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

/**
 * Loads the SeasonalOffsets from an XML file.
 * 
 * @author nick
 */
public class SeasonalWaterUseLoader {
	
	private static final QName YEAR = new QName("year");
	private static final QName DATE = new QName("date");
	private static final QName MEAN = new QName("mean");
	private static final QName MONTH = new QName("month");
	private static final QName FRAC_OF_MEAN = new QName("fracOfMean");

	// Private static classes
	
	private static class SingleYearSeasonalWaterUse implements SeasonalWaterUse{

		Year year;
		float offset = -1;
		
		public SingleYearSeasonalWaterUse(Year y){
			year = y;
		}
		
		@Override
		public float getYearlyMean(Calendar date) {
			return year.mean;
		}

		@Override
		public void resetMonthlyScaleFactor(Calendar date) {
			int month = date.get(Calendar.MONTH);
			offset = year.months[month];
		}

		@Override
		public float getMonthlyScaleFactor() {
			return offset;
		}
		
		public float getMonthlyScaleFactorFor(Calendar date) {
			int month = date.get(Calendar.MONTH);
			return year.months[month];
		}
			
		
	}

	
	private static class SO implements SeasonalWaterUse {
		
		int year0;
		int endingYear;
		float offset = -1;
		List<Year> years = new ArrayList<>();
		
		public SO(int endingYear) {
			this.endingYear = endingYear;
		}
		
		public void calcYear0() {
			year0 = Integer.MAX_VALUE;
			for (Year year : years) {
				if (year.year < year0) year0 = year.year;
			}
		}
		
		private int getYearIndex(Calendar date) {
			int year = date.get(Calendar.YEAR);
			if (year > endingYear){
				year = endingYear;
			}
			int yearIndex = year - year0;
			// by default just keep using the last year for which
			// we have data, when we don't have data through the ending year.
			if (yearIndex >= years.size()) yearIndex = years.size() - 1;
			return yearIndex;	
		}
		
		/* (non-Javadoc)
		 * @see cnh.crb.SeasonalWaterUse#getYearlyMean(java.util.Calendar)
		 */
		@Override
		public float getYearlyMean(Calendar date) {
			int yearIndex = getYearIndex(date);
			Year year = years.get(yearIndex);
			return year.mean;
		}

		@Override
		public void resetMonthlyScaleFactor(Calendar date) {
			int yearIndex = getYearIndex(date);
			Year year = years.get(yearIndex);
			// jan is 0
			int month = date.get(Calendar.MONTH);
			offset = year.months[month];
		}
		
		
		/* (non-Javadoc)
		 * @see cnh.crb.SeasonalOffset#getOffset(java.util.Calendar)
		 */
		@Override
		public float getMonthlyScaleFactor() {
			return offset;
		}
		
		public float getMonthlyScaleFactorFor(Calendar date) {
			int yearIndex = getYearIndex(date);
			Year year = years.get(yearIndex);
			// jan is 0
			int month = date.get(Calendar.MONTH);
			return year.months[month];
		}
		
		public Year getAverageYear(){
			Year year = new Year();
			for(Year y: years){
				for(int i = 0; i < 12; i++){
					year.mean      += y.mean;
					year.months[i] += y.months[i];
				}
			}
			year.mean = year.mean / years.size();
			for(int i = 0; i < 12; i++) year.months[i] = year.months[i]/years.size();
			return year;
		}
	}
	
	
	
	// Instance methods
	
	// Private
	private void parseMonth(StartElement element, Year year) {
		int month = Integer.parseInt(element.getAttributeByName(DATE).getValue());
		float frac = Float.parseFloat(element.getAttributeByName(FRAC_OF_MEAN).getValue());
		year.months[month - 1] = frac;
	}
	
	private Year parseYear(StartElement element, SO so) {
		Year year = new Year();
		so.years.add(year);
		year.year = Integer.parseInt(element.getAttributeByName(DATE).getValue());
		year.mean = Float.parseFloat(element.getAttributeByName(MEAN).getValue());
		return year;
	}
	
	private static class Year {
		int year;
		float mean;
		float[] months = new float[12];
	}
	
	
	// Public
	public SeasonalWaterUse load(String file, int endingYear) throws IOException, XMLStreamException {
		return this.load(file, endingYear, false);
	}

	
	public SeasonalWaterUse load(String file, int endingYear, boolean useAverage) throws IOException, XMLStreamException {
		SO so = new SO(endingYear);
		
		XMLInputFactory factory = XMLInputFactory.newInstance();
		XMLEventReader reader = null;
		try {
			reader = factory.createXMLEventReader(new BufferedReader(new FileReader(file)));
			Year year = null;
			 while (reader.hasNext()) {
			      XMLEvent evt = reader.nextEvent();
			      if (evt.isStartElement()) {
			    	  StartElement elm = evt.asStartElement();
			          QName elmName = elm.getName();
			    	  if (elmName.equals(YEAR)) {
			    		  year = parseYear(elm, so); 
			    	  } else if (elmName.equals(MONTH)) {
			    		  parseMonth(elm, year);
			    	  }
			      } 
			 }
		} finally {
			if (reader != null) try {
				reader.close();
			} catch (Exception ex) {}
		}
		
		so.calcYear0();
		return useAverage ? new SingleYearSeasonalWaterUse(so.getAverageYear()) : so;
	}
	
		

}
