package cnh.crb;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;

public class TimeKeeper implements IAction {
	
	private static Logger LOGGER = Logger.getLogger(TimeKeeper.class);
	
	public static class Day {
		
		// 1-based day of month
		int date;
		@Override
		public String toString() {
			return "Day [date=" + date + ", month=" + month + ", year=" + year
					+ "]";
		}
		// 0-based month
		int month;
		// year
		int year;
		
		public Day(int date, int month, int year) {
			this.date = date;
			this.month = month;
			this.year = year;
		}
		
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + date;
			result = prime * result + month;
			result = prime * result + year;
			return result;
		}
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Day other = (Day) obj;
			if (date != other.date)
				return false;
			if (month != other.month)
				return false;
			if (year != other.year)
				return false;
			return true;
		}
	}
	
	public static class TimeKeeperAction {
		
		IAction action;
		int priority;

		public TimeKeeperAction(IAction action, int priority) {
			this.action = action;
			this.priority = priority;
		}

		public IAction getAction() {
			return action;
		}

		public int getPriority() {
			return priority;
		}

	}

	public Calendar getCal() {
		return cal;
	}

	private Calendar cal;
	private int defaultIncrement;
	private List<TimeKeeperAction> dailyActions = new ArrayList<>();
	private List<TimeKeeperAction> weeklyActions = new ArrayList<>();
	private List<TimeKeeperAction> monthlyActions = new ArrayList<>();
	private List<TimeKeeperAction> yearlyActions = new ArrayList<>();
	
	private Map<Day,List<TimeKeeperAction>> specificDateActions = new LinkedHashMap<>();
	
	// This is to avoid incrementing past the first day of the simulation.
	boolean doIncrement = false;

	public TimeKeeper(Calendar cal, int defaultIncrement) {
		this.cal = cal;
		this.defaultIncrement = defaultIncrement;
	}

	public TimeKeeper(Calendar cal) {
		this(cal, 1);
	}

	@Override
	public void execute() {
		// don't increment very first day
		if (doIncrement)
			cal.add(Calendar.DAY_OF_YEAR, defaultIncrement);
		LOGGER.debug("TimeKeeper updating: new date is " + cal.getTime());
		// Schedule specific date actions
		Day today = new Day(cal.get(Calendar.DATE),cal.get(Calendar.MONTH),cal.get(Calendar.YEAR));
		if (specificDateActions.containsKey(today)){
			List<TimeKeeperAction> todayActions = specificDateActions.get(today);
			if (todayActions!=null){
				scheduleTimeKeeperActions(todayActions);
			}
			// Remove today
			specificDateActions.remove(today);
		}
		
		// Daily actions always happen
		scheduleTimeKeeperActions(dailyActions);
		if (cal.get(Calendar.DAY_OF_WEEK) == 1) {
			scheduleTimeKeeperActions(weeklyActions);
		}
		if (cal.get(Calendar.DAY_OF_MONTH) == 1) {
			scheduleTimeKeeperActions(monthlyActions);
		}
		if (cal.get(Calendar.DAY_OF_YEAR) == 1) {
			scheduleTimeKeeperActions(yearlyActions);
		}
		doIncrement = true;
	}

	private void scheduleTimeKeeperActions(List<TimeKeeperAction> timeKeeperActions) {
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		double currentTime = schedule.getTickCount();
		for (TimeKeeperAction tka : timeKeeperActions) {
			schedule.schedule(
					ScheduleParameters.createOneTime(currentTime, tka.getPriority()), tka.getAction());
		}
	}
	
	// Specific Date Actions Management
	public void registerSpecificDateAction(Day day, TimeKeeperAction tka){
		if (!specificDateActions.containsKey(day)){
			specificDateActions.put(day, new ArrayList<TimeKeeperAction>());
		}
		specificDateActions.get(day).add(tka);
	}
	
	public void clearSpecificDateActions(){
		specificDateActions.clear();
	}
	
	public void clearSpecificDateActions(Day day){
		if (specificDateActions.containsKey(day)){
			List<TimeKeeperAction> dayActions = specificDateActions.get(day);
			if (dayActions!=null){
				dayActions.clear();
			}
		}
	}
	
	public void clearSpecificDateAction(Day day, TimeKeeperAction tka){
		if (specificDateActions.containsKey(day)){
			List<TimeKeeperAction> dayActions = specificDateActions.get(day);
			if (dayActions!=null){
				dayActions.remove(tka);
			}
		}
	}

	// Daily Action List Management
	public void registerDailyAction(TimeKeeperAction tka) {
		dailyActions.add(tka);
	}

	public void clearDailyActions() {
		dailyActions.clear();
	}

	public void removeDailyAction(TimeKeeperAction tka) {
		dailyActions.remove(tka);
	}

	// Weekly Action List Management
	public void registerWeeklyAction(TimeKeeperAction tka) {
		weeklyActions.add(tka);
	}

	public void clearWeeklyActions() {
		weeklyActions.clear();
	}

	public void removeWeeklyAction(TimeKeeperAction tka) {
		weeklyActions.remove(tka);
	}

	// Monthly Action List Management
	public void registerMonthlyAction(TimeKeeperAction tka) {
		monthlyActions.add(tka);
	}

	public void clearMonthlyActions() {
		monthlyActions.clear();
	}

	public void removeMonthlyAction(TimeKeeperAction tka) {
		monthlyActions.remove(tka);
	}

	// Yearly Action List Management
	public void registerYearlyAction(TimeKeeperAction tka) {
		yearlyActions.add(tka);
	}

	public void clearYearlyActions() {
		yearlyActions.clear();
	}

	public void removeYearlyAction(TimeKeeperAction tka) {
		yearlyActions.remove(tka);
	}

}