package cnh.crb.agent;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import cnh.crb.hydro.DemandRequest;
import cnh.crb.hydro.FlowHydrograph;
import cnh.crb.hydro.FlowManager;
import cnh.crb.hydro.FlowNode;
import cnh.crb.hydro.RecipientFlowNode;
import cnh.crb.hydro.SourceFlowNode;
import cnh.crb.hydro.SupplyRequest;
import cnh.crb.hydro.WaterSink;
import cnh.crb.hydro.FlowRate.FlowRateUnits;

public abstract class AbstractWaterManager implements SourceFlowNode, RecipientFlowNode {
  

  protected String                                id;
	
  protected List<SourceFlowNode>                  waterSources;
  protected List<RecipientFlowNode>               waterRecipients;
  protected Map<SourceFlowNode, DemandRequest>    expectedInflow;
  protected Map<RecipientFlowNode, SupplyRequest> outflow;
  protected boolean                               ready;
  protected WaterSink                             sink;  
  
  
  protected AbstractWaterManager(String id){
	this.id         = id;
    waterSources    = new ArrayList<SourceFlowNode>();
    waterRecipients = new ArrayList<RecipientFlowNode>();
    expectedInflow  = new HashMap<SourceFlowNode, DemandRequest>();
    outflow         = new HashMap<RecipientFlowNode, SupplyRequest>();
  }
  
  public String getId(){ return id; }
  
  public void addWaterSource(SourceFlowNode source){
    if(!waterSources.contains(source)){
      waterSources.add(source);
      source.addWaterRecipient(this);
    }
  }
  
  public void addWaterRecipient(RecipientFlowNode recipient){
    if(!waterRecipients.contains(recipient)){
      waterRecipients.add(recipient);
      recipient.addWaterSource(this);
    }
  }
  
  public void removeWaterRecipient(RecipientFlowNode recipient){
	  waterRecipients.remove(recipient);
  }
  
  public void registerDemandRequest(RecipientFlowNode downstreamNode, FlowHydrograph demand, FlowManager flowManager){
    // Register the demand. If the demand is different
    // from a previously registered demand (possibly because
    // no demand has been registered yet for this recipient)
    // then change to a non-ready state
    SupplyRequest extant = outflow.get(downstreamNode);
    if(extant == null){
      outflow.put(downstreamNode, new SupplyRequest(demand));
      ready = false;
      flowManager.markNonReady(this);
    }
    else if(extant.registerDemand(demand)){
      ready = false;
      flowManager.markNonReady(this);
    }
  }
  
  public void registerSupplyToBeReceived(SourceFlowNode upstreamNode, FlowHydrograph supplyToBeReceived, 
		  FlowHydrograph additionalSupplyAvailable, FlowManager flowManager){ 
    // Register an amount to be received
    DemandRequest extant = expectedInflow.get(upstreamNode);
    if(extant == null){
      expectedInflow.put(upstreamNode,  new DemandRequest(supplyToBeReceived, additionalSupplyAvailable));
      ready = false;
      flowManager.markNonReady(this);
    }
    else{
	  if(extant.registerSupply(supplyToBeReceived,  additionalSupplyAvailable)){
	    ready = false;
	    flowManager.markNonReady(this);
	  }
    }
  }
  
  /**
   * Register demand with the source nodes. 
   * 
   * @param flowManager
   */
  protected void notifySourcesOfDemand(FlowManager flowManager){
	for(SourceFlowNode src: waterSources){
//		System.out.println(getId() + " notifying source " + src.getId() + " of demand " + expectedInflow.get(src).hydrograph1.getTotalFlowVolume(VolumeUnits.M3) + " m^3");
		src.registerDemandRequest(this, expectedInflow.get(src).hydrograph1, flowManager);
	}
  }

  /**
   * Notify recipient nodes of supply
   * @param flowManager
   */
  protected void notifyRecipientsOfSupply(FlowManager flowManager){
    for(RecipientFlowNode recipient: outflow.keySet()) recipient.registerSupplyToBeReceived(this, outflow.get(recipient).hydrograph1, new FlowHydrograph(), flowManager);
  }

  
  public void clearFlow(){
    expectedInflow.clear();
    outflow.clear();
    ready = false;
  }
  
  public boolean isReadyForFlow(){
    return ready;
  }
  
  protected boolean allRecipientsReadyToFlow(){
    for(RecipientFlowNode recipient: waterRecipients) if(!((FlowNode)recipient).isReadyForFlow()) return false;
    return true;
  }
  
  public List<SourceFlowNode> getSources(){
	  return waterSources;
  }
  
  // Called for water that is not to be delivered to downstream nodes but rather sent to the sink
  public abstract void sink();
  
  public void setSink(WaterSink sink){
	  this.sink = sink;
  }
  
  public WaterSink getSink(){ return sink; }
}
