package cnh.crb.agent;

import java.util.Date;

import repast.simphony.engine.environment.RunEnvironment;
import cnh.crb.hydro.DemandRequest;
import cnh.crb.hydro.FlowHydrograph;
import cnh.crb.hydro.FlowManager;
import cnh.crb.hydro.RecipientFlowNode;
import cnh.crb.hydro.SourceFlowNode;
import cnh.crb.hydro.Storage;
import cnh.crb.hydro.SupplyRequest;
import cnh.crb.hydro.WaterRequest;
import cnh.crb.hydro.FlowRate.FlowRateUnits;

/**
 * Manages water and serves water to consumers.
 * 
 * This implementation has a reservoir, a collection
 * of water sources, and a collection of water users.
 * 
 * @author murphy
 *
 */
public class BasicWaterManager extends AbstractWaterManager {
	
	protected Storage storage;

	private boolean useFullWaterFlow;
	
	protected double deltaStorage;
	
	protected class NotificationStatus{
		boolean recipientsNotified = false;
		boolean sourcesNotified    = false;
		
		public NotificationStatus(){ }
		
		public NotificationStatus(boolean recipients, boolean sources){
			recipientsNotified = recipients;
			sourcesNotified    = sources;
		}
		
		public void add(NotificationStatus other){
			recipientsNotified = recipientsNotified || other.recipientsNotified;
			sourcesNotified    = sourcesNotified    || other.sourcesNotified;
		}
	}
	
	public BasicWaterManager(Storage storage, String managerID){
		super(managerID);
		this.storage   = storage;
		useFullWaterFlow = RunEnvironment.getInstance().getParameters().getBoolean("useFullWaterFlow");
	}
  
  public FlowHydrograph getDemand(Date start, Date end){
	  FlowHydrograph H = new FlowHydrograph();
	  if(allRecipientsReadyToFlow()){
		  for(SupplyRequest supplyReq : outflow.values()){
//			  H.add(supplyReq.hydrograph2); // TODO Sometimes it should be 1, sometimes 2
			  H.add(supplyReq.hydrograph1);
		  }
	  }
	  else{
          for(RecipientFlowNode recipient: waterRecipients){
        	  H.add(recipient.getDemand(start, end));
          }
	  }
      return H;
  }

  public boolean prepareFlow(Date start, Date end, FlowManager flowManager){
	// If you're not using full water flow, stop here
	if(!useFullWaterFlow) return true;
    // If you're in a 'ready' state, just acknowledge
    if(ready) return true;
    // If you don't know all of your demand, do nothing
    if(!allRecipientsReadyToFlow()) return false;
    // If all of your recipients are ready but you are not
    // You need to calculate how you will supply your
    // recipients and, possibly, what you will demand from
    // your supplier.
    
    // Possibly you are in this state because you have just received the
    // demand from your recipients but you have not 
    // yet registered your own demand with your source. 
    // Alternatively, it may be because you were in a 'ready'
    // state, but then your expected inflow was 
    // changed by the upstream source(s) or your
    // outflow was changed by the downstream recipients
    NotificationStatus status = calculateService(start, end, flowManager);
    if(!status.sourcesNotified)    notifySourcesOfDemand(flowManager); 
    if(!status.recipientsNotified) notifyRecipientsOfSupply(flowManager);
    ready = true;
    return ready;
  }
  
  protected NotificationStatus calculateService(Date start, Date end, FlowManager flowManager){
	  NotificationStatus ret = new NotificationStatus();
	  // If you have no information about incoming supply
	  if(expectedInflow.size() == 0){
		  ret.add(calculateAmountsToSupplyToRecipients(start, end, flowManager));
		  ret.add(calculateAmountsToDemandFromSources(start, end, flowManager));
	  }
	  else{
		  // In this case there is information in the inflow record,
		  // but you are not in a ready state.
		  // This is not yet implemented
	  }
	  return ret;
  }
  
  protected NotificationStatus calculateAmountsToSupplyToRecipients(Date start, Date end, FlowManager flowManager){
	  for(RecipientFlowNode recipient: outflow.keySet()){
          SupplyRequest out = outflow.get(recipient);
          out.allocateSupply(out.hydrograph1.getSum(out.hydrograph2));
      }
	  return new NotificationStatus(false, false);
  }
  
  /**
   * Given a demand (from all recipients) but no constraints from
   * upstream sources, calculate what the demand from any upstream
   * sources should be.
   * 
   * For efficiency, this may not only calculate the values but
   * also notify either the sources or the recipients of the results
   * of the calculation (e.g. to avoid looping through recipients
   * twice); the return flag should indicate this.
   * 
   * The implementation in this parent class simply sums the demand from
   * all recipients and assumes one source, which it does not notify.
   * 
   * @param start
   * @param end
   * @param flowManager
   * @return
   */
  protected NotificationStatus calculateAmountsToDemandFromSources(Date start, Date end, FlowManager flowManager){
	  // Assume only one source (for now)
	  SourceFlowNode src = waterSources.get(0);
	  // The amount of the demand will be the sum of all requested amounts
	  FlowHydrograph H = getDemand(start, end);
	  expectedInflow.put(src, new DemandRequest(H, null)); // Assume it will be met
	  return new NotificationStatus(false, false);
  }
  
  
    
  
  /**
   * Assumes that the demand from recipients has been changed.
   * At this point need to decide if:
   *   Demand should be re-requested from sources, OR
   *   Supply can be re-allocated to recipients, OR
   *   Or BOTH, OR
   *   Keep the new values- It's OK to run a deficit
   *        or surplus- the storage will be fine
   *        
   * @param start
   * @param end
   */
  protected NotificationStatus respondToAlteredSupplyOrDemand(Date start, Date end){
    return new NotificationStatus();
  }
  
  
  public void flow(Date end){
	if(!useFullWaterFlow) return;
	for(WaterRequest src: expectedInflow.values())	storage.addInflowHydrograph(src.hydrograph1);  // Add all inflow
    modifyStorageFlow();
    deltaStorage = storage.getCurrentVolume();
    for(WaterRequest out: outflow.values())	storage.addOutflowHydrograph(out.hydrograph1); // Add all outflow
    storage.flowToDate(end);                                                                      // Update reservoir state
    deltaStorage = storage.getCurrentVolume() - deltaStorage;
    sink();
    clearFlow();
  }
  
  /**
   * Some child classes may need to modify the storage flow at this point  
   */
  protected void modifyStorageFlow(){ }
 
  public String toString(){ return id; }

  // Does nothing
  public void sink(){  }
  
  
}
