package cnh.crb.agent;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.StringWriter;
import java.text.DecimalFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

import org.apache.log4j.Logger;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;

import repast.simphony.context.Context;
import cnh.crb.WaterSourceLoader;
import cnh.crb.hydro.CylindricalStorage;
import cnh.crb.hydro.DemandRequest;
import cnh.crb.hydro.FlowHydrograph;
import cnh.crb.hydro.FlowManager;
import cnh.crb.hydro.FlowRate;
import cnh.crb.hydro.FlowRateGPD;
import cnh.crb.hydro.SourceFlowNode;
import cnh.crb.hydro.Storage;
import cnh.crb.hydro.UnitsAndConverters;
import cnh.crb.hydro.UnitsAndConverters.VolumeUnits;
import cnh.crb.hydro.WaterSink;
import cnh.crb.hydro.WaterSink.WATER_SINK_METHOD;

public class DataDrivenDemandWaterManager extends BasicWaterManager{

  private static final int MUNICIPAL    = 0;
  private static final int AGRICULTURAL = 1;
  private static final int RECHARGE     = 2;
  
  public static boolean includeMunicipalDemand    = true;
  public static boolean includeAgriculturalDemand = false;
  public static boolean includeRechargeDemand     = true;
	
  private static Logger LOGGER = Logger.getLogger(DataDrivenDemandWaterManager.class);
	
  public FlowRate.FlowRateUnits units;

  public    static Set<DataDrivenDemandWaterManager> instances;
  
  protected static Map<String, Map<Integer, double[][]>> data;
  protected int      currentYear   = -1;
  protected int      currentMonth  = -1;
  protected FlowRate rate;
  protected double   returnPercent;
  
  public static void reinitialize(){
	  instances = new HashSet<DataDrivenDemandWaterManager>();
	  data      = null;
  }
  
  public static Set<String> idsToSkip = new HashSet<String>();
  
  public static void initialize(String filename, Context<Object> context){
	  data      = new TreeMap<String, Map<Integer, double[][]>>();
	  instances = new HashSet<DataDrivenDemandWaterManager>();
	  Map<String, WaterSink> sinks = new HashMap<String, WaterSink>();
	  for(Object sink: context.getObjects(WaterSink.class)){
		  WaterSink s = (WaterSink)sink;
		  sinks.put(s.getId(), s);
	  }
	  Set<String> ids = new HashSet<String>();
	  try{
		  // Read the data from a file
		  HSSFWorkbook workbook = new HSSFWorkbook(new FileInputStream(filename));
		  int sheetIndex = workbook.getSheetIndex("FullData");
		  HSSFSheet sheet = workbook.getSheetAt(sheetIndex);
		  int rows = sheet.getPhysicalNumberOfRows();
		  for(int rowIndex = 1; rowIndex < rows; rowIndex++){
			  HSSFRow row = sheet.getRow(rowIndex);
			  if(row.getPhysicalNumberOfCells() > 0){
				  String baseName = row.getCell(8).getStringCellValue();                       // Column 'I', purchaser
				  int year = (int)row.getCell(0).getNumericCellValue();
				  if(!idsToSkip.contains(baseName)){
					  String sinkName = row.getCell(9).getStringCellValue().replace("&", "and");   // Column 'J', delivery location
					  String entityID = baseName + "_" + sinkName;
					  
					  
					  String waterUse = row.getCell(6).getStringCellValue();
					  int waterUseIndex                                 = MUNICIPAL;    // Default is M & I
					  if(waterUse.startsWith("AG"))       waterUseIndex = AGRICULTURAL;
					  if(waterUse.startsWith("RECHARGE")) waterUseIndex = RECHARGE;
	
					  LOGGER.debug("Entering data for entity " + entityID + " in year: " + year + " ");
					  int firstColumn = 16;
					  for(int col = firstColumn; col < (firstColumn + 12); col++){
					    int month = col - firstColumn;
					    // Must convert from acre-feet per month (!) to GPD
					    double afPerMonth = row.getCell(col).getNumericCellValue();
					    double gallonsPerDay = 
							  UnitsAndConverters.AFToGallons(afPerMonth) / 
							  (new GregorianCalendar(year, month, 1)).getActualMaximum(Calendar.DAY_OF_MONTH);
					    LOGGER.debug("   Value: acre-feet per month " + afPerMonth + " in month " + month);
					    addDatum(entityID, year, month, waterUseIndex, gallonsPerDay);
				      }
					  if(!ids.contains(entityID)){
						  DataDrivenDemandWaterManager newMgr = new DataDrivenDemandWaterManager(entityID); // Adds it to instances automatically
						  sinkName = WaterSourceLoader.nameMap.get(sinkName); // Look up 
						  WaterSink sink = sinks.get(sinkName);
						  if(sink != null){
							  LOGGER.debug("Setting sink for " + newMgr.getId() + " to " + sink.getId());
							  newMgr.setSink(sink);
						  }
						  context.add(newMgr);
						  ids.add(entityID);
					  }
				  }
				  else{
					  LOGGER.info("Skipping " + baseName + " for year " + year);
				  }
			  }
		  }
	  }
	  catch(IOException E){
		  LOGGER.info("IO Exception in DataDrivenWaterManager.initialize");
	  }
//	  report(2002, 0);
//	  report(2002, 1);
//	  report(2002, 2);

  }
  
  public static void addDatum(String ID, Integer year, int month, int use, double value){
	  LOGGER.debug("    ADDING DATUM: " + ID + " " + year + " " + month + " " + value);
	  Map<Integer, double[][]> yearData = data.get(ID);
	  if(yearData == null){
		  yearData = new TreeMap<Integer, double[][]>();
		  data.put(ID, yearData);		  
	  }
	  double[][] rawData = yearData.get(year);
	  if(rawData == null){
		  rawData = new double[3][12];
		  for(int i = 0; i < 3; i++) for(int j = 0; j < 12; j++) rawData[i][j] = 0;
		  yearData.put(year, rawData);
	  }
	  rawData[use][month] += value;
  }
  
  public static void setAll(Integer year, int month){
	  if(instances == null) return;
	  for(DataDrivenDemandWaterManager mgr: instances) mgr.getDemand(year, month);
  }
  
  public FlowRate getDemand(Integer year, int month){
	  // Memoize for performance
	  if(month != currentMonth || year != currentYear){
		  currentYear  = year;
		  currentMonth = month;
		  Map<Integer, double[][]> allData = data.get(getId());
		  if(allData == null || allData.keySet().size() == 0){
			  rate = new FlowRateGPD(0);
			  returnPercent = 0;
		  }
		  else{
			  double[][] yearData = allData.get(year);
			  if(yearData == null){
				  rate = new FlowRateGPD(0);
				  returnPercent = 0;
			  }
			  else{
				  double municipal    = includeMunicipalDemand    ? yearData[MUNICIPAL][month]    : 0;
				  double agricultural = includeAgriculturalDemand ? yearData[AGRICULTURAL][month] : 0;
				  double recharge     = includeRechargeDemand     ? yearData[RECHARGE][month]     : 0;
				  double totalRate = municipal + agricultural + recharge;
				  rate = new FlowRateGPD(totalRate);
				  returnPercent = totalRate > 0 ? (municipal + agricultural) / totalRate : 0;
			  }
		  }
	  }
	  if(rate.getRateGPD() > 0) LOGGER.debug(this.getId() + " date: " + month + "/" + year + " is flowing: " + rate.getString(FlowRate.FlowRateUnits.M3S) + " m3/s and returning " + (returnPercent * 100) + "%");
	  return rate;
  }
  
  private static Storage getDummyStorage(){
	  return new CylindricalStorage(new Date(), 10, 10000, "Dummy");
  }
  
  
  public DataDrivenDemandWaterManager(Storage storage, String managerID) {
	super(storage, managerID);
	if(instances == null)  instances = new HashSet<DataDrivenDemandWaterManager>();
	instances.add(this);
  }
  
  
  public DataDrivenDemandWaterManager(String managerID){
	  super(getDummyStorage(), managerID);
	  if(instances == null)  instances = new HashSet<DataDrivenDemandWaterManager>();
	  instances.add(this);
  }
  
	
  protected NotificationStatus calculateAmountsToDemandFromSources(Date start, Date end, FlowManager flowManager){
	  // Assume only one source (for now)
	  SourceFlowNode src = waterSources.get(0);
	  // The amount of the demand will be the sum of all requested amounts
	  FlowHydrograph H = getDemand(start, end);
	  expectedInflow.put(src, new DemandRequest(H, null)); // Assume it will be met
	  return new NotificationStatus(false, false);
  }
  
  public FlowHydrograph getDemand(Date start, Date end){
	  super.getDemand(start, end); // Calculate as per usual, but this will be EMPTY:
	  FlowHydrograph H = new FlowHydrograph(start, end, rate.getRate(), rate.getNativeUnits()); 
	  LOGGER.debug("In DataDrivenDemandWaterManager " + getId() + " getDemand:");
	  LOGGER.debug(H.report(rate.getNativeUnits()));
	  return H;
  }
  
  public void flow(Date end){
	  super.flow(end);
	  if(storage.getCurrentVolume() > 0) LOGGER.debug(getId() + " at end of flow My Reservoir Volume is: " + this.storage.getCurrentVolume());
  }
	
  public void sink(){
//	  System.out.println(getId() + " running sink (null?: " + (sink == null) + "; vol=0?: " + (storage.getCurrentVolume() == 0) + ")" );
	  if(sink != null){
		  if(storage.getCurrentVolume() > 0){
//			  System.out.println(getId() + " telling the sink " + sink.getId() + " that I'm eliminating " + returnPercent + " * " + storage.getCurrentVolume());
			  sink.eliminateWater(storage.getCurrentVolume() * returnPercent, VolumeUnits.M3, WATER_SINK_METHOD.RETURN_TO_SURFACE);
			  storage.setVolumeStored(0);
		  }
	  }
  }

  public static void report(int year){
	  String[] types = {"M&I", "Agr", "Rch" };
	  DecimalFormat decimalFormat = new DecimalFormat("#0.000");
	  Calendar c = Calendar.getInstance();
	  // Loop through and give a report of every entity by year... big, but valuable!
	  for(Entry<String, Map<Integer, double[][]>> entry: data.entrySet()){
		  StringWriter entityReport = new StringWriter();
		  entityReport.write("Data Driven Mgr ID: " + entry.getKey() + System.getProperty("line.separator"));
		  Map<Integer, double[][]> dataVals = entry.getValue();
		  boolean hasValues = false;
		  for(Entry<Integer, double[][]> yearVals : dataVals.entrySet()){
			  int loopYear = yearVals.getKey().intValue();
			  if((year == -1) || (loopYear == year)){
				  c.set(Calendar.YEAR, loopYear);
				  StringWriter yearReport = new StringWriter();
				  yearReport.write("   " + yearVals.getKey() + System.getProperty("line.separator"));
				  for(int i = 0; i < 3; i++){
					  double[][] vals = yearVals.getValue();
					  double total = 0;
					  StringWriter lineReport = new StringWriter();
					  lineReport.write("      " + types[i] + ": ");
					  for(int j = 0; j < 12; j++){
						  c.set(Calendar.MONTH, j);
						  double val = UnitsAndConverters.GallonsToAF(vals[i][j]) * c.getActualMaximum(Calendar.DAY_OF_MONTH);  // Convert back to AF/Month
						  lineReport.write(decimalFormat.format(val) + " ");
						  total += val;
					  }
					  if(total > 0){
						  yearReport.write(lineReport.toString() + System.getProperty("line.separator"));
						  hasValues = true;  
					  }
				  }
				  //if(hasValues) entityReport.write(yearReport.toString());
			  }
		  }
		  //if(hasValues) System.out.print(entityReport.toString());
	  }
  }

  public static void report(int year, int type){
	  String[] types = {"M&I", "Agr", "Rch" };
	  DecimalFormat decimalFormat = new DecimalFormat("#0.0");
	  Calendar c = Calendar.getInstance();
	  c.set(Calendar.YEAR, year);
	  StringWriter report = new StringWriter();
	  report.write(year + " " + types[type] + ":" + System.getProperty("line.separator"));
	  for(Entry<String, Map<Integer, double[][]>> entry: data.entrySet()){
		  String ID = entry.getKey();
		  Map<Integer, double[][]> dataVals = entry.getValue();
		  if(dataVals.containsKey(year)){
			  double[][] data = dataVals.get(year);
			  double total = 0;
			  StringWriter lineReport = new StringWriter();
			  lineReport.write("  " + ID.replaceAll(" ", "_") + ": ");
			  for(int j = 0; j < 12; j++){
				  c.set(Calendar.MONTH, j);
				  double val = UnitsAndConverters.GallonsToAF(data[type][j]) * c.getActualMaximum(Calendar.DAY_OF_MONTH);  // Convert back to AF/Month
				  lineReport.write(decimalFormat.format(val) + " ");
				  total += val;
			  }
			  if(total > 0) report.write(lineReport.toString() + System.getProperty("line.separator"));
		  }
	  }
	  System.out.print(report.toString());
  }
  
}
