package cnh.crb.analysis;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import cnh.crb.analysis.datasets.ColumnFilterResampler;
import cnh.crb.analysis.datasets.DataSetSource;
import cnh.crb.analysis.datasets.MonthlyAverageResampler;
import cnh.crb.analysis.datasets.RollingAverageResampler;
import cnh.crb.analysis.datasets.SingleDataSet;
import cnh.crb.analysis.parametersets.BatchParametersSet;


/**
 * Given a data set of simulation runs (which may be
 * from a single batch or multiple batches, and may include
 * single runs or merged runs (across random number seeds, e.g))
 * compares the runs to the Tucson data using some measure of
 * fit, then ranks the runs according to their closeness on
 * this measure and reports the water use for the top
 * N according to this ranking.
 * @author murphy
 *
 */
public class RankRunsByFitToTucsonData {
	
	public static void main(String[] args){
		// Point to the data to be compared
		DataSetSource dsSource = 
				new DataSetSource(
						"",                                           // Bulk data does not exist for merged runs
						"./out/analysis/MERGED/WATER_USE/",           // Library of run data           
						"RUN",                                        // Prefix on individual runs
						"./out/analysis/MERGED/merged_params.txt");   // Parameters file
		
		
		// First load the Tucson data
		Calendar start = Calendar.getInstance();
		start.set(1993, 0, 1, 0, 0, 0);
		start.set(Calendar.MILLISECOND, 0);
		SingleDataSet tucson = new SingleDataSet("./out/TucsonMonthlyAverageAsIfSimulated.csv", start);
		double[] scalar = {748.052};
		tucson.addCalculatedColumn("TucsonWaterUsed", SingleDataSet.Function.MULTIPLY_BY, scalar, "waterUsed");
		MonthlyAverageResampler maResampler = new MonthlyAverageResampler(true);
		RollingAverageResampler raResampler = new RollingAverageResampler(12);
		ColumnFilterResampler remove = new ColumnFilterResampler("tick","TucsonWaterUsed");
		tucson = tucson.resample(maResampler, raResampler, remove);
		
		// Make a copy with which to work
		SingleDataSet analysis = new SingleDataSet(tucson, false);
		
		// Determine parameters
		BatchParametersSet batchParams = new BatchParametersSet("./out/analysis/MERGED/merged_params.txt");
		
		Map<String, Set<String>> valuesVaried = batchParams.getDistinctValues("run");
		System.out.println("Values varied:");
		int combinations = 1;
		for(String param: valuesVaried.keySet()){
			System.out.print(param + ": ");
			combinations *= valuesVaried.get(param).size();
			for(String val: valuesVaried.get(param)) System.out.print(val + " ");
			System.out.println();
		}
		System.out.println("Total possible combinations: " + combinations);
		System.out.println("Total runs: " + batchParams.getCountOfEntries());
		
		List<Integer> runsToExamine = new ArrayList<Integer>();
		runsToExamine.addAll(batchParams.getListOfRuns());
		// Short version for testing ONLY
//		while(runsToExamine.size() > 30) runsToExamine.remove(0); // TEST
		
		System.out.print("RUN,");
		String[] vals = {"priceWeight", "messagingSubsampling", "waterUseBottom", "messagingEffectiveness", "messagingBeginYear", "socialFactorWeight"};
		for(String val: vals) System.out.print(val + ",");
		System.out.println("DIFF");
		TreeMap<Double, Integer> topNRuns = new TreeMap<Double, Integer>();
		double Nthlowest = Double.MAX_VALUE;
		double Nthhighest = Double.MIN_VALUE;
		for(Integer run: runsToExamine){
			SingleDataSet ds = dsSource.getDataSet(run);
			ds.setStartDate(start);
			ds = ds.resample(maResampler, raResampler);
			SingleDataSet temp = new SingleDataSet(analysis, false);
			
			if(!temp.merge(ds, "RUN_" + run + "_")) System.out.println("MERGE FAIL FOR RUN " + run);
			double compare = temp.getAggregate(SingleDataSet.Function.SQUARE_DIFF, "TucsonWaterUsed", "RUN_" + run + "_" + "waterUsed");
			boolean trapLowest = true;
			int topNToTrap = 50;
			if(trapLowest){
				if(compare < Nthlowest){
					topNRuns.put(compare, run);
					if(topNRuns.size() > topNToTrap){
					    double highest = 0;
					    Nthlowest = 0;
					    for(Double d: topNRuns.keySet()) highest = Math.max(highest, d);
					    Integer removed = topNRuns.remove(highest);
					    for(Double d: topNRuns.keySet()) Nthlowest = Math.max(Nthlowest, d);
//					    System.out.println("Removed run " + removed + " to insert " + run + " with value " + compare + " New nth Lowest = " + Nthlowest);
					}
				}
			}
			else{
				
				if(compare > Nthhighest){
					topNRuns.put(compare, run);
					if(topNRuns.size() > topNToTrap){
					    double lowest = Double.MAX_VALUE;
					    for(Double d: topNRuns.keySet()) lowest = Math.min(lowest, d);
					    Integer removed = topNRuns.remove(lowest);
					    Nthhighest = Double.MAX_VALUE;
					    for(Double d: topNRuns.keySet()) Nthhighest = Math.min(Nthhighest, d);
//					    System.out.println("Removed run " + removed + " to insert " + run + " with value " + diffSq + " New nth Highest = " + Nthhighest);
					}
				}
			}
			
//			temp.report(true);
			System.out.print(run + ",");
			for(String val: vals) System.out.print(batchParams.getParameterValue(run, val) + ",");
			System.out.println(compare);
			
		}
		

		// Report to NRuns
		System.out.print("RUN,");
		for(String val: vals) System.out.print(val + ",");
		System.out.println("DIFF");
		SingleDataSet temp = new SingleDataSet(analysis, false);
		for(Double runKey: topNRuns.keySet()){
			int run = topNRuns.get(runKey);
			SingleDataSet ds = dsSource.getDataSet(run);
			ds.setStartDate(start);
			ColumnFilterResampler removeComponentRuns = new ColumnFilterResampler("waterUsed");
			ds = ds.resample(maResampler, raResampler, removeComponentRuns);
			if(!temp.merge(ds, "RUN_" + run + "_")) System.out.println("MERGE FAIL FOR RUN " + run);
			System.out.print(run + ",");
			for(String val: vals) System.out.print(batchParams.getParameterValue(run, val) + ",");
			System.out.println(runKey);
		}
		temp.report(true);
		
		
		
	}
}
