package cnh.crb.analysis;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.TreeMap;

import cnh.crb.analysis.datasets.ColumnFilterResampler;
import cnh.crb.analysis.datasets.DataSetSource;
import cnh.crb.analysis.datasets.DataTrimmerResampler;
import cnh.crb.analysis.datasets.SingleDataSet;
import cnh.crb.analysis.datasets.TickShifterResampler;
import cnh.crb.analysis.parametersets.BatchParametersSet;



/**
 * Given a data set of simulation runs (which may be
 * from a single batch or multiple batches, and may include
 * single runs or merged runs (across random number seeds, e.g))
 * compares the runs to the Tucson data using some measure of
 * fit, then ranks the runs according to their closeness on
 * this measure and reports the water use for the top
 * N according to this ranking.
 * @author murphy
 *
 */
public class RankRunsByFitToTucsonDataMicro {
	
	public static final int DO_SERVICES = 0;
	public static final int DO_USAGE    = 0;
	
	public static final String versions[][] = {{ "SERVICE_COUNTS",  "SERVICE_COUNT" },
                                               { "USAGE_HISTOGRAM", "USAGE_HIST"}};

	public static final String[] vals = {"priceWeight", "messagingSubsampling", "waterUseBottom", "messagingEffectiveness", "messagingBeginYear", "socialFactorWeight"};

	public static final int VERSION = DO_USAGE;
	
	public static void main(String[] args){
		System.out.println("Running Ranking Runs");
		
		// First load the Tucson data
		Calendar start = Calendar.getInstance();
		start.set(1993, 0, 1, 0, 0, 0);
		start.set(Calendar.MILLISECOND, 0);
		SingleDataSet tucson = 
			VERSION == DO_SERVICES ? new SingleDataSet("./out/TucsonBlockServiceCountsAsIfSimulated.csv", start) :
				                     new SingleDataSet("./out/TucsonBlockUsageAsIfSimulated.csv", start);
		
//		System.out.println("INITIAL TUCSON: ");
//		tucson.report(true);
//		System.out.println("<DONE> INITIAL TUCSON: ");
		
		double[] scalar = {748.052};
		tucson.addCalculatedColumn("Block1_WaterUsed", SingleDataSet.Function.MULTIPLY_BY, scalar, "Block1_Water_Use");
		tucson.addCalculatedColumn("Block2_WaterUsed", SingleDataSet.Function.MULTIPLY_BY, scalar, "Block2_Water_Use");
		tucson.addCalculatedColumn("Block3_WaterUsed", SingleDataSet.Function.MULTIPLY_BY, scalar, "Block3_Water_Use");
		tucson.addCalculatedColumn("Block4_WaterUsed", SingleDataSet.Function.MULTIPLY_BY, scalar, "Block4_Water_Use");
		tucson.addCalculatedColumn("Block5_WaterUsed", SingleDataSet.Function.MULTIPLY_BY, scalar, "Block5_Water_Use");
		
//		System.out.println("INITIAL TUCSON CONVERTED: ");
//		tucson.report(true);
//		System.out.println("<DONE> INITIAL TUCSON CONVERTED: ");
		
		ColumnFilterResampler remove = new ColumnFilterResampler("tick","Block1_WaterUsed","Block2_WaterUsed","Block3_WaterUsed","Block4_WaterUsed","Block5_WaterUsed");

		tucson = tucson.resample(remove);

//		System.out.println("INITIAL TUCSON RESAMPLED: ");
//		tucson.report(true);
//		System.out.println("<DONE> INITIAL TUCSON RESAMPLED: ");

		
		// Now convert to percentages of the total
		tucson.addCalculatedColumn("Block1_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block1_WaterUsed","Block2_WaterUsed","Block3_WaterUsed","Block4_WaterUsed","Block5_WaterUsed" );
		tucson.addCalculatedColumn("Block2_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block2_WaterUsed","Block1_WaterUsed","Block3_WaterUsed","Block4_WaterUsed","Block5_WaterUsed" );
		tucson.addCalculatedColumn("Block3_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block3_WaterUsed","Block1_WaterUsed","Block2_WaterUsed","Block4_WaterUsed","Block5_WaterUsed" );
		tucson.addCalculatedColumn("Block4_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block4_WaterUsed","Block1_WaterUsed","Block2_WaterUsed","Block3_WaterUsed","Block5_WaterUsed" );
		tucson.addCalculatedColumn("Block5_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block5_WaterUsed","Block1_WaterUsed","Block2_WaterUsed","Block3_WaterUsed","Block4_WaterUsed" );
//		tucson.addCalculatedColumn("TOTALPCT", SingleDataSet.Function.SUM, "Block1_Pct", "Block2_Pct", "Block3_Pct", "Block4_Pct", "Block5_Pct" ); // TEST- should sum to 1
		
//		System.out.println("INITIAL TUCSON RESAMPLED TO PCT: ");
//		tucson.report(true);
//		System.out.println("<DONE> INITIAL TUCSON RESAMPLED TO PCT: ");

		
		ColumnFilterResampler removeAllButPct = new ColumnFilterResampler("tick","Block1_Pct","Block2_Pct","Block3_Pct","Block4_Pct","Block5_Pct");
		tucson = tucson.resample(removeAllButPct);


//		System.out.println("FINAL TUCSON RESAMPLED TO PCT: ");
//		tucson.report(true);
//		System.out.println("<DONE> FINAL TUCSON RESAMPLED TO PCT: ");

		
		// Determine parameters
		BatchParametersSet batchParams = new BatchParametersSet("./out/analysis/MERGED/merged_params.txt");
		
		// Set up data source
		DataSetSource dsSource = 
				new DataSetSource(
						"./out/analysis/MERGED/merged_params.txt", 
						"./out/analysis/MERGED/" + versions[VERSION][0] + "/",
						versions[VERSION][1], 
						"./out/analysis/MERGED/merged_params.txt");

		// Determine runs to test
		List<Integer> runsToExamine = new ArrayList<Integer>();
		runsToExamine.addAll(batchParams.getListOfRuns());
		
		// Resamplers:
		// Trim off data from before 1996
		Calendar c = Calendar.getInstance();
		c.set(1995, 11, 31, 0, 0);
		DataTrimmerResampler dtr = new DataTrimmerResampler(c.getTime(), null);
		
		
		// The output of the micro-data is not a perfect match to the output of the
		// 'monthlyAverager' resampler. Given data for Jan 1, Jan 2, ... Jan 31,
		// the monthlyAverager will calculate the daily average for the days in the
		// month and output that as a value for Jan 31. The micro data ouputs data 
		// on the first of the month, e.g. Feb 1.
		// To make these match, we'll shift the micro-data back one tick
		TickShifterResampler tsr = new TickShifterResampler(-1);
		
		// Remove raw data, leaving only aggregated values
		ColumnFilterResampler removeFromMerged = new ColumnFilterResampler("tick","Block1","Block2","Block3","Block4","Block5");
		
		// Remove raw data, leaving only aggregated values
		ColumnFilterResampler removeRaw = new ColumnFilterResampler("tick","M_Block1_Pct","M_Block2_Pct","M_Block3_Pct","M_Block4_Pct","M_Block5_Pct");
				
				
		
		// Start output:
		// Print headers
		System.out.print("RUN,");
		for(String val: vals) System.out.print(val + ",");
		System.out.println("DIFF");
		
		// Loop through runs, collecting diff values
		TreeMap<Double, Integer> topNRuns = new TreeMap<Double, Integer>();
		double Nthlowest = Double.MAX_VALUE;
		double Nthhighest = Double.MIN_VALUE;
		for(Integer run: runsToExamine){
			// Get the data set
			SingleDataSet ds = dsSource.getDataSet(run);
			ds.setStartDate(start); // Set the start date
			
			// Clean up absurd column names
			ds.renameColumn("[0.0 - 11.0)",      "Block1");
			ds.renameColumn("[11.0 - 16.0)",     "Block2");
			ds.renameColumn("[16.0 - 31.0)",     "Block3");
			ds.renameColumn("[31.0 - 45.0)",     "Block4");
			ds.renameColumn("[45.0 - Infinity]", "Block5");

//			System.out.println("DS PRE-RESAMPLED");
//			ds.report(true);
//			System.out.println("<DONE> DS PRE-RESAMPLED");
			
			// Shift the ticks, trim the data, and remove raw data
			ds = ds.resample(tsr, dtr, removeFromMerged);

			// Calculate percentages
			ds.addCalculatedColumn("M_Block1_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block1","Block2","Block3","Block4","Block5" );
			ds.addCalculatedColumn("M_Block2_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block2","Block1","Block3","Block4","Block5" );
			ds.addCalculatedColumn("M_Block3_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block3","Block1","Block2","Block4","Block5" );
			ds.addCalculatedColumn("M_Block4_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block4","Block1","Block2","Block3","Block5" );
			ds.addCalculatedColumn("M_Block5_Pct", SingleDataSet.Function.PCT_OF_TOTAL, "Block5","Block1","Block2","Block3","Block4" );

			// Remove raw data
			ds = ds.resample(removeRaw);
			
//			System.out.println("DS RESAMPLED");
//			ds.report(true);
//			System.out.println("<DONE> DS RESAMPLED");
			
			// Working copy of the Tucson data set
			SingleDataSet temp = new SingleDataSet(tucson, false);
			if(!temp.merge(ds, "RUN_" + run + "_")) System.out.println("MERGE FAIL FOR RUN " + run);
			
//			System.out.println("TEMP POST MERGE");
//			temp.report(true);
//			System.out.println("<DONE> TEMP POST MERGE");
			
			double compare_b1 = temp.getAggregate(SingleDataSet.Function.SQUARE_DIFF, "Block1_Pct", "RUN_" + run + "_M_Block1_Pct");
			double compare_b2 = temp.getAggregate(SingleDataSet.Function.SQUARE_DIFF, "Block2_Pct", "RUN_" + run + "_M_Block2_Pct");
			double compare_b3 = temp.getAggregate(SingleDataSet.Function.SQUARE_DIFF, "Block3_Pct", "RUN_" + run + "_M_Block3_Pct");
			double compare_b4 = temp.getAggregate(SingleDataSet.Function.SQUARE_DIFF, "Block4_Pct", "RUN_" + run + "_M_Block4_Pct");
			double compare_b5 = temp.getAggregate(SingleDataSet.Function.SQUARE_DIFF, "Block5_Pct", "RUN_" + run + "_M_Block5_Pct");
			
			// Can use any or all of these
//			double compare = compare_b1 + compare_b2 + compare_b3 + compare_b4 + compare_b5;
			double compare = compare_b3;
			
			boolean trapLowest = true;
			int topNToTrap = 50;
			if(trapLowest){
				if(compare < Nthlowest){
					topNRuns.put(compare, run);
					if(topNRuns.size() > topNToTrap){
					    double highest = 0;
					    Nthlowest = 0;
					    for(Double d: topNRuns.keySet()) highest = Math.max(highest, d);
					    Integer removed = topNRuns.remove(highest);
					    for(Double d: topNRuns.keySet()) Nthlowest = Math.max(Nthlowest, d);
					}
				}
			}
			else{
				
				if(compare > Nthhighest){
					topNRuns.put(compare, run);
					if(topNRuns.size() > topNToTrap){
					    double lowest = Double.MAX_VALUE;
					    for(Double d: topNRuns.keySet()) lowest = Math.min(lowest, d);
					    Integer removed = topNRuns.remove(lowest);
					    Nthhighest = Double.MAX_VALUE;
					    for(Double d: topNRuns.keySet()) Nthhighest = Math.min(Nthhighest, d);
					}
				}
			}
			
			System.out.print(run + ",");
			for(String val: vals) System.out.print(batchParams.getParameterValue(run, val) + ",");
			System.out.println(compare);
			
		}
		
		// Report to NRuns
//		System.out.print("RUN,");
//		for(String val: vals) System.out.print(val + ",");
//		System.out.println("DIFF");
//		for(Double runKey: topNRuns.keySet()){
//			int run = topNRuns.get(runKey);
//			System.out.print(run + ",");
//			for(String val: vals) System.out.print(batchParams.getParameterValue(run, val) + ",");
//			System.out.println(runKey);
//		}
		
		
	}
}
