package cnh.crb.analysis;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import cnh.crb.analysis.datasets.ColumnFilterResampler;
import cnh.crb.analysis.datasets.DataSetSource;
import cnh.crb.analysis.datasets.MonthlyAverageResampler;
import cnh.crb.analysis.datasets.RollingAverageResampler;
import cnh.crb.analysis.datasets.SingleDataSet;
import cnh.crb.analysis.parametersets.BatchParametersSet;

/**
 * 
 * @author murphy
 *
 */
public class RetrieveRunsWithTucsonData {

	
	/**
	 * Creates an output that lists the Tucson water used values (as 12-month
	 * rolling averages of monthly averaged daily consumption) along with
	 * runs with specific price weight and social factor weight values
	 * @param args
	 */
	public static void main(String[] args){
		// Some useful preliminary instances
		double[] ccfToGallons = {748.052};  // Convert Hundred-Cubic-Foot to Gallons
		
		// A calendar corresponding to the start of the Tucson data:
		Calendar start = Calendar.getInstance();
		start.set(1993, 0, 1, 0, 0, 0);
		
		// Resamplers to modify the data sets
		MonthlyAverageResampler maResampler = new MonthlyAverageResampler(true);   // Creates the monthly averages
		RollingAverageResampler raResampler = new RollingAverageResampler(12);     // Creates the 12-month rolling average
		
		
		// First load the Tucson data
		SingleDataSet tucson = new SingleDataSet("./out/TucsonMonthlyAverageAsIfSimulated.csv", start);
		
		// Convert from CCF to Gallons
		tucson.addCalculatedColumn("TucsonWaterUsed", SingleDataSet.Function.MULTIPLY_BY, ccfToGallons, "waterUsed");  
		
		// Remove the Tucson Water Used column
		ColumnFilterResampler remove = new ColumnFilterResampler("tick","TucsonWaterUsed");
		
		// Resample the data to create Monthly Average and Rolling Average
		tucson = tucson.resample(maResampler, raResampler, remove);

		// Make a copy in which to work
		SingleDataSet analysis = new SingleDataSet(tucson, false);
		
		// Determine parameters
		DataSetSource dsSource = 
				new DataSetSource(
						"./out/combined_data_2014_08_14/ModelOutput.2014.Aug.14.11_39_04.txt",                    // Model output data 
						"./out/analysis/2014_Aug_14/",                                                            // Place where individual run data are stored
						"RUN",                                                                                    // Name prepended to each run
						"./out/combined_data_2014_08_14/ModelOutput.2014.Aug.14.11_39_04.batch_param_map.txt");   // Name of the parameters file for this data
		
		BatchParametersSet batchParams = dsSource.getParametersSet();
		
		// Determine the parameter values that were varied in these runs
		Map<String, Set<String>> valuesVaried = batchParams.getDistinctValues("run");
		System.out.println("Values varied:");
		int combinations = 1;
		for(String param: valuesVaried.keySet()){
			System.out.print(param + ": ");
			combinations *= valuesVaried.get(param).size();
			for(String val: valuesVaried.get(param)) System.out.print(val + " ");
			System.out.println();
		}
		System.out.println("Total possible combinations: " + combinations);
		System.out.println("Total runs: " + batchParams.getCountOfEntries());

		
		// Find the runs with specific values for price weight and social factor weight
		Map<String, List<String>> valuesMap = new TreeMap<String, List<String> >();
		List<String> priceWeightValuesToAdd = new ArrayList<String>();
		priceWeightValuesToAdd.add("0.2");
		valuesMap.put("priceWeight", priceWeightValuesToAdd);
		
		List<String> socialFactorWeightValuesToAdd = new ArrayList<String>();
		socialFactorWeightValuesToAdd.add("0.3");
		valuesMap.put("socialFactorWeight", socialFactorWeightValuesToAdd);
		
		List<Integer> runList = batchParams.getRunsWithMatchingValues(valuesMap);
		for(Integer run: runList) System.out.print(run + " "); // Display them
		System.out.println();
		
		// Begin merge into a single data set
		
		
		for(Integer run: runList){
			SingleDataSet ds = dsSource.getDataSet(run);
			ds.setStartDate(start);
			ds = ds.resample(maResampler, raResampler);
			analysis.merge(ds, "RUN_" + run + "_");
		}
		
		// Write the merged data set to the screen.
		analysis.report(true);
		
	}
	
}
