package cnh.crb.analysis.datasets;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.TreeMap;

import org.apache.commons.io.FileUtils;
import org.apache.commons.math3.stat.correlation.PearsonsCorrelation;
import org.apache.commons.math3.stat.regression.SimpleRegression;

public class SingleDataSet {

	public enum Function { SUM, ABS_DIFFERENCE, DIFFERENCE, AVERAGE, PRODUCT, MULTIPLY_BY, DELTA, SCALE_TO_INITIAL, SCALE_TO_VALUE, PERCENT_CHANGE, SQUARE_DIFF, LEAST_SQUARES, PCT_OF_TOTAL };
	public static int[][] functionArgCounts = 
		{{2, Integer.MAX_VALUE},  // SUM
		 {2, 2},                  // ABS_DIFFERENCE
		 {2, 2},                  // DIFFERENCE
		 {2, Integer.MAX_VALUE},  // AVERAGE
		 {2, Integer.MAX_VALUE},  // PRODUCT
		 {1, 1},                  // MULTIPLY_BY
		 {1, 1},                  // DELTA
		 {1, 1},                  // SCALE_TO_INITIAL
		 {1, 1},                  // SCALE_TO_VALUE
		 {1, 1},                  // PERCENT_CHANGE
		 {2, 2},                  // SQUARE_DIFF
		 {2, 2},                  // LEAST_SQUARES
	     {2, Integer.MAX_VALUE}}; // PCT_OF_TOTAL
	
	
	public static final Calendar defaultCalendar = Calendar.getInstance();
	
	
	public static boolean compareHeaders(SingleDataSet A, SingleDataSet B){
		// Note: List's built in 'equals' function MIGHT work- depends on String comparison...
		int aSize = A.headers.size();
		if(aSize != B.headers.size()) return false;
		for(int i = 0; i < aSize; i++) if(A.headers.get(i).compareTo(B.headers.get(i)) != 0) return false;
		return true;
	}
	
	public static boolean compareHeaders(SingleDataSet ... dataSets){
		if(dataSets.length == 0) return false;
		if(dataSets.length == 1) return true;
		SingleDataSet compare = dataSets[0];
		for(int i = 1; i < dataSets.length; i++) if(!compareHeaders(compare, dataSets[i])) return false;
		return true;
	}
	
	public static boolean compareTicks(SingleDataSet A, SingleDataSet B){
		if(A.data.size() != B.data.size()) return false;
		for(Double key: A.data.keySet()) if(!B.data.containsKey(key)) return false;
		return true;
	}
	
	public static boolean compareTicks(SingleDataSet ... dataSets){
		if(dataSets.length == 0) return false;
	    if(dataSets.length == 1) return true;
	    SingleDataSet compare = dataSets[0];
	    for(int i = 1; i < dataSets.length; i++) if(!compareTicks(compare, dataSets[i])) return false;
	    return true;
	}
	
	
    public static SingleDataSet aggregateDataSets(SingleDataSet.Function function, boolean keepOriginals, SingleDataSet ... dataSets){
    	if(dataSets.length == 0) return null;
    	if(function == Function.MULTIPLY_BY      ||
    	   function == Function.DELTA            ||
    	   function == Function.SCALE_TO_VALUE   ||
    	   function == Function.SCALE_TO_INITIAL ||
    	   function == Function.PERCENT_CHANGE   ||
    	   function == Function.PCT_OF_TOTAL) return null;
    	SingleDataSet ret = new SingleDataSet(dataSets[0], false);            // Must be at least one because of earlier test for zero
    	if(dataSets.length == 1) return ret;                                  // But if there is only one, return it
    	if(!compareHeaders(dataSets)) System.out.println(" HEADERS DO NOT MATCH");
    	if(!compareTicks(dataSets)) System.out.println(" TICKS DO NOT MATCH");
    	if(!compareHeaders(dataSets) || !compareTicks(dataSets)) return null; // If the data sets don't match, return null
    	// If the number of data sets is not appropriate to the function:
    	if(dataSets.length < functionArgCounts[function.ordinal()][0] || 
    	   dataSets.length > functionArgCounts[function.ordinal()][1]) return null;
    	for(Double tick: ret.data.keySet()){
    		List<String> destinationData = ret.data.get(tick);
    		for(int c = 0; c < ret.headers.size(); c++){
    			String val = destinationData.get(c);
    			Double initialValue = Double.NaN;
    			try{
    				initialValue = Double.parseDouble(val);
    			}
    			catch(NumberFormatException nfe){}
    			if(!Double.isNaN(initialValue)){
    				for(int i = 1; i < dataSets.length; i++){
    					String otherVal = dataSets[i].data.get(tick).get(c);
    					Double otherInitialValue = Double.NaN;
    					try{
    						otherInitialValue = Double.parseDouble(otherVal);
    					}
    					catch(NumberFormatException nfe){}
    					switch(function){
    					  case SUM:
    					  case AVERAGE:
    						  initialValue += otherInitialValue;
    						  break;
    					  case PRODUCT:
    						  initialValue *= otherInitialValue;
    						  break;
    					  case DIFFERENCE:
    					  case ABS_DIFFERENCE:
    					  case SQUARE_DIFF:
    						  initialValue -= otherInitialValue;
    					  default: // Do nothing
    					}
    				}
    				// At the end of the for loop
    				switch(function){
					  case AVERAGE:
						  initialValue /= (double)dataSets.length;
						  break;
					  case ABS_DIFFERENCE:
						  initialValue = Math.abs(initialValue);
					  case SQUARE_DIFF:
						  initialValue *= initialValue;
					  default: // Do nothing
    				}
    				destinationData.set(c, "" + initialValue);
    			}
    		}
    	}
    	if(keepOriginals){
    		int i = 1;
    		for(SingleDataSet dataSet: dataSets){
//    			System.out.println("MERGING + " + dataSet.id);
    			ret.merge(dataSet, (dataSet.id != null ? dataSet.id : "OTHER_" + i) + "_");
    			i++;
    		}
    	}
    	return ret;
    }
	

	
    public String                          id;      // Only used for some things
	public List<String>                    headers;
	public TreeMap<Double, List <String> > data;
	
	public double tickToDateConversion = 1;               // Default is that one tick is one day
	private Calendar startDate = Calendar.getInstance();   
	
	// For internal use only
	private SingleDataSet(){
		headers = new ArrayList<String>();
		data = new TreeMap<Double, List<String>>();
		setStartDate(startDate);
	}
	
	/**
	 * Constructs an instance of this by reading from
	 * a file.
	 * @param inputFile
	 */
	public SingleDataSet(List<String> emptyHeaders, Calendar start){
	  headers = new ArrayList<String>();
	  headers.addAll(emptyHeaders);
	  data = new TreeMap<Double, List<String>>();
	  setStartDate(start == null ? defaultCalendar : start);
	}
	
	/**
	 * This constructor copies the structure but NOT
	 * the data!
	 * @param copy
	 */
	public SingleDataSet(SingleDataSet copy){
		this(copy, true);
	}
	
	public SingleDataSet(SingleDataSet copy, boolean structureOnly){
		headers = new ArrayList<String>();
		headers.addAll(copy.headers);
		
		this.tickToDateConversion = copy.tickToDateConversion;
		this.startDate            = (Calendar)(copy.startDate.clone());
		
		data = new TreeMap<Double, List<String>>(); // Leave this empty
		if(!structureOnly){
			// deep clone
			for(Entry<Double, List<String>> entry: copy.data.entrySet()){
				List<String> toAdd = new ArrayList<String>();
				toAdd.addAll(entry.getValue());
				data.put(entry.getKey(), toAdd);
			}
		}
	}
	
	public SingleDataSet(String inputFile){
		this(inputFile, null);
	}
	public SingleDataSet(String inputFile, Calendar newStartDate){
		setStartDate(newStartDate == null ? defaultCalendar : newStartDate);
		data = new TreeMap<Double, List <String> >();
		File f = new File(inputFile);
		
		FileInputStream inputStream = null;
		Scanner         sc          = null;
		try {
		  inputStream = new FileInputStream(f);
		  sc = new Scanner(inputStream, "UTF-8");
		  
		  // First get the headers
		  headers = new ArrayList<String>();
		  int tickColumn = -1;
		  if(sc.hasNextLine()){
			  String[] headerTokens = sc.nextLine().replace('"', ' ').trim().split(",");
			  for(int i = 0; i < headerTokens.length; i++){
				  String header = headerTokens[i].trim();
				  if(header.compareTo("tick") == 0){
					  tickColumn = i;
				  }else{
					  headers.add(header); // Skip the first column
				  }
			  }
		  }
		  
		  String nextLine = "";
		  String[] tokens = new String[0];
		  while(sc.hasNextLine()){
			List<String> dataValues = new ArrayList<String>();
			nextLine = sc.nextLine();
			tokens = nextLine.replace('"', ' ').trim().split(",");

			double tick = -1;
			for(int i = 0; i < tokens.length; i++){
				if(i == tickColumn) tick = Double.parseDouble(tokens[i]);
				else				dataValues.add(tokens[i]);
			}
			data.put(tick, dataValues);	  
		  }
		}
        catch (Exception e) {
		  e.printStackTrace();
		} 
		 finally {
		    if (inputStream != null) {
		        try {
					inputStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		    }
		    if (sc != null) {
		        sc.close();
		    }
		}
	}
	
	
	/**
	 * Reads the data from one run out of a batch run output file.
	 * @param inputFile
	 * @param runNumber
	 */
	public SingleDataSet(String inputFile, int runNumber){
		setStartDate(defaultCalendar);
		File f = new File(inputFile);
		
		FileInputStream inputStream = null;
		Scanner         sc          = null;
		try {
		  inputStream = new FileInputStream(f);
		  sc = new Scanner(inputStream, "UTF-8");
		  
		  // First get the headers
		  headers = new ArrayList<String>();
		  int tickColumn = -1;
		  if(sc.hasNextLine()){
			  String[] headerTokens = sc.nextLine().replace('"', ' ').trim().split(",");
			  for(int i = 1; i < headerTokens.length; i++){
				  String header = headerTokens[i].trim();
				  if(header.compareTo("tick") == 0){
					  tickColumn = i;
				  }else{
					  headers.add(header); // Skip the first column
				  }
			  }
		  }
		  
		  // Now proceed until the requested run is found
		  boolean inRun = false;
		  String nextLine = "";
		  String[] tokens = new String[0];
		  while (sc.hasNextLine() && (inRun == false)) {
			nextLine = sc.nextLine(); 
			tokens = nextLine.replace('"', ' ').trim().split(",");
			if(Double.parseDouble(tokens[0]) == runNumber) inRun = true;
		  }
		  // Now loop through all the lines related to this run
		  data = new TreeMap<Double, List<String>>();
		  do{
			List<String> dataValues = new ArrayList<String>();
			double tick = -1;
			for(int i = 1; i < tokens.length; i++){
				if(i == tickColumn) tick = Double.parseDouble(tokens[i]);
				else				dataValues.add(tokens[i]);
			}
			data.put(tick, dataValues);
			if(sc.hasNextLine()){
		      nextLine = sc.nextLine();
			  tokens = nextLine.replace('"', ' ').trim().split(",");
			  if(Double.parseDouble(tokens[0]) != runNumber) inRun = false;
			}	  
		  } while(sc.hasNextLine()  && (inRun == true));
		}
        catch (Exception e) {
		  e.printStackTrace();
		} 
		 finally {
		    if (inputStream != null) {
		        try {
					inputStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		    }
		    if (sc != null) {
		        sc.close();
		    }
		}
	}
	
	
	public void writeToFile(String outputFile){
		List<String> outLines = new ArrayList<String>();
		String headerOutLine = "tick";
		for(String header: headers) headerOutLine += ((headerOutLine.length() > 0 ? "," : "") + header);
		outLines.add(headerOutLine);
		
		for(Entry<Double, List<String> > dataLine: data.entrySet()){
			String dataOutLine = dataLine.getKey() + "";
			for(String datum: dataLine.getValue()){
			    try{ Double.parseDouble(datum); }
			    catch(NumberFormatException nfe){ datum = "\"" + datum + "\""; }
			    dataOutLine += (dataOutLine.length() > 0 ? "," : "") + datum;
			}
			outLines.add(dataOutLine);
		}
		try{
		  FileUtils.writeLines(new File(outputFile), outLines);
		}
		catch(Exception E){
			System.out.println(E.getMessage());
		}
		
	}
	
	/**
	 * Return false if not found, or if new name would
	 * duplicate a name already in the data set. If these
	 * tests fail no change is made to the data set.
	 * 
	 * @param originalName
	 * @param newName
	 * @return
	 */
	public boolean renameColumn(String originalName, String newName){
		if(!headers.contains(originalName) ||
			headers.contains(newName)) return false;
		int i = headers.indexOf(originalName);
		headers.remove(i);
		headers.add(i, newName);
		return true;
	}
	
	
	public void report(){
		report(false);
	}

	public void report(boolean useDate){
		System.out.print((useDate ? "[date]" : "tick") + " | ");
		for(String header: headers) System.out.print(header + " | ");
		System.out.println();
		DateFormat df = new SimpleDateFormat("yyyy-MM-dd");
		for(Double tick: data.keySet()){
			System.out.print((useDate ? df.format(convertTickToCalendar(tick).getTime()) : tick) + " | ");
			for(String datum: data.get(tick)) System.out.print(datum + " | ");
			System.out.println();
		}
	}
	
	
	public void setStartDate(Calendar newStartDate){
	  startDate = (Calendar)newStartDate.clone();
	  startDate.add(Calendar.DATE, -1);
	}
	
	public Calendar convertTickToCalendar(double tick){
		Calendar cal = (Calendar)startDate.clone();
		double convertedTick = tick * tickToDateConversion; // Subtract one because ticks start with 1
		cal.add(Calendar.DATE, (int)Math.floor(convertedTick));
		if(convertedTick != Math.floor(convertedTick)){
			double remainder = tick - Math.floor(convertedTick);
			remainder *= 24; // Convert remaining portion of day to Hours
			int hours = (int)Math.floor(remainder);
			cal.add(Calendar.HOUR, hours);
			remainder -= hours;
			remainder *= 60; // Convert to minutes
			int minutes = (int)Math.floor(remainder);
			cal.add(Calendar.MINUTE, minutes);
			remainder -= minutes;
			remainder *= 60; // Convert to seconds
			int seconds = (int)Math.floor(remainder);
			cal.add(Calendar.SECOND, seconds);
		}
		// That's all, folks- no sub-seconds...
		return cal;
	}
	
	/**
	 * Adds the columns from the other data set to this one.
	 * 
	 * Tests to ensure that no columns in the other
	 * data set have the same names as columns in this one and that the
	 * list of ticks is identical.
	 * @param other
	 */
	public boolean merge(SingleDataSet other, String columnPrefix){
		for(String header: other.headers) 
			if(headers.contains(columnPrefix + header)){
				System.out.println("MERGE FAIL DUE TO DUPLICATE HEADERS (" + columnPrefix + header + ")");
				return false;
			}
		for(Entry<Double, List<String>> entry: other.data.entrySet())
			if(!data.containsKey(entry.getKey())){
				System.out.println("MERGE FAIL DUE TO MISMATCHED TICKS (" + entry.getKey() + " is not present in other");
				return false;
			}
		if(other.data.entrySet().size() != data.entrySet().size()){
			System.out.println("MERGE FAIL DUE TO DIFFERENT ENTRY SET SIZES ( local = " + data.entrySet().size() + " other = " + other.data.entrySet().size());
			return false;
		}
		for(String header: other.headers) headers.add(columnPrefix + header);
		for(Entry<Double, List<String>> entry: other.data.entrySet())
			    for(String otherDatum: entry.getValue()) data.get(entry.getKey()).add(otherDatum);
		return true;
	}
	
	
    public SingleDataSet resample(DataResampler ... resamplers){
    	SingleDataSet source = this;
		for(DataResampler resampler: resamplers) source = resampler.resample(source);
		System.gc();
		return source;
	}

    public boolean addCalculatedColumn(String newColumnName, Function function, String ... columns){
    	return addCalculatedColumn(newColumnName, function, null, columns);
    }

    
    public boolean addCalculatedColumn(String newColumnName, Function function, double[] args, String ... columns){
    	// Validation and prep
    	if((columns.length < functionArgCounts[function.ordinal()][0])  ||
    	   (columns.length > functionArgCounts[function.ordinal()][1])  ||
    	   headers.contains(newColumnName)                              ||
    	   ((args == null || args.length == 0) && 
    			              ((function == Function.MULTIPLY_BY)  || 
    	                       (function == Function.SCALE_TO_VALUE)))) return false;
    	
    	List<Integer> columnList = new ArrayList<Integer>();
    	for(String column: columns){
    		if(!headers.contains(column)) return false;
    		else columnList.add(headers.indexOf(column));
    	}
    	
    	SingleDataSet toMerge = new SingleDataSet();
    	toMerge.headers.add(newColumnName);
    	double previous   = Double.NaN;
    	for(Entry<Double, List<String>> entry: data.entrySet()){
    		// Grab all the values
    		List<Double> values = new ArrayList<Double>();
    		for(Integer index: columnList){
			  double val = Double.NaN;
			  try{ val = Double.parseDouble(entry.getValue().get(index)); }
			  catch(NumberFormatException nfe){}
			  values.add(val);
			}
    		
    		double valueToAdd = Double.NaN;
    		
    		switch(function){
    		  case SUM:
    			  valueToAdd = 0;
    			  for(Double value: values) valueToAdd += value;
    			  break;
    		  case ABS_DIFFERENCE:
    			  valueToAdd = Math.abs(values.get(0) - values.get(1));
    			  break;
    		  case DIFFERENCE:
    			  valueToAdd = values.get(0) - values.get(1);
    			  break;
    		  case PRODUCT:
    			  valueToAdd = 1;
    			  for(Double value: values) valueToAdd *= value;
    			  break;
    		  case MULTIPLY_BY:
    			  valueToAdd = values.get(0) * args[0];
    			  break;
    		  case DELTA:
    			  valueToAdd = (Double.isNaN(previous) ? 0 : values.get(0) - previous);
    			  previous = valueToAdd;
    			  break;
    		  case SCALE_TO_INITIAL:
    			  if(Double.isNaN(previous)) previous = values.get(0);
    			  valueToAdd = values.get(0) / previous;
    			  break;
    		  case SCALE_TO_VALUE:
    			  valueToAdd /= args[0];
    			  break;
    		  case PERCENT_CHANGE:
    			  valueToAdd = (Double.isNaN(previous) ? 0 : values.get(0) - previous) / previous;
    			  previous = valueToAdd;
    			  break;
    		  case SQUARE_DIFF:
    			  valueToAdd = Math.pow(values.get(1) - values.get(0), 2); 
    			  break;
    		  case LEAST_SQUARES: // N/A
    			  break;
    		  case PCT_OF_TOTAL:
    			  valueToAdd = 0;
    			  for(Double value: values)  valueToAdd += value;
    			  valueToAdd = values.get(0) / valueToAdd;
    			  break;
    		  case AVERAGE:
    		  default:
    			  valueToAdd = 0;
    			  for(Double value: values) valueToAdd += value;
    			  valueToAdd /= columns.length;
    			  break;
    		}

    		if(Double.isNaN(valueToAdd)) return false;
    		List<String> listToAdd = new ArrayList<String>();
    		listToAdd.add(valueToAdd + "");
    		toMerge.data.put(entry.getKey(), listToAdd);
    	}
    	
    	merge(toMerge, "");
    	return true;
    }
    
    public double getAggregate(Function function, String ... columnHeaders){
    	for(String header: columnHeaders) if(!headers.contains(header)) return Double.NaN;
    	switch(function){
    	  case DIFFERENCE:
    	  case PRODUCT:
    	  case MULTIPLY_BY:
    	  case DELTA:
    	  case SCALE_TO_INITIAL:
    	  case SCALE_TO_VALUE:
    	  case PERCENT_CHANGE:
    	  case SQUARE_DIFF:{
    		  int i = headers.indexOf(columnHeaders[0]);
    		  int j = headers.indexOf(columnHeaders[1]);
    		  double sum = 0;
    		  double count = 0;
    		  for(List<String> dataLine: data.values()){
    			  double val1 = Double.NaN;
    			  double val2 = Double.NaN;
    			  try{
    				  val1 = Double.parseDouble(dataLine.get(i));
    				  val2 = Double.parseDouble(dataLine.get(j));
    			  }
    			  catch(NumberFormatException E){}
    			  sum += (val1 - val2) * (val1 - val2);
    			  count++;
              }
    		  return Math.sqrt(sum/count);
    	  }
    	  case ABS_DIFFERENCE:{
    		  int i = headers.indexOf(columnHeaders[0]);
    		  int j = headers.indexOf(columnHeaders[1]);
    		  double sum = 0;
    		  for(List<String> dataLine: data.values()){
    			  double val1 = Double.NaN;
    			  double val2 = Double.NaN;
    			  try{
    				  val1 = Double.parseDouble(dataLine.get(i));
    				  val2 = Double.parseDouble(dataLine.get(j));
    			  }
    			  catch(NumberFormatException E){}
    			  sum += Math.abs(val1 - val2);
              }
    		  return sum;
    	  }
    	  case SUM:{
    		  double sum = 0;
    		  int i = headers.indexOf(columnHeaders[0]);
    		  for(List<String> dataLine: data.values()){
    			  double val = Double.NaN;
    		      try{
    		    	  val = Double.parseDouble(dataLine.get(i));
    		      }
    		      catch(NumberFormatException E){}
    			  sum += val;
              }
              return sum;
    	  }
    	  case AVERAGE:{
    		  double sum = 0;
    		  double count = 0;
    		  int i = headers.indexOf(columnHeaders[0]);
    		  for(List<String> dataLine: data.values()){
    			  double val = Double.NaN;
    			  count++;
    		      try{
    		    	  val = Double.parseDouble(dataLine.get(i));
    		      }
    		      catch(NumberFormatException E){}
    			  sum += val;
              }
              return sum / (double)count;
    	  }
    	  case LEAST_SQUARES:{
    		  int i = headers.indexOf(columnHeaders[0]);
    		  int j = headers.indexOf(columnHeaders[1]);
    		  SimpleRegression ordinaryLeastSquares = new SimpleRegression();
    		  for(List<String> dataLine: data.values()){
    			  double val1 = Double.NaN;
    			  double val2 = Double.NaN;
    			  try{
    				  val1 = Double.parseDouble(dataLine.get(i));
    				  val2 = Double.parseDouble(dataLine.get(j));
    			  }
    			  catch(NumberFormatException E){}
    			  ordinaryLeastSquares.addData(val1, val2);
              }
//    		  return ordinaryLeastSquares.getR();
    		  return ordinaryLeastSquares.getSlope();
    	  }
       }
       return Double.NaN;
    }
    
    
    public double regression(String column1, String column2){
    	if(!headers.contains(column1) || !headers.contains(column2)) return Double.NaN;
    	int c1 = headers.indexOf(column1);
    	int c2 = headers.indexOf(column2);
    	double[] array1 = new double[data.size()];
    	double[] array2 = new double[data.size()];
    	
    	int i = 0;
    	for(List<String> dataLine: data.values()){
    		array1[i] = Double.parseDouble(dataLine.get(c1));
    		array2[i] = Double.parseDouble(dataLine.get(c2));
    		i++;
    	}
    	return (new PearsonsCorrelation()).correlation(array1, array2);
    }
    
	public static void main(String[] args){
		SingleDataSet dataSet200 = new SingleDataSet("./out/combined_data_2014_09_08/ModelOutput.2014.Sep.08.02_42_01.txt", 200);
		SingleDataSet dataSet111 = new SingleDataSet("./out/combined_data_2014_09_08/ModelOutput.2014.Sep.08.02_42_01.txt", 111);
		Calendar cal = Calendar.getInstance();
		cal.set(1993, 1, 0, 0, 0, 0);
		dataSet200.setStartDate(cal);
		dataSet111.setStartDate(cal);
		MonthlyAverageResampler maResampler = new MonthlyAverageResampler();
		RollingAverageResampler raResampler = new RollingAverageResampler(12);
		SingleDataSet resampled200 = dataSet200.resample(maResampler, raResampler);
		SingleDataSet resampled111 = dataSet111.resample(maResampler, raResampler);
		System.out.println();
		resampled200.renameColumn("waterUsed", "Run_200_waterUsed");
		resampled200.merge(resampled111, "Run_111_");
		resampled200.addCalculatedColumn("Average", Function.AVERAGE, "Run_200_waterUsed", "Run_111_waterUsed");
		resampled200.addCalculatedColumn("Difference", Function.DIFFERENCE, "Run_200_waterUsed", "Run_111_waterUsed");
		resampled200.addCalculatedColumn("Sum", Function.SUM, "Run_200_waterUsed", "Run_111_waterUsed");
		resampled200.report(true);
		System.out.println(resampled200.regression("Run_200_waterUsed", "Run_111_waterUsed"));
		System.out.println("Done.");
	}
}
