package cnh.crb.analysis.parametersets;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.io.FileUtils;

import repast.simphony.util.collections.Pair;

public class BatchParametersSet {
	
	private String                             id;
	private Map<Integer, Map<String, String> > paramSet;   // The map of all parameters and values, by Run Number
	private List<String>                       paramList;  // The list of all parameters by parameter name
	
	private static int setCount = 0;
	
	private static String getShortId(){
		setCount++;
		return "BATCH_SET_" + setCount;
	}

	
	/**
	 * Returns true if all elements in the first map are equal
	 * to corresponding elements in the second map. Used to compare
	 * sets of parameters.
	 * @param paramSetA
	 * @param paramSetB
	 * @return
	 */
	private static boolean allParametersAreEqual(Map<String, String> paramSetA, Map<String, String> paramSetB){
		if(paramSetA.size() != paramSetB.size()) return false;
		for(Entry<String, String> entry: paramSetA.entrySet()){
			if((!paramSetB.containsKey(entry.getKey())) || 
			   (entry.getValue().compareTo(paramSetB.get(entry.getKey())) != 0)){
				return false;
			}
		}
		return true;
	}
	
	
	/**
	 * Given a set of parameter sets and a list of parameters to ignore,
	 * returns a collection of unique sets of parameter values
	 * and the runs that have those values. The collection structure is:
	 * 
	 * List<
	 *   Pair<
	 *     Map< String, String >
	 *     Map< String, List< Integer > > > >
	 *     
	 * This is a list of pairs; each pair includes two elements:
	 * 
	 *      The (unique) set of parameters and values
	 *      The runs that this set has, mapped by Batch Name
	 * @param paramSets
	 * @param parametersToIgnore
	 * @return
	 */
	public static List<Pair<Map<String, String>, Map<BatchParametersSet, List<Integer>>>> listRunsWithCommonParameters(List<BatchParametersSet> paramSets, String ... parametersToIgnore){
		List<Pair<Map<String, String>, Map<BatchParametersSet, List<Integer>>>> ret = new ArrayList<Pair<Map<String, String>, Map<BatchParametersSet, List<Integer>>>>();
		
		for(BatchParametersSet paramSet: paramSets){
			List<Pair<Map<String, String>, List<Integer>>> basicList = paramSet.listRunsWithCommonParameters(parametersToIgnore);
			for(Pair<Map<String, String>, List<Integer>> pair: basicList ){
				Map<String, String> paramValues = pair.getFirst();
				// Find an existing match?
				boolean matched = false;
				for(Pair<Map<String, String>, Map<BatchParametersSet, List<Integer>>> matchPair: ret){
					Map<String, String> matchValues = matchPair.getFirst();
					if(allParametersAreEqual(paramValues, matchValues)){
//						System.out.println("Found a new match with " + pair.getSecond().size() + " runs (" + ret.size() + ")");
						matched = true;
						matchPair.getSecond().put(paramSet, pair.getSecond());
					}
				}
				if(!matched){
				  // Must add new entry to ret
//				  System.out.println("Adding a new entry with " + pair.getSecond().size() + " runs (" + (ret.size() + 1) + ")");
				  Map<BatchParametersSet, List<Integer>> toAdd = new HashMap<BatchParametersSet, List<Integer>>();
				  toAdd.put(paramSet, pair.getSecond());
				  Pair<Map<String, String>, Map<BatchParametersSet, List<Integer>>> pairToAdd = 
						  new Pair<Map<String, String>, Map<BatchParametersSet, List<Integer>>>(paramValues, toAdd);
				  ret.add(pairToAdd);
				}
			}
		}
		System.out.println("Returning with " + ret.size());
		return ret;
	}
	
	
	public BatchParametersSet(List<String> parameters){
		id = getShortId();
		paramSet  = new TreeMap<Integer, Map<String, String>>();
		paramList = new ArrayList<String>();
		for(String param: parameters) paramList.add(param); // Preserve order
	}
	
	public BatchParametersSet(String fileName){
		this(fileName, "BATCH_SET_" + getShortId());
	}
	
	public BatchParametersSet(String fileName, String shortID){
		id = shortID.substring(0, Math.max(shortID.length(), 20));
		paramSet  = new TreeMap<Integer, Map<String, String>>();
		paramList = new ArrayList<String>();
		// Load file, populating these sets
		FileInputStream inputStream = null;
		Scanner sc = null;
		try{
		  File file = new File(fileName);
		    
		  inputStream = new FileInputStream(file);
		  sc = new Scanner(inputStream, "UTF-8");
		  
		  // First do the header line
		  if(sc.hasNextLine()){
			String nextLine = sc.nextLine().replace('"', ' ').trim();
			String[] headers = nextLine.split(",");
			for(String headerEntry: headers) paramList.add(headerEntry.trim());
		  }
		  // Now do the individual runs
		  while (sc.hasNextLine()) {
		    String nextLine = sc.nextLine().replace('"', ' ').trim();
		    String[] data = nextLine.split(",");
			int col = 0;
			int runNumber = -1;
			Map<String, String> paramValues = new HashMap<String, String>();
			for(String entry: data){
			  if(col == 0) runNumber = Integer.parseInt(entry);
			  paramValues.put(paramList.get(col), entry);
              col++;
			}
			paramSet.put(runNumber, paramValues);
		  }
		}
		catch(Exception E){
			System.out.println(E.getMessage());
			// Do snazzy error trapping here
		}
		finally {
		    if (inputStream != null) {
		        try {
					inputStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		    }
		    if (sc != null) {
		        sc.close();
		    }
		}
	}
	
	public void addRun(Integer runNumber, Map<String, String> values){
		paramSet.put(runNumber, values);
	}
	
	/**
	 * Writes the entire parameter set to a file as a csv
	 */
	public void writeToFile(String outputFileName){
		int[] runNumbers = new int[paramSet.size()];
		int i = 0;
		for(Integer integer: paramSet.keySet()){
			runNumbers[i] = integer;
			i++;
		}
		writeToFile(outputFileName, runNumbers);
	}
	
	/**
	 * Writes the parameters for the specified runs
	 * as a csv file.
	 * @param runNumber
	 */
	public void writeToFile(String outputFileName, int ... runNumbers){
		// Bail if the file already exists
		try{
			if((new File(outputFileName).exists())){
				System.out.println("File '" + outputFileName + "' already exists; not writing...");
				return;
			}
		}
		catch(Exception E){
			System.out.println("COULD NOT PERFORM CHECK OF FILE EXIST...");
			System.out.println(E.getMessage());
			return;
		}
		List<String> outlines = new ArrayList<String>();
		
		TreeSet<Integer> runNumberSet = new TreeSet<Integer>();
		for(int run: runNumbers) runNumberSet.add(run);
		// Write the headers
		String headerLine = "run";
		for(String param: paramList) { headerLine += (headerLine.length() > 0 ? "," : "") + param; }
		// Write the header line
		outlines.add(headerLine);
		for(Integer runNumber: runNumberSet){
			String nextLine = "" + runNumber;
			Map<String, String> parameters = paramSet.get(runNumber);
			for(String parameter: parameters.keySet()){
				String valueToWrite = parameters.get(parameter);
				try{
					Double.parseDouble(valueToWrite); // If it's not a number, it's a string; should have quotation marks
				}
				catch(NumberFormatException E){
					valueToWrite = "\"" + valueToWrite + "\"";
				}
				nextLine += (nextLine.length() > 0 ? "," : "") + valueToWrite;
			}
			// Write the next line. Why do I find it hard to write the next line?
			outlines.add(nextLine);
		}
		// Write the file
		try{
			FileUtils.writeLines(new File(outputFileName), outlines);
		}
		catch(Exception E){
			System.out.println("FAILED writing '" + outputFileName + "'");
		}
		
	}
	
	/**
	 * Returns a map of the parameter keys and values such that all
	 * runs have the same value for the listed parameters.
	 * @return
	 */
	public Map<String, String> getParametersCommonToAllRuns(String ... parametersToIgnore){
		Map<String, String> commonParams = null;
		for(Integer runNumber: paramSet.keySet()){
			if(commonParams == null){ // On first loop through
				commonParams = new TreeMap<String, String>();
				commonParams.putAll(paramSet.get(runNumber));
				for(String paramToIgnore: parametersToIgnore) commonParams.remove(paramToIgnore);
				continue; // Bypass the rest of the for loop
			}
			Map<String, String> testParams = paramSet.get(runNumber);
			Set<String> paramsToRemove = new TreeSet<String>();
			for(String param: commonParams.keySet())
				if(commonParams.get(param).compareTo(testParams.get(param)) != 0) paramsToRemove.add(param); // Can't remove while looping
			for(String param: paramsToRemove) commonParams.remove(param);
			if(commonParams.size() == 0) break; // None apply
		}
		return commonParams;
	}
	
	/**
	 * Gets a list of elements where each element is a pair that includes
	 * 
	 *   1- A Map of String, String values that represents the unique combination
	 *           of parameters
	 *   2- The list of run numbers that have that combination
	 *   
	 * The argument that is passed to this function is a list of parameters to
	 * ignore. Runs that differ only on these parameter values will be grouped
	 * as if they are identical, and the parameter will not be listed in the
	 * parameter map. Commonly this set of ignored parameters will include the 
	 * random number seed. Note that the parameters returned by getParametersCommonToAllRuns
	 * are AUTOMATICALLY ignored by this function and do not appear in the maps
	 * returned.
	 * @param parametersToIgnore
	 * @return
	 */
	public List<Pair<Map<String, String>, List<Integer>>> listRunsWithCommonParameters(String ... parametersToIgnore){
		Map<String, String> commonParams = getParametersCommonToAllRuns(parametersToIgnore); // Don't need to check these
		List<String> parametersToCheck = new ArrayList<String>();
		parametersToCheck.addAll(paramList);
		parametersToCheck.removeAll(commonParams.keySet());
		for(String paramToIgnore: parametersToIgnore) parametersToCheck.remove(paramToIgnore);
		
		List<Pair<Map<String, String>, List<Integer>>> ret = new ArrayList<Pair<Map<String, String>, List<Integer>>>();
		for(Integer runNumber: paramSet.keySet()){
			if(ret.size() == 0){ // First time through loop
				Map<String, String> firstMap = new TreeMap<String, String>();
				Map<String, String> paramVals = new TreeMap<String, String>();
				paramVals.putAll(paramSet.get(runNumber));
				for(String parameterToCheck: parametersToCheck) firstMap.put(parameterToCheck, paramVals.get(parameterToCheck));
				List<Integer> firstList = new ArrayList<Integer>();
				firstList.add(runNumber);
				ret.add(new Pair<Map<String, String>, List<Integer>>(firstMap, firstList));
				continue; // Skip the rest and just continue the loop
			}
			// Check previously entered values for a match:
			Map<String, String> paramVals = new TreeMap<String, String>();
			paramVals.putAll(paramSet.get(runNumber));
			for(String paramToIgnore: commonParams.keySet()) paramVals.remove(paramToIgnore);
			for(String paramToIgnore: parametersToIgnore)    paramVals.remove(paramToIgnore);
			boolean found = false;
			for(Pair<Map<String, String>, List<Integer>> pair: ret){
				Map<String, String> compareMap = pair.getFirst();
				if(allParametersAreEqual(compareMap, paramVals)){
					pair.getSecond().add(runNumber);
					found = true;
				}
				if(found == true) break;
			}
			if(found == false){
				List<Integer> newList = new ArrayList<Integer>();
				newList.add(runNumber);
				ret.add(new Pair<Map<String, String>, List<Integer>>(paramVals, newList));
			}
		}
		return ret;
	}
	
	/**
	 * Returns a String that shows the results of the listRunsWithCommonParameters.
	 * String can be in list or table format, and can include either the list of runs
	 * or just the count of matching runs.
	 * @param parametersToIgnore
	 */
	public String showRunsWithCommonParameters(boolean useTableForm, boolean listRuns, String ... parametersToIgnore){
		StringWriter ret = new StringWriter();
		
		List<Pair<Map<String, String>, List<Integer>>> masterList = listRunsWithCommonParameters(parametersToIgnore);
		Set<String> paramsList = new TreeSet<String>();
		int longestParamWidth = 0;
		for(Pair<Map<String, String>, List<Integer>> entry: masterList){
			for(String key: entry.getFirst().keySet()){
				paramsList.add(key);
				longestParamWidth = Math.max(longestParamWidth, key.length());
				longestParamWidth = Math.max(longestParamWidth, entry.getFirst().get(key).length());
			}
		}
		longestParamWidth += 1;
		String buffer = "";
		for(int i = 0; i < longestParamWidth; i++) buffer += " ";
		if(useTableForm){
			for(String param: paramsList){
				String paramForm = (param + buffer).substring(0, longestParamWidth);
				ret.write(paramForm);
			}
			ret.write("Runs");
			ret.write(System.getProperty("line.separator"));
		}
		
		for(Pair<Map<String, String>, List<Integer>> entry: masterList){
			if(useTableForm){
				for(String param: paramsList){
					String valForm = entry.getFirst().get(param);
					if(valForm != null) valForm = (valForm + buffer).substring(0, longestParamWidth);
					ret.write(valForm);
				}
			}
			if(listRuns){
				String runList = "";
				for(Integer run: entry.getSecond()) runList += runList.length() > 0 ? ", " : run;
				ret.write(runList);
			}
			else{
				ret.write("[" + entry.getSecond().size() + " runs]");
			}
			ret.write(System.getProperty("line.separator"));
		}
		return ret.toString();
	}
	

	
	/**
	 * Get the count of runs contained in this set.
	 * @return
	 */
	public int getCountOfEntries(){
		return this.paramSet.size();
	}
	
	/**
	 * Gets a map that lists each parameter for which values vary among this set
	 * and the list of distinct values it has.
	 * @param paramsToIgnore
	 * @return
	 */
	public Map<String, Set<String>> getDistinctValues(String ... paramsToIgnore){
		Map<String, Set<String>> ret = new HashMap<String, Set<String>>();
		Set<String> params = new TreeSet<String>();
		params.addAll(paramList);
		for(String paramToIgnore: paramsToIgnore)                                        params.remove(paramToIgnore);
		for(String paramToIgnore: getParametersCommonToAllRuns(paramsToIgnore).keySet()) params.remove(paramToIgnore);
		for(Integer runNumber: paramSet.keySet()){
			if(ret.size() == 0){
				for(String param: params){
					Set<String> valueList = new TreeSet<String>();
					valueList.add(paramSet.get(runNumber).get(param));
					ret.put(param, valueList);
				}
			}
			else{
				for(String param: params){
					Set<String> valueList = ret.get(param);
					valueList.add(paramSet.get(runNumber).get(param));
				}
			}
		}
		return ret;
	}
	
	/**
	 * Returns a string report of the results of the getDistinctValues call
	 * @param paramsToIgnore
	 * @return
	 */
	public String showDistinctValues(String ... paramsToIgnore){
		Map<String, Set<String>> map = getDistinctValues(paramsToIgnore);
		String ret = "";
		for(Entry<String, Set<String>> entry: map.entrySet()){
			ret += entry.getKey() + ": ";
			for(String val: entry.getValue()) ret+= val + " ";
			ret += System.getProperty("line.separator");
		}
		return ret;
	}
	
	/**
	 * Gets a list of the runs that have the specified value for the
	 * specified parameter.
	 * @param param
	 * @param value
	 * @return
	 */
	public List<Integer> getRunsWithMatchingValue(String param, String value){
		List<Integer> ret = new ArrayList<Integer>();
		Map<String, String> map = new HashMap<String, String>();
		map.put(param, value);
		for(Integer runNumber: paramSet.keySet())
			if(allParametersAreEqual(map, paramSet.get(runNumber))) ret.add(runNumber);
		return ret;
	}
	
	

	/**
	 * Gets all runs with values that match one of the values in the specified
	 * list, per parameter.
	 * @param valuesToMatch
	 * @return
	 */
	public List<Integer> getRunsWithMatchingValues(Map<String, List<String>> valuesToMatch){
		List<Integer> ret = new ArrayList<Integer>();
		for(Integer runNumber: paramSet.keySet()){
			boolean matches = true;
			Map<String, String> paramVals = paramSet.get(runNumber);
			for(String param: valuesToMatch.keySet()){
				String val = paramVals.get(param);
				List<String> compare = valuesToMatch.get(param);
				if(!compare.contains(val)){
					matches = false;
					break;
				}
			}
			if(matches) ret.add(runNumber);
		}
		return ret;
	}
	
	public String getParameterValue(Integer run, String parameter){
		return paramSet.get(run).get(parameter);
	}
	
	public Set<Integer> getListOfRuns(){
		Set<Integer> allRuns = new TreeSet<Integer>();
		for(Integer i: paramSet.keySet()) allRuns.add(i);
		return allRuns;
	}
	
	public static void main(String[] args){
		BatchParametersSet batchParams = new BatchParametersSet("./out/combined_data_2014_09_08/ModelOutput.2014.Sep.08.02_42_01.batch_param_map.txt");
//		List<Pair<Map<String, String>, List<Integer>>> commonList = batchParams.listRunsWithCommonParameters("randomSeed", "run", "socialFactorWeight");
//		Map<String, List<String>> valuesToMatch = new TreeMap<String, List<String>>();
//		String[][] values = { {"priceWeight", "0.2"},
//				              {"messagingSubsampling", "0.5"},
//		                      {"messagingEffectiveness", "0.1"},
//		                      {"waterUseBottom", "0.0"},
//		                      {"messagingBeginYear","1996"}
//		};
//		for(String[] value: values){
//			List<String> toAdd = new ArrayList<String>();
//			toAdd.add(value[1]);
//			valuesToMatch.put(value[0], toAdd);
//		}
//		
//		List<Integer> runsToExamine = batchParams.getRunsWithMatchingValues(valuesToMatch);
//		for(Integer run: runsToExamine) System.out.println(run + " " + batchParams.getParameterValue(run, "socialFactorWeight"));
		System.out.println(batchParams.showDistinctValues("run"));
		System.out.println(batchParams.showRunsWithCommonParameters(true, true, "run"));
	}
	
	
	
}
