package cnh.crb.econ;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import repast.simphony.util.collections.Pair;

/**
 * Cost calculator in which consumption above specific thresholds is
 * billed at a higher rate. May include a flat fee for service. Note that ONLY
 * the excess above the threshold is billed at the higher rate (c.f. the 'Step
 * Rate' calculator).
 * 
 * @author murphy
 * 
 */
public class BlockRateCostCalculator implements CostCalculator {

	protected double baseFee = 0;

	protected List<Pair<Double, Double>> rateBlocks;

	/**
	 * Initializes this BlockRateBillCalculator with two arrays. The first contains
	 * the threshold values, the second the rates for values equal to or below 
	 * the corresponding threshold. The rates array should contain an additional
	 * final entry that is the rate for consumption above the greatest threshold.
	 * 
	 * Note that thresholds are in absolute consumption (that is, consumption
	 * above zero, not consumption above the preceding cutoff).
	 * 
	 * @param thresholds
	 * @param rates
	 * @param flatFee
	 */
	public BlockRateCostCalculator(double[] thresholds, double[] rates, double flatFee) {
		if (thresholds.length != rates.length - 1) 
			throw new IllegalArgumentException("Rates must contain one more entry than thresholds");
		
		// use the map to sort thresholds.
		TreeMap<Double, Double> map = new TreeMap<Double, Double>();
		for (int i = 0; i < rates.length - 1; i++) {
			map.put(thresholds[i], rates[i]);
		}
		map.put(Double.MAX_VALUE, rates[rates.length - 1]);
		
		rateBlocks = new ArrayList<Pair<Double, Double>>();
		
		// rate block stores the ifference between one threshold
		// and its preceeding one
		double prev = 0;
		for (Map.Entry<Double, Double> entry : map.entrySet()) {
			rateBlocks.add(new Pair<Double, Double>(entry.getKey() - prev, entry.getValue()));
			prev = entry.getKey();
		}
		
		baseFee = flatFee;
	}

	/**
	 * Convenience, if there is no flat fee.
	 * 
	 * @param thresholds
	 * @param rates
	 */
	public BlockRateCostCalculator(double[] thresholds, double[] rates) {
		this(thresholds, rates, 0.0);
	}

	/**
	 * Gets the cost for that amount of water used.
	 */
	public double calculateCost(Date date, double ccf, BillableEntity entity) {
		return amount(ccf);
	}

	protected double amount(double ccf) {
		double billAmount = 0;
		double unbilledUsage = ccf;
		for (Pair<Double, Double> rateBlock : rateBlocks) {
			double threshold = rateBlock.getFirst();
			double rate = rateBlock.getSecond();
			if (unbilledUsage > threshold) {
				billAmount +=  threshold * rate;
				unbilledUsage -= threshold;
			} else {
				billAmount += unbilledUsage * rate;
				break;
			}
		}
		return billAmount + baseFee;
	}
}
