package cnh.crb.econ;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;

/**
 * Implements a Cost calculator that can have multiple different rate structures
 * depending on the month of the year.
 * 
 * @author murphy
 * 
 */
public class SeasonalCostCalculator implements CostCalculator {

	private Map<Integer, CostCalculator> calculatorByMonth = new HashMap<Integer, CostCalculator>();
	private Calendar cal = GregorianCalendar.getInstance();

	/**
	 * Sets a Cost calculator to be used in specific months. For example:
	 * 
	 * SeasonalCostCalculator sbc = new SeasonalCostCalculator();
	 * 
	 * CostCalculator winter = new CostCalculator(); // Of any kind...
	 * sbc.setCalculator(winter, 0, 1, 2, 3, 4, 10, 11);
	 * 
	 * CostCalculator summer = new CostCalculator(); // Of any kind
	 * sbc.setcalculator(summer, 5, 6, 7, 8, 9);
	 * 
	 * @param calculator
	 * @param monthList
	 */
	public void setCalculator(CostCalculator calculator, int... monthList) {
		for (int i : monthList)
			calculatorByMonth.put(i, calculator);
	}
	
	/**
	 * Calculates the cost for the specified date and amount.
	 * 
	 * @param date
	 * @param amt
	 * 
	 * @return the cost for the specified date and amount.
	 */
	public double calculateCost(Date date, double amount, BillableEntity entity) {
		cal.setTime(date);
		int month = cal.get(Calendar.MONTH);
		CostCalculator calc = calculatorByMonth.get(month);
		if (calc == null)
			throw new IllegalArgumentException("Missing calculator for month: " + month);
		return calc.calculateCost(date, amount, entity);
		
	}
}
