/**
 * 
 */
package cnh.crb.econ;

import java.util.Date;
import java.util.Map;
import java.util.TreeMap;

/**
 * Calculates the surcharge, if any, given the amount of water used and a value
 * to compare that use against.
 * 
 * @author nick
 */
public abstract class SurchargeCalculator implements CostCalculator{

	public static class Surcharge {
		double rate1;
		double rate2;
		double rate2ThresholdAsFraction;
		
		public Surcharge(double rate1, double rate2, double rate2Threshold) {
			this.rate1 = rate1;
			this.rate2 = rate2;
			this.rate2ThresholdAsFraction = rate2Threshold;
		}

		public double calculateCharge(double comparandum, double amount) {
			if (amount <= comparandum) return 0;
			double charge = (amount - comparandum) * rate1;
			double secondChargeLevel = comparandum * rate2ThresholdAsFraction;
			if (amount > secondChargeLevel)
				charge += (amount - secondChargeLevel) * rate2;
			return charge;
		}
	}

	private TreeMap<Date, Surcharge> surchargeMap = new TreeMap<>();

	public SurchargeCalculator(Map<Date, Surcharge> surcharges) {
		this.surchargeMap.putAll(surcharges);
	}

	/**
	 * Calculates a surcharge for the amount of water used on the specified date.
	 * Child classes should implement a means to retrieve the number against which
	 * the specified usage is compared
	 * 
	 * @param date
	 * @param amtToCompareTo
	 * @param amountUsed
	 * 
	 * @return the calculated surcharge.
	 */
	public double calculateCost(Date date, double amt, BillableEntity entity) {
		Surcharge surcharge = getSurcharge(date);
		double amtToCompareTo = getAmountToCompare(date, amt, entity);
		System.out.println("CALCULATING COST: " + amtToCompareTo + " " + amt);
		return (amtToCompareTo >= 0 ? surcharge.calculateCharge(amtToCompareTo, amt) : 0);
	}

	/**
	 * Gets the surcharge (object) associated with the specified date
	 * @param date
	 * @return
	 */
	private Surcharge getSurcharge(Date date) {
		Map.Entry<Date, Surcharge> entry = surchargeMap.floorEntry(date);
		if (entry == null)
			throw new IllegalArgumentException("No data for the period that includes " + date);
		return entry.getValue();
	}
	
	
	/**
	 * Child classes should provide the way that instances get the amount
	 * from the billable entity that will be used for comparison.
	 * @param date
	 * @param amt
	 * @param entity
	 * @return
	 */
	public abstract double getAmountToCompare(Date date, double amt, BillableEntity entity);
	

}
