/**
 * 
 */
package cnh.crb.econ.io;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import cnh.crb.econ.BlockRateCostCalculator;
import cnh.crb.econ.CostCalculator;
import cnh.crb.econ.IncreasingStepRateCostCalculator;
import cnh.crb.econ.MultiSeasonCostCalculator;
import cnh.crb.econ.SeasonalCostCalculator;

/**
 * @author nick
 */
public class RateLoader {

	private static class SeasonCalc {
		String id;
		Map<String, String> entries = new HashMap<>();

		void createCalculator(Map<String, CostCalculator> map) throws IOException {
			SeasonalCostCalculator calc = new SeasonalCostCalculator();

			if (entries.size() == 0) {
				throw new IOException("Invalid rate file. Bad seasonal_calculator "
						+ this.id + " contains not entries.");
			}
			
			for (String id : entries.keySet()) {
				CostCalculator bc = map.get(id);
				if (bc == null)
					throw new IOException("Invalid rate file. Bad seasonal_calc " + this.id
							+ ": rate table " + id + " not found.");

				String[] sMonths = entries.get(id).split(",");
				int[] months = new int[sMonths.length];
				try {
					for (int i = 0; i < sMonths.length; ++i) {
						months[i] = Integer.parseInt(sMonths[i].trim());
					}
				} catch (NumberFormatException ex) {
					throw new IOException("Invalid rate file. Bad seasonal_calc " + this.id
							+ ": invalid months");
				}
				calc.setCalculator(bc, months);
			}
			map.put(id, calc);
		}
	}

	private static class MultiSeasonCalc {
		String id;
		Map<String, String> entries = new HashMap<>();

		void createCalculator(Map<String, CostCalculator> map) throws IOException {
			MultiSeasonCostCalculator calc = new MultiSeasonCostCalculator();
			if (entries.size() == 0) {
				throw new IOException("Invalid rate file. Bad multi_seasonal_calculator "
						+ this.id + " contains not entries.");
			}
			for (String id : entries.keySet()) {
				CostCalculator bc = map.get(id);
				if (bc == null)
					throw new IOException("Invalid rate file. Bad multi_seasonal_calculator "
							+ this.id + ": seasonal_calculator " + id + " not found.");

				if (!(bc instanceof SeasonalCostCalculator)) {
					throw new IOException("Invalid rate file. Bad multi_seasonal_calculator "
							+ this.id + ": seasonal_calculator " + id
							+ " is not a seasonal_calculator.");
				}

				String[] sYears = entries.get(id).split(",");
				int[] years = new int[sYears.length];
				try {
					for (int i = 0; i < sYears.length; ++i) {
						years[i] = Integer.parseInt(sYears[i].trim());
					}
				} catch (NumberFormatException ex) {
					throw new IOException("Invalid rate file. Bad multi_seasonal_calc " + this.id
							+ ": invalid years");
				}
				calc.addSeasonalCalculator((SeasonalCostCalculator) bc, years);
			}
			map.put(id, calc);
		}
	}

	private static final String BLOCK = "block";
	private static final String STEP = "step";
	private static final String MAX = "max";

	private static final QName RATES = new QName("rates");
	private static final QName RATE_TABLE = new QName("rate_table");
	private static final QName RATE_BLOCK = new QName("rate_block");
	private static final QName SEASONAL_CALC = new QName("seasonal_calculator");
	private static final QName ENTRY = new QName("entry");
	private static final QName USE = new QName("use");
	private static final QName TYPE = new QName("type");
	private static final QName ID = new QName("id");
	private static final QName FLAT_FEE = new QName("flat_fee");
	private static final QName THRESHOLD = new QName("threshold");
	private static final QName RATE = new QName("rate");
	private static final QName MONTHS = new QName("months");

	private static final QName MULTI_SEASONAL_CALC = new QName("multi_seasonal_calculator");
	private static final QName YEARS = new QName("years");

	private Map<String, CostCalculator> map = new HashMap<>();
	private List<SeasonCalc> seasonCalcs = new ArrayList<>();
	private List<MultiSeasonCalc> multiCalcs = new ArrayList<>();

	/**
	 * Loads the rate tables etc from the specified file and returns the
	 * appropriate CostCalculator.
	 * 
	 * @param file
	 * @return the calculator for the loaded rate tables.
	 * 
	 * @throws IOException
	 * @throws XMLStreamException
	 */
	public CostCalculator load(String file) throws IOException, XMLStreamException {
		map.clear();
		seasonCalcs.clear();
		multiCalcs.clear();

		XMLInputFactory factory = XMLInputFactory.newInstance();
		XMLEventReader reader = null;
		String calcName = null;
		try {
			reader = factory.createXMLEventReader(new BufferedReader(new FileReader(file)));

			while (reader.hasNext()) {
				XMLEvent evt = reader.nextEvent();
				if (evt.isStartElement()) {
					StartElement elm = evt.asStartElement();
					QName elmName = elm.getName();
					if (elmName.equals(RATES)) {
						calcName = getAttribute(elm, USE);
					} else if (elmName.equals(RATE_TABLE)) {
						parseRateTable(reader, elm);
					} else if (elmName.equals(SEASONAL_CALC)) {
						parseSeasonalCalc(reader, elm);
					} else if (elmName.equals(MULTI_SEASONAL_CALC)) {
						parseMultiSeasonalCalc(reader, elm);
					}
				}
			}

		} finally {
			if (reader != null)
				try {
					reader.close();
				} catch (Exception ex) {
				}
		}

		for (SeasonCalc calc : seasonCalcs) {
			calc.createCalculator(map);
		}

		for (MultiSeasonCalc calc : multiCalcs) {
			calc.createCalculator(map);
		}

		CostCalculator calc = map.get(calcName);
		if (calc == null)
			throw new IOException("Invalid rate file. Calculator to use '" + calcName
					+ "' not found.");
		return calc;
	}

	private void parseMultiSeasonalCalc(XMLEventReader reader, StartElement elm)
			throws IOException, XMLStreamException {
		MultiSeasonCalc calc = new MultiSeasonCalc();
		calc.id = getAttribute(elm, ID);
		while (reader.hasNext()) {
			XMLEvent evt = reader.nextEvent();
			if (evt.isEndElement() && evt.asEndElement().getName().equals(MULTI_SEASONAL_CALC))
				break;
			else if (evt.isStartElement() && evt.asStartElement().getName().equals(ENTRY)) {
				StartElement st = evt.asStartElement();
				String seasonId = getAttribute(st, SEASONAL_CALC);
				String months = getAttribute(st, YEARS);
				calc.entries.put(seasonId, months);
			}
		}
		multiCalcs.add(calc);
	}

	private void parseSeasonalCalc(XMLEventReader reader, StartElement elm) throws IOException,
			XMLStreamException {
		SeasonCalc calc = new SeasonCalc();
		calc.id = getAttribute(elm, ID);
		while (reader.hasNext()) {
			XMLEvent evt = reader.nextEvent();
			if (evt.isEndElement() && evt.asEndElement().getName().equals(SEASONAL_CALC))
				break;
			else if (evt.isStartElement() && evt.asStartElement().getName().equals(ENTRY)) {
				StartElement st = evt.asStartElement();
				String rate_id = getAttribute(st, RATE_TABLE);
				String months = getAttribute(st, MONTHS);
				calc.entries.put(rate_id, months);
			}
		}
		seasonCalcs.add(calc);
	}

	private void parseRateTable(XMLEventReader reader, StartElement elm) throws IOException,
			XMLStreamException {
		String type = getAttribute(elm, TYPE);
		String id = getAttribute(elm, ID);
		double flatFee = Double.parseDouble(getAttribute(elm, FLAT_FEE));

		List<Double> thresholds = new ArrayList<Double>();
		List<Double> rates = new ArrayList<Double>();

		while (reader.hasNext()) {
			XMLEvent evt = reader.nextEvent();
			if (evt.isEndElement() && evt.asEndElement().getName().equals(RATE_TABLE))
				break;
			else if (evt.isStartElement() && evt.asStartElement().getName().equals(RATE_BLOCK)) {
				StartElement st = evt.asStartElement();
				String threshold = getAttribute(st, THRESHOLD);
				String rate = getAttribute(st, RATE);
				if (!threshold.equals(MAX)) {
					thresholds.add(Double.valueOf(threshold));
				}
				rates.add(Double.valueOf(rate));
			}
		}

		CostCalculator calc = null;
		if (type.equals(BLOCK)) {
			calc = new BlockRateCostCalculator(createArray(thresholds), createArray(rates), flatFee);
		} else if (type.equals(STEP)) {
			calc = new IncreasingStepRateCostCalculator(createArray(thresholds),
					createArray(rates), flatFee);
		}
		if (calc == null)
			throw new IOException("Invalid rates file. Bad Rate table: " + id);
		map.put(id, calc);
	}

	private double[] createArray(List<Double> vals) {
		double[] array = new double[vals.size()];
		for (int i = 0; i < vals.size(); ++i) {
			array[i] = vals.get(i);
		}
		return array;
	}

	private String getAttribute(StartElement elm, QName name) throws IOException {
		Attribute attrib = elm.getAttributeByName(name);
		if (attrib == null)
			throw new IOException("Invalid rates file. rates element is missing required "
					+ name.getLocalPart() + " attribute");
		return attrib.getValue();
	}

}
