/**
 * 
 */
package cnh.crb.econ.io;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import cnh.crb.econ.SurchargeCalculator;
import cnh.crb.econ.SurchargeCalculator.Surcharge;

/**
 * Reads a surcharge table xml file and creates a SurchargeCalculator from it.
 * 
 * @author nick
 */
public abstract class SurchargeLoader {
	private static final QName ENTRY = new QName("entry");
	private static final QName YEAR = new QName("year");
	private static final QName MONTH = new QName("month");
	private static final QName RATE_1 = new QName("rate_1");
	private static final QName RATE_2 = new QName("rate_2");
	private static final QName THRESHOLD = new QName("rate_2_threshold");
	
	/**
	 * Loads  surcharge table xml file and returns the SurchargeCalculator created
	 * from it.
	 * 
	 * @param file
	 * @return
	 * @throws IOException
	 * @throws XMLStreamException
	 */
	public SurchargeCalculator load(String file) throws IOException, XMLStreamException {

		Calendar cal = GregorianCalendar.getInstance();
		cal.set(Calendar.MILLISECOND, 0);
		cal.set(Calendar.HOUR, 0);
		cal.set(Calendar.SECOND, 0);
		
		XMLInputFactory factory = XMLInputFactory.newInstance();
		XMLEventReader reader = null;
		Map<Date, SurchargeCalculator.Surcharge> map = new HashMap<>();
		
		try {
			reader = factory.createXMLEventReader(new BufferedReader(new FileReader(file)));

			while (reader.hasNext()) {
				XMLEvent evt = reader.nextEvent();
				if (evt.isStartElement()) {
					StartElement elm = evt.asStartElement();
					QName elmName = elm.getName();
					if (elmName.equals(ENTRY)) {
						String sYear = getAttribute(elm, YEAR);
						String sMonth = getAttribute(elm, MONTH);
						String sRate1 = getAttribute(elm, RATE_1);
						String sRate2 = getAttribute(elm, RATE_2);
						String sThresh = getAttribute(elm, THRESHOLD);

						try {
							int year = Integer.parseInt(sYear);
							int month = Integer.parseInt(sMonth);
							cal.set(year, month, 1, 0, 0, 0);
							double rate1 = Double.parseDouble(sRate1);
							double rate2 = Double.parseDouble(sRate2);
							double threshold = Double.parseDouble(sThresh);
							map.put(cal.getTime(), new SurchargeCalculator.Surcharge(rate1, rate2, threshold));
						} catch (NumberFormatException ex) {
							throw new IOException("Invalid attribute. " + ex.getMessage());
						}
					}

				}
			}
		} finally {
			if (reader != null)
				try {
					reader.close();
				} catch (Exception ex) {
				}
		}

		return getSurchargeCalculator(map);
	}

	private String getAttribute(StartElement elm, QName name) throws IOException {
		Attribute attrib = elm.getAttributeByName(name);
		if (attrib == null)
			throw new IOException("Invalid simple rate table. entry element is missing required "
					+ name.getLocalPart() + " attribute");
		return attrib.getValue();
	}
	
	
	/**
	 * Child classes should implement a specific surcharge calculator from this.
	 * 
	 * @param map
	 */
	public abstract SurchargeCalculator getSurchargeCalculator(Map<Date, Surcharge> map);

}
