/**
 * 
 */
package cnh.crb.household;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import org.apache.log4j.Logger;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameters;
import repast.simphony.space.graph.Network;
import repast.simphony.space.graph.RepastEdge;
import repast.simphony.util.ContextUtils;
import simphony.util.messages.MessageCenter;
import cnh.crb.CNHConstants;
import cnh.crb.SeasonalWaterUseInstance;
import cnh.crb.WaterUseAgent;
import cnh.crb.WaterUseCalculator;
import cnh.crb.climate.Climate;
import cnh.crb.climate.ClimateImpacts;
import cnh.crb.econ.Bill;
import cnh.crb.econ.BillableEntity;
import cnh.crb.hydro.DemandRequest;
import cnh.crb.hydro.FlowHydrograph;
import cnh.crb.hydro.FlowManager;
import cnh.crb.hydro.FlowRate;
import cnh.crb.hydro.SourceFlowNode;
import cnh.crb.hydro.Tap;
import cnh.crb.hydro.UnitsAndConverters;
import cnh.crb.hydro.UnitsAndConverters.VolumeUnits;
import cnh.crb.hydro.WaterSink;
import cnh.crb.util.Utilities;

import com.vividsolutions.jts.geom.Point;

/**
 * A Household.
 * 
 * @author NC, Mark A.
 */
public class Household implements WaterUseAgent, BillableEntity {
	
	static MessageCenter msg = MessageCenter.getMessageCenter(Household.class);
	
	public  static final int    INCOME_NA             = -99999999;
	public  static final String UPDATE_SOCIAL_FACTORS = "UPDATE_SOCIAL_FACTORS";
	private static       Logger LOGGER                = Logger.getLogger(Household.class);
	
	public static class HouseholdData {
		// From census data:
		String id;
		int    income;
		int    size;
		int    age;
		Race   race;
		Point  location;

		// For water use calculator:
		double baselineWaterUse;
		
		double socialFactors,   socialWeight;
		double priceFactor,     priceWeight;
		double exogenousEffect, exogenousWeight;
		double lowFlowFixtureWeight;
		
		// For price memory and influence:
		int    priceMemSize;
		double maxPriceFactorIncrease;
		double priceFactorIncreaseScale;
		double priceFactorDecreaseDecay;
		double priceFactorBaseDecay;
		
		boolean useFullWaterFlow;
	}

	// Household instance variables:
	
	// A pointer to the global calendar
	private Calendar cal;
	
	// Census data
	private String  id;
	private Race    race;        // Race of the (head of?) the household
	private int     income;      // Household income 
	private int     size;        // Number of people in household
	private int     age;         // Age of the head of the household
	private Point   location;    // Location of the household
	
	// Household characteristics for water use:
	protected double priceFactor;
	protected double socialFactors;
	
	// Internal water use calculator and climate functions
	protected WaterUseCalculator calc;
	protected Climate            climate;
	protected ClimateImpacts     climateImpacts;
	
	// Networks that influence this one:
	private InfluenceSet    infSet = new InfluenceSet();
	private Network<Object> network;
	
	// Noise factor:
	private double noise = 0;
	
	// For FlowNode:
	protected SourceFlowNode  source;
	protected DemandRequest   expectedInflow;
	protected boolean         ready;
	private   FlowHydrograph  waterUseHydrograph = new FlowHydrograph();

	// Record keeping
	protected Tap      tap;
	private   Date     lastUsageCalculation;
	protected boolean  useFullWaterFlow;
	private   double   waterUsed              = 0; // Last water calculation
	protected double   currentDemand          = 0; // Running total since last Flow
	
	protected WaterSink sink;

	// Calculation of changes in social influence:
	protected double tmpSocialFactors; 
	protected double lastSocialInfluenceBaselineUsage = 0;
	
	// Price memory:
	protected HouseholdWaterBillMemory billMemory;
	public HouseholdWaterBillMemory getBillMemory() {
		return billMemory;
	}
	
	protected double                   lowFlowFixtureReduction = 0;
	
	protected int                      priceMemSize;
	protected double                   maxPriceFactorIncrease;
	protected double                   priceFactorIncreaseScale;
	protected double                   priceFactorDecreaseDecay;
	protected double                   priceFactorBaseDecay;
	protected boolean                  useUnitPrice = false; // Debugging

	public Household(HouseholdData data, Calendar calendar, Climate climate, ClimateImpacts climateImpacts) {
		// From data:
		this.id                       = data.id;
		this.race                     = data.race;
		this.income                   = data.income;
		this.size                     = data.size;
		this.age                      = data.age;
		this.location                 = data.location;
		this.maxPriceFactorIncrease   = data.maxPriceFactorIncrease;
		this.priceFactorIncreaseScale = data.priceFactorIncreaseScale;
		this.priceFactorDecreaseDecay = data.priceFactorDecreaseDecay;
		this.priceFactorBaseDecay     = data.priceFactorBaseDecay;
		this.priceFactor              = data.priceFactor;
		this.priceMemSize             = data.priceMemSize;
		this.socialFactors            = data.socialFactors;
	    this.tmpSocialFactors         = data.socialFactors;
	    this.useFullWaterFlow         = data.useFullWaterFlow;
	    
	    
	    // Initialize time and record keeping:
		cal                  = calendar;
		this.climate         = climate;
		this.climateImpacts  = climateImpacts;
		lastUsageCalculation = cal.getTime();
		calc                 = new HouseholdWaterUseCalculator(data, this);
		tap                  = new Tap(calendar.getTime(), 0, VolumeUnits.GALLONS);
		billMemory           = new HouseholdWaterBillMemory(data.priceMemSize);
		
		// For flow
		ready                = false;
	}
	
	private double calculateInitSurchargeAvg(double baselineWaterUse) {
		Calendar cal = GregorianCalendar.getInstance();
		cal.setTime(this.cal.getTime());
		double usageCCF = UnitsAndConverters.GallonsToCCF(baselineWaterUse); 
		
		cal.set(Calendar.MONTH, Calendar.NOVEMBER);
		double avg = usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		cal.set(Calendar.MONTH, Calendar.DECEMBER);
		avg = usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		cal.set(Calendar.MONTH, Calendar.JANUARY);
		avg += usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		cal.set(Calendar.MONTH, Calendar.FEBRUARY);
		avg += usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		cal.set(Calendar.MONTH, Calendar.MARCH);
		avg += usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		cal.set(Calendar.MONTH, Calendar.APRIL);
		avg += usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		return avg / 6;
	}
	
	private double calculateInitSewerAvg(double baselineWaterUse) {
		Calendar cal = GregorianCalendar.getInstance();
		cal.setTime(this.cal.getTime());
		cal.set(Calendar.MONTH, Calendar.DECEMBER);
		
		double usageCCF = UnitsAndConverters.GallonsToCCF(baselineWaterUse); 
		double avg = usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		cal.set(Calendar.MONTH, Calendar.JANUARY);
		avg += usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		cal.set(Calendar.MONTH, Calendar.FEBRUARY);
		avg += usageCCF * SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactorFor(cal);
		
		return avg / 3;
	}

	
	// Simple Getters:
	
	public double          getWaterUsed(){	             return waterUsed;               }
	public String          getId(){		                 return id;	                     }
	public Race            getRace(){	                 return race;                    }
	public int             getIncome(){                  return income;                  }
	public int             getSize(){                    return size;                    }
	public int             getAge(){        	         return age;                     }
	public Point           getLocation(){  	             return location;                }
	public double          getPriceFactor(){             return priceFactor; 	         }
	public double          getSocialFactors(){	         return socialFactors;           }
	public double          getWaterUsageNoise(){         return noise;	                 }
	public InfluenceSet    getInfluenceSet(){            return infSet;                  }
	public Climate         getClimate(){                 return climate;                 }
	public ClimateImpacts  getClimateImpacts(){          return climateImpacts;          }
	public double          getLowFlowFixtureReduction(){ return lowFlowFixtureReduction; }

	
	// Simple Setters:
	
	public void setWaterUsageNoise(double noise){  noise = noise;  }
	
	
	public void incrementLowFlowFixtureReduction(double reduction){
		lowFlowFixtureReduction = Math.min(lowFlowFixtureReduction + reduction, 1);
	}

	
	
	// Managing the networks of agents:
	
	/**
	 * Adds the specified household to the set of households that influence
	 * this household.
	 * 
	 * @param household
	 */
	public void addToInfluenceSet(Household household) {
		infSet.addHousehold(household);
	}

	/**
	 * Clears this household's influence set.
	 */
	public void clearInfluenceSet() {
		infSet.clear();
	}
	

	
	// Updating Social Factors:
	
	/**
	 * Updates the social factors based on messaging intensity
	 */
	public void updateSocialFactorsWithIntensity(double intensity){
		incrementTmpSocialFactors(intensity);
	}

	/**
	 * Updates the social norms property of this Household based on its
	 * geographic neighbors' social norms.
	 */
	public void updateSocialFactorsFromNgh() {
		incrementTmpSocialFactors(infSet.calculateInfluenceChange(socialFactors)); //
	}
	
	/**
	 * Updates this Household's social factors from its network neighbors.
	 * 
	 * Social influence calculated by looking at social connections and seeing
	 * if that connection has a probability of influencing a given individual.
	 * The calculation looks at differences between the two agents and tries to
	 * make this agent more similar to the agent in the social network.
	 */
	@SuppressWarnings("unchecked")
	@ScheduledMethod(start = 1, interval = 1, priority = CNHConstants.UPDATE_FROM_NET_PRIORITY)
	public void updateSocialFactorsFromNet() {
		Parameters p = RunEnvironment.getInstance().getParameters();

		if (network == null) {
			network = (Network<Object>) ContextUtils.getContext(this).getProjection(
					CNHConstants.HOUSEHOLD_NETWORK_ID);
		}

		if (network != null) {
			double convergenceFactor = p.getDouble(CNHConstants.SOCIAL_FACTOR_CONVERGENCE_NWK);
			double threshold         = p.getDouble(CNHConstants.SOCIAL_FACTOR_THRESHOLD_NWK);
			for (RepastEdge<Object> edge : network.getOutEdges(this)){
				Household alter = (Household) edge.getTarget();
				double diff = alter.getSocialFactors() - getSocialFactors(); // Compare social factors (not tmpSocialFactors)
				if (Math.abs(diff) < threshold)	incrementTmpSocialFactors(convergenceFactor * diff);
			}
		}
	}

	/**
	 * Should be scheduled to occur either with UPDATE_FROM_PRIOR_USAGE_PRIORITY
	 * or UPDATE_FROM_CURRENT_USAGE_PRIORITY based on whether current tick usage
	 * should be included or not.
	 */
	public void updateSocialFactorsFromUsage() {
		double currentUsage = tap.getUsage(UPDATE_SOCIAL_FACTORS, cal.getTime()).getVolumeUsed(VolumeUnits.GALLONS);
		if(lastSocialInfluenceBaselineUsage > 0)  
			incrementTmpSocialFactors((lastSocialInfluenceBaselineUsage - currentUsage) / lastSocialInfluenceBaselineUsage);
		lastSocialInfluenceBaselineUsage = currentUsage;
	}

	/**
	 * Increments and clamps tmpSocialFactors to [-1,1]
	 * 
	 * @param increment
	 */
	protected void incrementTmpSocialFactors(double increment) {
		tmpSocialFactors += increment;
		tmpSocialFactors = Utilities.clamp(tmpSocialFactors, -1, 1);
	}

	
	
	
	// Adjusting Price effects
	
	/**
	 * Adjusts the price effect based on the received bill
	 * vs. the previous bills.
	 * 
	 * If the received bill is higher than the average of the
	 * previous bills, adjusts the price effect upwards by a small
	 * additional amount. The amount is:
	 * 
	 * C/A * (1/S) * Max
	 * 
	 * where C is the current bill A is the average over previous
	 * bills, S is a scaling factor and Max is the maximum adjustment
	 * per unit of the scaling factor.
	 * 
	 * A reasonable value for 'Max' might be 0.3; this means that
	 * the increase in price factor (which ranges from 0 to 1) 
	 * is based on a unit of magnitude of 0.3.  The 'S' parameter is 
	 * essentially the magnitude increase that will lead to this 
	 * unit of impact. If, for example, S is 2, then a full magnitude
	 * increase requires 2xA- if the bill 
	 * is double the average, the increase would be the full 0.3; if
	 * the bill is triple the average, the impact factor increase will be 0.45;
	 * and if quadruple the average, the impact factor increase is .6.
	 * If S is 3, then the the full 0.3 increase is only possible if
	 * the bill is triple the average, and a 0.6 increase would require
	 * a 6x increase above the average.  
	 * @param currentBill
	 */
	private void updatePriceEffect(Bill currentBill) {
		double current = 
				useUnitPrice ? currentBill.getUnitPrice() : currentBill.getBillAmount();
		double average = 
				useUnitPrice ? billMemory.getAverageUnitPrice(priceMemSize) :
					           billMemory.getAverageCost(priceMemSize);
		if (average == 0) return; // This means price factor won't change until enough of a record exists
		LOGGER.info("Bill price      = " + current);
		LOGGER.info("Avg             = " + average);
		LOGGER.info("Max increase    = " + maxPriceFactorIncrease);

        msg.debug("START BILL DATE " + currentBill.getStartDate() + " CURRENT: " + current + " AVERAGE: " + average);

		if (current > average) {
			LOGGER.info("Increase will be: " + ((current - average)/ average) * (1/priceFactorIncreaseScale) * maxPriceFactorIncrease);
			priceFactor += ((current - average)/ average) * (1/priceFactorIncreaseScale) * maxPriceFactorIncrease;
		} else { // Current bill is less than or equal to the average
			priceFactor *= Math.min(
					(1 - Math.pow((1 - current / average), priceFactorDecreaseDecay)),
					priceFactorBaseDecay);
		}

		priceFactor = Utilities.clamp(priceFactor, 0, 1); // Keep between zero and 1
	}
	
	
	
	
    // Calculating Water Use	

	/**
	 * Calculate the amount of water used for the current period.
	 */
	@ScheduledMethod(start = 1, interval = 1, priority = CNHConstants.CALC_USE_PRIORITY)
	public void calculateWaterUse() {
		LOGGER.info("Tick: " + RunEnvironment.getInstance().getCurrentSchedule().getTickCount() 
				+ ", calculating water use " + cal.getTime());
		// Update the social factors with buffered value
		socialFactors = tmpSocialFactors;
		waterUsed = calc.calculateUsage();
		currentDemand += waterUsed;
		if(useFullWaterFlow){
			Calendar c2 = (Calendar)cal.clone();
			c2.add(Calendar.DAY_OF_YEAR, 1);
			// Add this day's water usage to the demand flow hydrograph
			if (c2.getTime().after(lastUsageCalculation)) {
				FlowHydrograph use = new FlowHydrograph(lastUsageCalculation, c2.getTime(), waterUsed,
						FlowRate.FlowRateUnits.GPD);
				this.waterUseHydrograph.add(use);	
			}
			lastUsageCalculation = c2.getTime();
		}
	}

	
	
	// Billing

	/**
	 * Gets the water Usage as provided by a flow since the last reading with 
	 * this TapID. Re-sets the reading associated with this TapID such that subsequent readings will
	 * get the Usage between this reading and the subsequent reading.
	 * 
	 * @param tapID
	 * @param volumeUnits
	 * 
	 * @return the water usage
	 */
	public Tap.Usage getMeteredWaterUse(String tapID) {
		return tap.getUsage(tapID, cal.getTime());
	}

	
	/**
	 * Gets the metered water use during the time between the start of tap 1
	 * and the start of tap 2.
	 * @param tapID1
	 * @param tapID2
	 * @return
	 */
	public Tap.Usage getMeteredWaterUse(String tapID1, String tapID2){
		return tap.getUsage(tapID1,  tapID2, cal.getTime());
	}

	public void dropMeteredWaterUse(String tapID){
		tap.dropAccumulator(tapID);
	}
		
	public void receiveBill(Bill bill) {
		updatePriceEffect(bill);
		billMemory.recordBill(bill);
	}
	
	// For this household's role as a Flow Node
	
	public void addWaterSource(SourceFlowNode src){	
		source = src; 	// Households have only one water source (for now)
	}

	/**
	 * Get the water demand in the form of a Hydrograph for the specified start
	 * date to the specified end date.
	 */
	public FlowHydrograph getDemand(Date start, Date end) {
		return waterUseHydrograph.resetBounds(start, end);
	}

	public boolean prepareFlow(Date start, Date end, FlowManager flowManager){
		if(!useFullWaterFlow) return true;
		if (ready)
			return true;
		if (expectedInflow == null) {
			FlowHydrograph demand = getDemand(start, end);
			source.registerDemandRequest(this, demand, flowManager);
			expectedInflow = new DemandRequest(demand, null);
		}
		ready = true;
		return ready;
	}

	public void clearFlow() {
		currentDemand = 0;
		ready = false;
		if(useFullWaterFlow){
			expectedInflow = null;
			waterUseHydrograph = new FlowHydrograph();
		}
	}

	public boolean isReadyForFlow() {
		return ready;
	}

	public void flow(Date end) {
		double totalInflowInGallons = 
				useFullWaterFlow ? expectedInflow.hydrograph1.getTotalFlowVolume(VolumeUnits.GALLONS) : currentDemand;
		tap.accumulate(totalInflowInGallons, VolumeUnits.GALLONS);
		sink(totalInflowInGallons);
		clearFlow();
	}

	// This is the agent being told what the supply will be
	// For now, it is just going to accept what it gets (!)
	public void registerSupplyToBeReceived(SourceFlowNode upstreamNode, FlowHydrograph supplyToBeReceived,
			FlowHydrograph additionalSupplyAvailable, FlowManager flowManager) {
		if (expectedInflow == null) {
			expectedInflow = new DemandRequest(supplyToBeReceived, additionalSupplyAvailable);
			ready = false;
			flowManager.markNonReady(this);
		} else {
			if (expectedInflow.registerSupply(supplyToBeReceived, additionalSupplyAvailable)) {
				ready = false;
				flowManager.markNonReady(this);
			}
		}
	}
	
	
  // Called for water that is not to be delivered to downstream nodes but rather sent to the sink
  public void sink(double totalInflowInGallons){
	  if(sink != null)  sink.eliminateWater(totalInflowInGallons, VolumeUnits.GALLONS, WaterSink.WATER_SINK_METHOD.RETURN_TO_SURFACE);
  }

  public void setSink(WaterSink sink){
	LOGGER.info("Household " + getId() + " setting sink to " + sink.getId());
    this.sink = sink;
  }
	  
  public WaterSink getSink(){ 
	return sink; 
  }


	
	
	
}
