package cnh.crb.household;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;

import org.apache.commons.collections15.Predicate;
import org.geotools.geometry.jts.JTSFactoryFinder;

import au.com.bytecode.opencsv.CSVReader;
import cnh.crb.climate.Climate;
import cnh.crb.climate.ClimateImpacts;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.GeometryFactory;

/**
 * Creates households from a population csv file.
 * 
 * @author nick
 * 
 */
public class HouseholdLoader {
	
	static {
		System.setProperty("org.geotools.referencing.forceXY", "true");
	}

	private static final int ID_IDX = 0;
	private static final int RACE_IDX = 3;
	private static final int INCOME_IDX = 4;
	private static final int SIZE_IDX = 5;
	private static final int AGE_IDX = 6;
	private static final int LAT_IDX = 7;
	private static final int LON_IDX = 8;

	private CSVReader reader;
	private Household household;
	private Predicate<Household.HouseholdData> filter;
	private HouseholdDataInitializer initializer;
	private Calendar calendar;
	private Climate climate;
	private ClimateImpacts climateImpacts;

	/**
	 * Creates a HouseholdCreator that will read the household data from the
	 * specified file. Only HouseholdData that passes the filter will become a
	 * Household.
	 * 
	 * @param popFile
	 * @param filter
	 * 
	 * @throws IOException
	 *             if the popFile cannot be found or if there is an error
	 *             reading it.
	 */
	public HouseholdLoader(String popFile,
			Predicate<Household.HouseholdData> filter, HouseholdDataInitializer initializer,
			Calendar cal, Climate climate, ClimateImpacts climateImpacts) throws IOException {
		this.reader = new CSVReader(new BufferedReader(new FileReader(popFile)));
		this.filter = filter;
		this.initializer = initializer;
		// skip file header
		reader.readNext();
		calendar = cal;
		this.climate        = climate;
		this.climateImpacts = climateImpacts;
		readHousehold();
	}

	/**
	 * @return true if there are more households remaining to be read, otherwise
	 *         false.
	 */
	public boolean hasNext() {
		return household != null;
	}

	private void readHousehold() throws IOException {
		household = null;
		while (household == null) {
			String[] vals = reader.readNext();
			if (vals == null) break;
			Household.HouseholdData data = new Household.HouseholdData();
			data.id = vals[ID_IDX].trim();
			data.race = Race
					.fromHouseholdCode(Integer.parseInt(vals[RACE_IDX]));
			data.income = Household.INCOME_NA;
			try {
				data.income = Integer.parseInt(vals[INCOME_IDX]);
				// exception when code is not a number
				// which signifies the N/A value
			} catch (NumberFormatException ex) {
			}

			data.size = Integer.parseInt(vals[SIZE_IDX]);
			data.age = Integer.parseInt(vals[AGE_IDX]);
			double lat = Double.parseDouble(vals[LAT_IDX]);
			double lon = Double.parseDouble(vals[LON_IDX]);
			GeometryFactory geometryFactory = JTSFactoryFinder
					.getGeometryFactory();
			data.location = geometryFactory
					.createPoint(new Coordinate(lon, lat));
			if (filter.evaluate(data)) {
				this.initializer.initialize(data);
				household = new Household(data, calendar, climate, climateImpacts);
			}
			
		}
	}

	public Household nextHousehold() throws IOException {
		Household tmp = household;
		readHousehold();
		return tmp;
	}
	
	/**
	 * Closes this HouseholdLoader, closing any resources it has open.
	 * @throws IOException
	 */
	public void close() throws IOException {
		reader.close();
	}
}
