package cnh.crb.household;

import repast.simphony.random.RandomHelper;
import cnh.crb.SeasonalWaterUseInstance;
import cnh.crb.WaterUseCalculator;
import cnh.crb.climate.Climate;
import cnh.crb.climate.ClimateImpacts;
import cnh.crb.util.Utilities;

/**
 * Calculates water usage for a household.
 * 
 */
public class HouseholdWaterUseCalculator implements WaterUseCalculator {

	private   Household       household;
	
	protected double          baselineWaterUse;     // Baseline of water use (does not change)

	protected double          exogenousEffects;     // Influence of exogenous effects
	
	protected double          socialWeight;         // Weight of social norms/perceptions
	protected double          priceWeight;          // Weight of price
	protected double          exogenousWeight;      // weight of exogenous events
	
	protected double          lowFlowFixtureWeight;
		
	/**
	 * Default constructor
	 */
	public HouseholdWaterUseCalculator(Household.HouseholdData data,
			Household household) {
		
		baselineWaterUse     = data.baselineWaterUse; 
		priceWeight          = data.priceWeight;
		exogenousWeight      = data.exogenousWeight;
		socialWeight         = data.socialWeight;
		lowFlowFixtureWeight = data.lowFlowFixtureWeight; 
		
		// Exogenous effects
		exogenousEffects  = data.exogenousEffect;
		
		// Pointer to household
		this.household    = household;
	}

	/**
	 * Water use calculated by price, exogenous events, and social
	 * norms/perceptions. These are all controlled by weights that regulate how
	 * much each value influences water consumption.
	 */
	public double calculateUsage() {
		double socialNudge    = household.getSocialFactors();
		double priceNudge     = household.getPriceFactor();
		double exogenousNudge = Utilities.clamp(RandomHelper.nextDoubleFromTo(0.0, exogenousEffects),0, 1.0);
		double nudgeFactor =
				socialNudge    * socialWeight + 
				priceNudge     * priceWeight + 
				exogenousNudge * exogenousWeight;
	
		// If the nudgeFactor is greater than 1 we cap it at 1.
		// This is only concerned with the positive nudgeFactors
		// since they apply downward pressure for water consumption
		// and we don't want negative water consumption
		if (nudgeFactor > 1) nudgeFactor = 1;

		Climate climate               = household.getClimate();
		ClimateImpacts climateImpacts = household.getClimateImpacts();

		double climateScale  =
			(1 + 
			    ( climateImpacts.getFractionChangePerDegreeC()             * (climate.getAnnualTemperature()    - climate.getBaselineAnnualTemperature()) - 
				  climateImpacts.getFractionDecreasePer10mmPrecipitation() * ((climate.getAnnualPrecipitation() - climate.getBaselineAnnualPrecipitation())/10d)));
				
		double seasonalScale = SeasonalWaterUseInstance.getWaterUse().getMonthlyScaleFactor();
				
		double waterUsed = baselineWaterUse * seasonalScale * climateScale * (1 - nudgeFactor);
		waterUsed *= (1 + household.getWaterUsageNoise()) * (1 - (household.getLowFlowFixtureReduction() * lowFlowFixtureWeight));
		return waterUsed;
	}

}
