/**
 * 
 */
package cnh.crb.household;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.GeodeticCalculator;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.operation.TransformException;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.index.ItemVisitor;
import com.vividsolutions.jts.index.SpatialIndex;
import com.vividsolutions.jts.index.strtree.STRtree;

/**
 * Creates influence sets for Households.
 * 
 * @author Nick Collier
 */
public class InfluenceSetCreator {

	private static class WithinVisitor implements ItemVisitor {

		private Household source;
		private double distance;
		
		public WithinVisitor(double distance) {
			this.distance = distance;
		}

		public void reset(Household source) {
			this.source = source;
		}

		@Override
		public void visitItem(Object obj) {
			if (!source.equals(obj)) {
				Household target = ((Household) obj);
				try {
					double distance = JTS.orthodromicDistance(source
							.getLocation().getCoordinate(), target.getLocation().getCoordinate(),
							DefaultGeographicCRS.WGS84);
					if (distance <= this.distance) {
						source.addToInfluenceSet(target);
					}
				} catch (TransformException e) {
					// this should never occur given the CRS
					e.printStackTrace();
				}
			}
		}
	}

	private SpatialIndex index = new STRtree();
	private List<Household> households = new ArrayList<>();

	/**
	 * Adds the specified household to the list of households to create
	 * influence sets for.
	 * 
	 * @param household
	 */
	public void addHousehold(Household household) {
		households.add(household);
		index.insert(
				new Envelope(household.getLocation().getEnvelopeInternal()),
				household);
	}
	
	private Envelope calcEnvelope(Household h, double distance) {
		GeodeticCalculator calc = new GeodeticCalculator(DefaultGeographicCRS.WGS84);
	    // mind, this is lon/lat
		Coordinate coord = h.getLocation().getCoordinate();
		//System.out.println("coord: " + coord);
	    calc.setStartingGeographicPoint(coord.x, coord.y);
	    calc.setDirection(0, distance);
	    Point2D north = calc.getDestinationGeographicPoint();
	    //System.out.println("north: " + north);
	    calc.setDirection(90, distance);
	    Point2D east = calc.getDestinationGeographicPoint();
	    //System.out.println("east: " + east);
	    calc.setDirection(180, distance);
	    Point2D south = calc.getDestinationGeographicPoint();
	    //System.out.println("south: " + south);
	    calc.setDirection(-90, distance);
	    Point2D west = calc.getDestinationGeographicPoint();
	    //System.out.println("west: " + west);
	    Envelope env = new Envelope(west.getX(), east.getX(), south.getY(), north.getY());
	    //System.out.println(env);
	    		
	    return env;
	}

	/**
	 * Creates an influence set for each household that has been add to this
	 * InfluenceSetCreator. For each household, each houshold that is within the
	 * specified distance of that household is added to the influence set for
	 * that houshold.
	 * 
	 * @param distance
	 *            the distance in meters
	 * 
	 */
	public void run(double distance) {
		WithinVisitor visitor = new WithinVisitor(distance);
		for (Household h : households) {
			Envelope env = calcEnvelope(h, distance);
			visitor.reset(h);
			index.query(env, visitor);
		}
	}
}
