/**
 * 
 */
package cnh.crb.household;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import repast.simphony.context.space.graph.NetworkGenerator;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.Network;
import repast.simphony.space.graph.RepastEdge;
import repast.simphony.util.SimUtilities;
import cern.jet.random.Gamma;

/**
 * 
 * Buils KE type networks. See Klemm and Eguiluz,
 * "Growing scale-free network with small world behavior" in Phys. Rev. E 65.
 * Then sets edges weights to between 0 - 1 following a gamma distribution.
 * 
 * @author NC
 */
public class KENetworkGenerator implements NetworkGenerator<Object> {

	private static class Range {

		double lowerBound, upperBound;
		int index;

		public Range(int index, double lowerBound, double upperBound) {
			this.lowerBound = lowerBound;
			this.upperBound = upperBound;
			this.index = index;
		}

		public boolean inRange(double val) {
			return lowerBound <= val && val < upperBound;
		}
	}

	private int m;
	private int edgeCount = 0;
	private double alpha, lambda;

	public KENetworkGenerator(int m, double edgeWeightAlpha, double edgeWeightLambda) {
		this.m = m;
		alpha = edgeWeightAlpha;
		lambda = edgeWeightLambda;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * repast.simphony.context.space.graph.NetworkGenerator#createNetwork(repast
	 * .simphony.space.graph.Network)
	 */
	@Override
	public Network<Object> createNetwork(Network<Object> network) {
		edgeCount = 0;
		createEdges(network);
		setEdgeWeights(network);
		return network;
	}
	
	private void setEdgeWeights(Network<Object> network) {
		double[] weights = new double[edgeCount];
		Gamma rh = RandomHelper.createGamma(alpha, lambda);
		double max = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < edgeCount; ++i) {
			double val = rh.nextDouble();
			weights[i] = val;
			max = Math.max(max, val);
		}
		
		int i = 0;
		for (Object source : network.getNodes()) {
			for (RepastEdge<Object> edge : network.getOutEdges(source)) {
				edge.setWeight(weights[i] / max);
				++i;
			}
		}
		
	}

	private void createEdges(Network<Object> network) {
		// put nodes in list -- easier to work with
		List<Object> nodes = new ArrayList<Object>();
		for (Object obj : network.getNodes()) {
			nodes.add(obj);
		}

		SimUtilities.shuffle(nodes, RandomHelper.getUniform());
		// create fully connected graph of first m nodes
		// and put those nodes in the "active nodes" list
		// linked list for better removal speed
		List<Object> activeNodes = new LinkedList<Object>();
		for (int i = 0; i < m; ++i) {
			Object source = nodes.get(i);
			activeNodes.add(source);
			for (int j = 0; j < m; ++j) {
				if (i != j) {
					network.addEdge(source, nodes.get(j));
					++edgeCount;
				}
			}
		}

		// create edges for the remaining nodes
		for (int i = m, n = nodes.size(); i < n; ++i) {
			Object source = nodes.get(i);
			double sum = 0;

			// make an edge between source and all the "active nodes"
			for (Object target : activeNodes) {
				network.addEdge(source, target);
				++edgeCount;
				sum += 1.0 / network.getInDegree(target);
			}

			// calculate probabilities of removal for each active node
			// prob is inversely proportional to its current degree
			List<Range> ranges = new ArrayList<>();
			double lowerBound = 0;
			int index = 0;
			for (Object node : activeNodes) {
				double upperBound = lowerBound + (1.0 / network.getInDegree(node) / sum);
				ranges.add(new Range(index, lowerBound, upperBound));
				lowerBound = upperBound;
				++index;
			}

			// remove a node based on its probabilty of removal
			double p = RandomHelper.nextDouble();
			for (Range range : ranges) {
				if (range.inRange(p)) {
					activeNodes.remove(range.index);
					break;
				}
			}

			// add the current node to the activeNodes list
			activeNodes.add(source);
		}

	}
}
