package cnh.crb.hydro;

import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;

import simphony.util.messages.MessageCenter;
import cnh.crb.hydro.FlowRate.FlowRateUnits;

/**
 * A FlowHydrograph is a time-series record of rates of flow.
 * 
 * This implementation allows discontinuous rates of change by
 * storing data as:
 * 
 * Start Time      End Time       Start Flow Rate     End Flow Rate
 * t0              t1             xxx m^3/sec         yyy m^3/sec
 * t1              t2             zzz m^3/sec         aaa m^3/sec
 * etc.
 * 
 * @author murphy
 *
 */
public class FlowHydrograph {
	
	public static long countOfInstances = 0;
	
	static MessageCenter msg = MessageCenter.getMessageCenter(FlowHydrograph.class);
  
	// Inner classes
	public class EmptyFlowInterval extends FlowInterval{
	  public EmptyFlowInterval(long start, long end){
	    super(start, end, new FlowRateM3S(0), new FlowRateM3S(0));
	  }
	}
	
	// Instance variables
	protected Vector<FlowInterval> flowIntervals  = new Vector<FlowInterval>();
	
	// Constructors
	public FlowHydrograph(){ 
	  countOfInstances++;
	}
	
	public FlowHydrograph(Date start, Date end, double constantRate, FlowRateUnits units) throws IllegalArgumentException{
	  countOfInstances++;
	  if(end.getTime() <= start.getTime()) throw new IllegalArgumentException();
	  
	  addFlow(new FlowInterval(start, end, FlowRate.create(constantRate, units), FlowRate.create(constantRate, units)));
	}
	
	public FlowHydrograph(FlowHydrograph other){
	  countOfInstances++;
	  this.add(other);
	}
	
	public void finalize(){
		countOfInstances--;
	}
	
	// Methods
	public void addFlow(FlowInterval flowInt){
	  FlowInterval toAdd = new FlowInterval(flowInt);
	  
	  // If there are no other flowInts, just add this one
	  if(flowIntervals.size() == 0){ 
	    flowIntervals.add(toAdd);
	    return;
	  }
	  
	  // If the one to be added precedes the first one in the list:
      if(toAdd.startSecond < flowIntervals.firstElement().startSecond)
          flowIntervals.add(0, new EmptyFlowInterval(toAdd.startSecond, flowIntervals.firstElement().startSecond));

      // If the one to be added goes beyond the last one in the list:
      if(toAdd.endSecond > flowIntervals.lastElement().endSecond)
          flowIntervals.add(new EmptyFlowInterval(flowIntervals.lastElement().endSecond, toAdd.endSecond));
        
	  // Loop to the first flow interval in the flows collection that ends after the
	  // one to be added begins
      int i = 0;
	  while(i < flowIntervals.size() && flowIntervals.get(i).endSecond <= toAdd.startSecond) i++;
	  
	  // If the first one in the list starts before the one to be added:
	  // Split it into two parts:
	  FlowInterval first = flowIntervals.get(i);
	  if(first.startSecond < toAdd.startSecond){
	    flowIntervals.remove(i);
	    FlowInterval[] split = first.split(toAdd.startSecond);
	    flowIntervals.add(i, split[0]);
	    i++; // Move the index
	    flowIntervals.add(i, split[1]);
	  }
	
      // Now loop through the rest of the list
	  // From here on we assume that the start time for the indexed element
	  // in the extant list is the same as the start time for the element
	  // to be added
	  while(toAdd != null && i < flowIntervals.size()){
	    first = flowIntervals.get(i);
	  
	    // If the element to be added is longer than the indexed element
	    if(first.endSecond < toAdd.endSecond){
	      // Split the element to be added
	      FlowInterval[] split = toAdd.split(first.endSecond);
	      toAdd = split[0];
	      flowIntervals.remove(i);
	      flowIntervals.add(i, new FlowInterval(first.startSecond, first.endSecond,
	          first.startFlowRate.getSum(toAdd.startFlowRate), first.endFlowRate.getSum(toAdd.endFlowRate)));
	      toAdd = split[1];
	      i++;
	    }
	    else{ // Element to be added ends before or at the same time as the indexed element
	      FlowInterval[] split = first.split(toAdd.endSecond);
	      flowIntervals.remove(i);
	      flowIntervals.add(i, new FlowInterval(toAdd.startSecond, toAdd.endSecond,
	          first.startFlowRate.getSum(toAdd.startFlowRate), toAdd.endFlowRate.getSum(split[0].endFlowRate)));
	      i++;
	      if(split[1] != null) flowIntervals.add(i, split[1]); // Null = they both end at same time
	      toAdd = null;
	    }    
	  }
      condense();		
    }
	
	// If two adjacent flow intervals are actually
	// colinear, remove and combine into one
	public void condense(){
	  int i = 1;
	  while(i < flowIntervals.size()){
	    FlowInterval merged = flowIntervals.get(i-1).merge(flowIntervals.get(i));
	    if(merged != null){
	      i--;
	      flowIntervals.remove(i);
	      flowIntervals.remove(i);
	      flowIntervals.add(i, merged);
	    }
	    i++;
	  }
	}
	
	
	public void add(FlowHydrograph other){
	  if(other == null) return;
    for(FlowInterval fi : other.flowIntervals) addFlow(fi);
	}
	
	public void subtract(FlowHydrograph other){
	  if(other == null) return;
	  for(FlowInterval fi: other.flowIntervals) 
	      addFlow(new FlowInterval(fi.startSecond, fi.endSecond, fi.startFlowRate.getProduct(-1), fi.endFlowRate.getProduct(-1)));
	}

	public FlowHydrograph getSum(FlowHydrograph other){
	  FlowHydrograph ret = new FlowHydrograph(this);
	  if(other != null) ret.add(other);
	  return ret;
	}
	
	public FlowHydrograph getDiff(FlowHydrograph other){
	  FlowHydrograph ret = new FlowHydrograph(this);
    if(other != null) ret.subtract(other);
    return ret;
	}

	public FlowHydrograph getProduct(double scalar){
	  FlowHydrograph ret = new FlowHydrograph();
	  for(FlowInterval fi: flowIntervals) {
      ret.addFlow(new FlowInterval(fi.startSecond, fi.endSecond, fi.startFlowRate.getProduct(scalar), fi.endFlowRate.getProduct(scalar)));
	  }
	  return ret;
	}
	
	public FlowHydrograph getInverse(){ return getProduct(-1); }
	
	public FlowRate getFlowRate(Date t){ return getFlowRate(t.getTime()); }
	
	public FlowRate getFlowRate(long t){
	  if((t <  flowIntervals.firstElement().startSecond) || (t >= flowIntervals.lastElement().endSecond))  return null; // Range check- quick
	  for(FlowInterval fi: flowIntervals) if(fi.includes(t)) return fi.getFlowAt(t);
	  return null; // Cannot happen but Java requires this
	}
	
	public Date getDateAtWhichFirstGoesNegative(){
	  for(FlowInterval fi: flowIntervals){
	    Date d = fi.dateCrossFromPositiveToNegative();
	    if(d != null) return d;
	  }
	  return null;
	}
	
	public boolean isEmpty(){ return flowIntervals.size() == 0; }
	
	
	
	
	/**
	 * Returns this FlowHydrograph divided into two components;
	 * the first includes all periods of positive flow, the second
	 * all periods of negative flow.
	 * @return
	 */
	public FlowHydrograph[] getSignedHydrograph(){
    FlowHydrograph[] ret = new FlowHydrograph[2];
    ret[0] = new FlowHydrograph();
    ret[1] = new FlowHydrograph();
    for(FlowInterval interval: flowIntervals){
      FlowInterval[] intervals = interval.splitAtZero();
      ret[ (intervals[0].includesNegativeRates() ? 1 : 0) ].addFlow(intervals[0]);
      if(intervals[1] != null) 
          ret[ (intervals[1].includesNegativeRates() ? 1 : 0) ].addFlow(intervals[1]);
    }
    return ret;
	}
	
	
	
	public boolean equals(FlowHydrograph other){
	  if(other == null) return false;
	  if(other.flowIntervals.size()   != flowIntervals.size()) return false;
	  for(int i = 0; i < flowIntervals.size(); i++){
	    if(!flowIntervals.get(i).equals(other.flowIntervals.get(i))) return false;
	  }
	  return true;
	}
	
	public FlowHydrograph resetBounds(Date t1, Date t2){ return resetBounds(t1.getTime(), t2.getTime()); }
	
	public FlowHydrograph resetBounds(long t1, long t2){
	  FlowHydrograph ret = new FlowHydrograph();
	  ret.addFlow(new EmptyFlowInterval(t1, t2));
	  for(FlowInterval fi: flowIntervals){
	    if(fi.startSecond < t1 && fi.endSecond > t1){ ret.addFlow(fi.split(t1)[1].split(t2)[0]);  } // Add appropriate portion
	    else if(fi.startSecond >= t1 && fi.startSecond < t2){ ret.addFlow(fi.split(t2)[0]); }
	  }
	  return ret;
	}
	
	public FlowHydrograph dropDatesBefore(Date t1){ return dropDatesBefore(t1.getTime()); }
	
	public FlowHydrograph dropDatesBefore(long t1){ return resetBounds(t1, flowIntervals.lastElement().endSecond ); }
		
	/**
	 * Returns true iff this hydrograph includes a value for
	 * the entire time period covered by the other hydrograph
	 * and the flow rate in this hydrograph is greater than
	 * or equal to the flow rate in the other hydrograph at every point.
	 * @param other
	 * @return
	 */
	public boolean isGreaterThan(FlowHydrograph other){
	  // If other hydrograph starts before or ends after this one, must be false
	  if((flowIntervals.firstElement().startSecond > other.flowIntervals.firstElement().startSecond) ||
	     (flowIntervals.lastElement().endSecond < other.flowIntervals.lastElement().endSecond)) return false;
	  // Must collect a list of transition points from each hydrograph and test at each
	  Set<Date> dates = new HashSet<Date>();
	  Date earliest = other.flowIntervals.firstElement().startTime;
	  Date last     = other.flowIntervals.lastElement().endTime;
	  Date d;
	  for(FlowInterval fi: flowIntervals){
	    d = fi.startTime;
	    if( (d.getTime() >= earliest.getTime()) && (d.getTime() < last.getTime()) ) dates.add(d);
	  }
	  d = flowIntervals.lastElement().endTime;
    if( (d.getTime() >= earliest.getTime()) && (d.getTime() < last.getTime()) ) dates.add(d);
  
    for(FlowInterval fi: other.flowIntervals){
      d = fi.startTime;
      if( (d.getTime() >= earliest.getTime()) && (d.getTime() < last.getTime()) ) dates.add(d);
    }
    d = other.flowIntervals.lastElement().endTime;
    if( (d.getTime() >= earliest.getTime()) && (d.getTime() < last.getTime()) ) dates.add(d);
  
	  for(Date dateAt: dates) if(other.getFlowRate(dateAt).compare(this.getFlowRate(dateAt)) == 1) return false;
	  return true;
	  
	}
	
	public boolean includesNegativeValues(){
	  for(FlowInterval fi: flowIntervals) if(fi.includesNegativeRates()) return true; // Could be memoized on add flow...
	  return false;
	}

	
	public String report(FlowRateUnits units){
    String lf = System.getProperties().getProperty("line.separator");
	  StringWriter ret = new StringWriter();
	  for(FlowInterval flowInt: flowIntervals) ret.write(flowInt.toString(units) + lf);
	  return ret.toString();
	}
	
	public double getTotalFlowVolume(UnitsAndConverters.VolumeUnits volumeUnits){
	  double ret = 0;
	  for(FlowInterval flowInt: flowIntervals) ret += flowInt.getTotalVolume(volumeUnits);
	  return ret;
	}
	
	public static void main(String[] args){
	  try{
  	  FlowHydrograph H = new FlowHydrograph();
  	  SimpleDateFormat format = new SimpleDateFormat("yyyy.MM.dd HH:mm:ss");
  	  FlowInterval fi1 = new FlowInterval(format.parse("2013.09.26 10:00:00"), 
  	                                      format.parse("2013.09.26 11:00:00"), 
  	                                      new FlowRateM3S(10), 
  	                                      new FlowRateM3S(10));
  	  FlowInterval fi2 = new FlowInterval(format.parse("2013.09.26 13:40:00"), 
                                          format.parse("2013.09.26 14:15:00"), 
                                          new FlowRateM3S(10), 
                                          new FlowRateM3S(10));

  	  
  	  H.addFlow(fi1);
  	  H.addFlow(fi2);
  	  msg.debug(H.report(FlowRateUnits.M3S));
  	  FlowHydrograph r = H.resetBounds(format.parse("2013.09.26 10:30:00"), format.parse("2013.09.26 14:00:00"));
  	msg.debug("");
  	msg.debug(r.report(FlowRateUnits.M3S));
	  }
	  catch(Exception E){
	    E.printStackTrace(System.out);
	  }
	  msg.debug("DONE.");
	}
	
}
