package cnh.crb.hydro;

import java.text.SimpleDateFormat;
import java.util.Date;

import cnh.crb.hydro.FlowRate.FlowRateUnits;

public class FlowInterval{
  
  private static double ALLOWABLE_ERROR = .0000005;
  

  protected Date      startTime;
  protected Date      endTime;
  protected long      startSecond;
  protected long      endSecond;
  protected double    durationInMilliSeconds; // Memoized for performance
  protected FlowRate  startFlowRate;
  protected FlowRate  endFlowRate;
      
  public FlowInterval(Date start, Date end, FlowRate initialRate, FlowRate finalRate){
    startTime              = start;
    endTime                = end;
    startSecond            = startTime.getTime();
    endSecond              = endTime.getTime();
    durationInMilliSeconds = endSecond - startSecond; // Should throw exception if duration <= 0
    startFlowRate          = initialRate.copy();
    endFlowRate            = finalRate.copy();
  }
  
  public FlowInterval(long start, long end, FlowRate initialRate, FlowRate finalRate){
    startTime              = new Date(start);
    endTime                = new Date(end);
    startSecond            = start;
    endSecond              = end;
    durationInMilliSeconds = endSecond - startSecond; // Should throw exception if duration <= 0
    startFlowRate          = initialRate.copy();
    endFlowRate            = finalRate.copy();
  }
  
  public FlowInterval(FlowInterval fi){
    startTime              = fi.startTime;
    endTime                = fi.endTime;
    startSecond            = fi.startSecond;
    endSecond              = fi.endSecond;
    durationInMilliSeconds = fi.durationInMilliSeconds;
    startFlowRate          = fi.startFlowRate.copy();
    endFlowRate            = fi.endFlowRate.copy();
  }
  
  /**
   * If date passed is outside range, will extrapolate the linear trend,
   * even if negative
   * @param t
   * @return
   */
  public FlowRate getFlowAt(Date t){ return getFlowAt(t.getTime());    }
  
  public FlowRate getFlowAt(long t){
    return startFlowRate.getSum(endFlowRate.getDiff(startFlowRate).getProduct((double)(t - startSecond) / durationInMilliSeconds));
  }
  
  
  
  public FlowRate getAverageFlowRate(){
    return startFlowRate.getSum(endFlowRate).getProduct(.5);
  }
  
  public double getTotalVolume(UnitsAndConverters.VolumeUnits volumeUnits){
	return UnitsAndConverters.convertVolume(
    		getAverageFlowRate().getRateM3S() * durationInMilliSeconds / 1000, 
    		UnitsAndConverters.VolumeUnits.M3, volumeUnits);
  }
  
  public boolean includes(Date t){ return includes(t.getTime()); }        
  
  public boolean includes(long t){
    return ((t >= startSecond) && (t < endSecond));
  }
  
  public boolean includesNegativeRates(){
    return ((startFlowRate.getRate()) < 0 || (endFlowRate.getRate() < 0));
  }
  
  public Date dateCrossFromPositiveToNegative(){
    if(startFlowRate.getRate() < 0)  return startTime; // Units of rate don't matter
    if(endFlowRate.getRate() >= 0)   return null;      // If the startTime was zero or positive, then if end time is also >=0, does not cross
    if(startFlowRate.getRate() == 0) return startTime; // Know that endFlowRate is negative
    // Must interpolate
    return new Date(
        startSecond + 
            (long)((endSecond - startSecond) * (startFlowRate.getRate() / (startFlowRate.getRate() - endFlowRate.getRate()))));
  }
  
  public FlowInterval[] split(Date t){ return split(t.getTime());    }
  
  public FlowInterval[] split(long second){
    FlowInterval[] ret = new FlowInterval[2];
    if(second >= endSecond){
      ret[0] = this;
      ret[1] = null;
      return ret;
    }
    FlowRate splitRate = getFlowAt(second);
    ret[0] = new FlowInterval(startSecond, second, startFlowRate, splitRate);
    ret[1] = new FlowInterval(second, endSecond, splitRate, endFlowRate);
    return ret;
  }
  
  public FlowInterval[] splitAtZero(){
    FlowInterval[] ret = new FlowInterval[2];
    if( ( (startFlowRate.getRate() >= 0) && (endFlowRate.getRate() >= 0) ) ||
        ( (startFlowRate.getRate() <= 0) && (endFlowRate.getRate() <= 0) )){
      ret[0] = this;
      ret[1] = null;
      return ret;
    }
    return split(
        startSecond + 
          (long)((endSecond - startSecond) * 
              (Math.abs(startFlowRate.getRateM3S()) / (Math.abs(endFlowRate.getRateM3S()) + Math.abs(startFlowRate.getRateM3S())))));
  }
  
  // Returns null if they are not colinear
  public FlowInterval merge(FlowInterval other){
    if((this.endSecond != other.startSecond) || 
       (Math.abs(this.getSlope() - other.getSlope()) > ALLOWABLE_ERROR)) return null;
    FlowInterval ret = new FlowInterval(this.startSecond, other.endSecond, this.startFlowRate, other.endFlowRate);
//    if((Math.abs(ret.getFlowAt(this.endSecond).getRateM3S()    - this.endFlowRate.getRateM3S())    > ALLOWABLE_ERROR)  ||
//       (Math.abs(ret.getFlowAt(other.startSecond).getRateM3S() - other.startFlowRate.getRateM3S()) > ALLOWABLE_ERROR)) return null;
    return ret;
  }
  
  public boolean equals(FlowInterval other){
//    System.out.println("FLOW INTERVAL EQUALS:");
//    System.out.println("This Start Second: " + this.startSecond + " other: " + other.startSecond + " " + (this.startSecond  == other.startSecond));
//    System.out.println("This End   Second: " + this.endSecond   + " other: " + other.endSecond   + " " + (this.endSecond    == other.endSecond));
//    System.out.println("This startFlow:    " + this.startFlowRate.getRate(FlowRateUnits.GPD) + " other: " + other.startFlowRate.getRate(FlowRateUnits.GPD) + " " + (this.startFlowRate.compare(other.startFlowRate)));
//    System.out.println("This endFlow:      " + this.endFlowRate.getRate(FlowRateUnits.GPD) + " other: " + other.endFlowRate.getRate(FlowRateUnits.GPD) + " " + (this.endFlowRate.compare(other.endFlowRate)));
    
    return this.startSecond  == other.startSecond               &&
           this.endSecond    == other.endSecond                 &&
           this.startFlowRate.compare(other.startFlowRate) == 0 &&
           this.endFlowRate.compare(other.endFlowRate)     == 0;
  }
  
  private double getSlope(){
	  return startSecond == endSecond ? Double.NaN :
		  ((endFlowRate.getRateM3S() - startFlowRate.getRateM3S()) / (endSecond - startSecond));
  }
  
  public String toString(FlowRateUnits units){
    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy.MM.dd HH:mm:ss");
    return "START: " + dateFormat.format(startTime) + "  " + startFlowRate.getString(units) + "            " + 
            "END: "  + dateFormat.format(endTime)   + "  " + endFlowRate.getString(units); // TODO Add times
  }
  
}
