package cnh.crb.hydro;

import java.util.Date;

import cnh.crb.hydro.FlowRate.FlowRateUnits;
import repast.simphony.engine.environment.RunEnvironment;


/**
 * A NaturalWaterSource is a water source that
 * is simply set from outside the simulation dynamics.
 * Other classes can interrogate it but not make decisions
 * about it- the water available is simply available
 * by fiat. (For example, rainfall.)
 * @author murphy
 *
 */
public class NaturalWaterSource implements SourceFlowNode {

  private String id;
	
  // Natural sources have a single hydrograph set at initialization
  protected FlowHydrograph hydrograph;
  
  // For acting as a FlowNode
  protected RecipientFlowNode  recipient;        // Recipient of flow
  protected SupplyRequest      demands;          // Amount requested/sent
  protected boolean            ready;            // Ready to flow
  protected boolean            useFullWaterFlow; 
  
  public NaturalWaterSource(String sourceId, FlowHydrograph flowHydrograph){
	id = sourceId;
    hydrograph = flowHydrograph;
    useFullWaterFlow = RunEnvironment.getInstance().getParameters().getBoolean("useFullWaterFlow");
  }
  
  public String getId(){ return id; }
  
  public void addWaterRecipient(RecipientFlowNode recipientNode){  
    recipient = recipientNode;
  }

  public boolean isReadyForFlow(){    
    return ready;   
  }

  public boolean prepareFlow(Date start, Date end, FlowManager flowManager){
	if(!useFullWaterFlow) return true;
    if(ready) return true;
    
    // Are all downstream recipients ready?
    if(!recipient.isReadyForFlow()) return false;
    
    // If downstream recipient is ready it means this node has
    // received its demand
    // Allocate water as available to the manager
    FlowHydrograph availableToFlow = hydrograph.resetBounds(start, end);
    FlowHydrograph surplus = demands.allocateSupply(availableToFlow);
    recipient.registerSupplyToBeReceived(this, demands.getAmountToBeSent(), surplus, flowManager);  // Amount still available to send
    ready = true;
    return ready;   
  }

  public void clearFlow() { 
    demands = null;
    ready = false;
  }

  public void flow(Date end) {  
    // Natural sources have no internal state and do not need
    // to be updated when 'flow' actually occurs
    clearFlow();
  }

  public void registerDemandRequest(RecipientFlowNode downstreamNode, FlowHydrograph demand, FlowManager flowManager) {  
    if(!downstreamNode.equals(recipient)) return; // ERROR!
    demands = new SupplyRequest(demand);
    ready = false;
    flowManager.markNonReady(this);
  }

  /**
   * Call this to update the natural water source
   */
  public void update(){
	  
  }
  
}
