package cnh.crb.hydro;

import cnh.crb.hydro.FlowRate.FlowRateUnits;


/**
 * A SupplyRequest is a WaterRequest used by a SourceFlowNode
 * to track the amounts that it will supply.
 * 
 * For a SupplyRequest:
 * 
 *   Hydrograph1 = The amount to be sent
 *   Hydrograph2 = The amount that the recipient has requested
 *       that cannot be met ('unmet')
 * 
 * @author murphy
 *
 */
public class SupplyRequest extends WaterRequest {

  /**
   * When this constructor is called, it assumes that
   * what is being created is a record of demand that
   * has not yet been met. Hence, it places all of the
   * water indicated by the hydrograph in the 'unmet'
   * hydrograph2 slot.
   * @param newDemand
   */
  public SupplyRequest(FlowHydrograph newDemand){
    super(null, newDemand);
  }
  
  public SupplyRequest(FlowHydrograph hydro1, FlowHydrograph hydro2) {
    super(hydro1, hydro2);
  }
  
  /**
   * Registers the demand received by this SourceFlowNode
   * requested by the downstream recipient. If there has
   * been no demand registered, places all of this in
   * 'unmet'; returns true if this is a change from any 
   * extant value, false if it is not a change.
   * @param newDemand
   * @return
   */
  public boolean registerDemand(FlowHydrograph newDemand){
    // If recipient is asking for exactly what is being sent,
    // then just send it; however, 'unmet demand' can now
    // be considered zero. This, however, is not an actionable
    // change- no recalculation is needed.
    if(hydrograph1 != null && hydrograph1.equals(newDemand)){
      hydrograph2 = null;
      return false;
    }
    // If the demand sent is exactly the same as before and there
    // (still) has been no supply added, this is also not a change
    if((hydrograph1 == null || hydrograph1.isEmpty()) && hydrograph2 != null && hydrograph2.equals(newDemand)){
      return false;
    }
    // Otherwise:
    // New amount to be sent will be equal to as much
    // of the old send amount that is contained in demand
    FlowHydrograph overage   = hydrograph1.getDiff(newDemand);
    FlowHydrograph posNeg[] = overage.getSignedHydrograph();
    FlowHydrograph surplus   = posNeg[0];
    FlowHydrograph shortage  = posNeg[1];
    hydrograph1 = hydrograph1.getDiff(surplus); // New demand
    hydrograph2 = shortage.getInverse();        // New unmet demand
    return true;
  }
  
  public FlowHydrograph getAmountToBeSent(){   return hydrograph1;  }
  
  /**
   * Given an amount available, modify this as if the amount
   * available was going to be distributed to the recipient
   * under the assumption that hydrograph1 is the amount
   * that is already going to be sent (and presumably expected)
   * and hydrograph2 is the amount of unmet demand. The
   * result will be new hydrographs for amount to be sent (hydrograph1)
   * and unmet demand (hydrograph2); any surplus will be returned.
   * @param available
   * @return
   */
  public FlowHydrograph allocateSupply(FlowHydrograph available){
    FlowHydrograph totalDemand = hydrograph1.getSum(hydrograph2);
    FlowHydrograph overage     = available.getDiff(totalDemand);
    FlowHydrograph[] posNeg    = overage.getSignedHydrograph();
    FlowHydrograph unmet       = posNeg[1].getInverse();
    FlowHydrograph surplus     = posNeg[0];
    hydrograph1                = available.getDiff(surplus);
    hydrograph2                = unmet;
    return surplus;
  }

  public String report(FlowRateUnits units){
	  String ret = "Supply Request: " + System.getProperty("line.separator");
	  ret += "hydrograph1: " + System.getProperty("line.separator");
	  ret += hydrograph1.report(units);
	  ret += "hydrograph2: " + System.getProperty("line.separator");
	  ret += hydrograph2.report(units);
	  return ret;
  }
  
}
