package cnh.crb.hydro;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;

import cnh.crb.household.Household;
import cnh.crb.hydro.UnitsAndConverters.VolumeUnits;
import cnh.crb.util.Utilities;

/**
 * Represents a 'service' or 'tap' in a water system. This is essentially a
 * meter, but with the capability of recording arbitrary readings, identified by
 * name. Each time a reading is taken, the (named) result is stored, along with
 * the date and the current value of the 'meter'. It is then possible to extract
 * usage since the last reading. To get usage between two dates, take a named
 * reading at the start and end dates and calculate the difference.
 * 
 * Internally these are all stored as M^3. Conversion to Gallons or CCF is
 * provided.
 * 
 * @author murphy
 * 
 */
public class Tap {

	// Static classes

	/**
	 * Represents a reading of the meter. Can retrieve value with any volume units.
	 * 
	 * @author murphy
	 * 
	 */
	public static class Reading {
		private Date        dateOfReading;
		private double      value;
		private VolumeUnits units;

		public Reading(Date date, double val, VolumeUnits volumeUnits) {
			dateOfReading = date;
			value         = val;
			units         = volumeUnits;
		}

		public Date getDate() {
			return dateOfReading;
		}

		public double getValue(VolumeUnits volumeUnits) {
			return UnitsAndConverters.convertVolume(value, units, volumeUnits);
		}

		public double getValue() {
			return value;
		} // Units can be determined from getUnits();

		public VolumeUnits getUnits() {
			return units;
		}
	}

	/**
	 * Class to record usage; usage is not only a volume and a date (like a
	 * Reading), but also a number of days.
	 * 
	 * @author murphy
	 * 
	 */
	public static class Usage {

		private Date        startDate;
		private double      volume;
		private int         numberOfDays;
		private VolumeUnits units;

		private Usage(Date startDate, double volume, int numberOfDays, VolumeUnits units){
			this.startDate    = startDate;
			this.volume       = volume;
			this.numberOfDays = numberOfDays;
			this.units        = units;
		}
		
        public Usage(Reading reading1, Reading reading2, VolumeUnits volumeUnits) {
			this.startDate = reading1.getDate();
			numberOfDays = (int) Utilities.daysBetweenDates(startDate, reading2.getDate());
			volume       = reading2.getValue(volumeUnits)	- reading1.getValue(volumeUnits);
			units        = volumeUnits;
		}

		public int getDays(){
			return numberOfDays;
		}

		public Date getStartDate(){
			return this.startDate;
		}

		public VolumeUnits getUnits(){
			return units;
		}

		public double getUsage(){
			return volume;
		}

		// Note: returns the base value; units can be determined using getUnits.
		public double getVolumeUsed(VolumeUnits volumeUnits) {
			return UnitsAndConverters.convertVolume(volume, units, volumeUnits);
		}

//		/**
//		 *  Gets the usage between two dates, by subtracting the
//		 *  laterUsage from the later one.
//		 *  
//		 *  Example: this = Usage starting 3/1/2000, 121 days, 1200 CCF
//		 *           laterUsage = Usage starting 5/1/2000, 60 days, 700 CCF
//		 *           returns Usage starting 3/1/2000, 61 days, 500 CCF
//		 *           
//		 *  If the 'later usage' is actually later, returns null.
//		 *  
//		 *  Units will be in the native units of this usage
//		 * @param 
//		 * @return
//		 */
//		public Usage subtract(Usage laterUsage){
//			if(laterUsage.startDate.before(startDate)) return null;
//			return 
//				new Usage(startDate, 
//						  volume - laterUsage.getVolumeUsed(this.units), 
//						  (int)Utilities.daysBetweenDates(startDate, laterUsage.startDate), 
//						  this.units);
//		}
		
		 
		/**
		 * Gets the monthly average for the period covered by this Usage instance.
		 * Calculated by determining the number of months and dividing the
		 * total usage. Months are calculated fractionally based on days in month.
		 * Value returned is in the native units of this Usage instance.
		 * @return
		 */
		public double getMonthlyAverage(){
			return getMonthlyAverage(units);
		}
		 
		/**
		 * Gets the monthly average for the period covered by this Usage instance.
		 * Calculated by determining the number of months and dividing the
		 * total usage. Months are calculated fractionally based on days in month.
		 * Value returned is in the specified units.
		 * @return
		 */
		public double getMonthlyAverage(VolumeUnits volumeUnits){
			if(numberOfDays == 0) return 0;
			GregorianCalendar cal = (GregorianCalendar)startDate.clone();
			int daysInMonth = Utilities.daysInMonth(cal);
			double monthCount = 0;
			for(int i = 0; i < numberOfDays;){
				// Can skip months...
				if( (cal.get(Calendar.DAY_OF_MONTH) == 0) ){
				  daysInMonth = Utilities.daysInMonth(cal);
				  if(numberOfDays > (i + daysInMonth) ) {
					monthCount++;
					i += daysInMonth;
					cal.add(Calendar.MONTH, 1);
				  }
				  else{
					monthCount += 1d/daysInMonth;
					i++;
				  }
				}
				else{
				  monthCount += 1d/daysInMonth;
				  i++;
				}
			}
			return this.getVolumeUsed(volumeUnits)/monthCount;
		}
		
		
		 

		
		public String toString() {
			return String.format(
					"Tap.Usage: " + startDate + " days: %d, usage: %f  %s",
					numberOfDays, volume, units);
		}


	}


	// Instance variables
	private double               meter;
	private Map<String, Reading> readings = new HashMap<String, Reading>();
	private VolumeUnits          units;
	private Date                 startDate;
	private double               initialVal;


	/**
	 * Creates a Tap.
	 * 
	 * @param initialDate
	 * @param initialValue
	 * @param volumeUnits
	 */
	public Tap(Date initialDate, double initialValue, VolumeUnits volumeUnits) {
		startDate = initialDate;
		initialVal = initialValue;
		meter = initialValue;
		units = volumeUnits;
	}

	/**
	 * Gets a reading of the total value for the meter. Does not track any named
	 * accumulator
	 * 
	 * @return
	 */
	public Reading readMeter() {
		return new Reading(startDate, meter, units);
	}

	/**
	 * Gets the previous reading for the named accumulator. If none exists,
	 * creates one that assumes it was the initial meter value and was taken on
	 * the initialization date.
	 * 
	 * @param accumulatorName
	 * @param date
	 * @return
	 */
	public Reading getPreviousReading(String accumulatorName) {
		if (readings.containsKey(accumulatorName))
			return readings.get(accumulatorName);

		// Otherwise it's new
		Reading newReading = new Reading(startDate, initialVal, units);
		readings.put(accumulatorName, newReading);
		return newReading;
	}

	/**
	 * Gets the usage, in native units, since the last reading on this
	 * accumulator. This will be the difference between the current meter value
	 * and the value that is recorded in the specified accumulator. Re-sets the
	 * accumulator to the current meter value.
	 * 
	 * @param accumulatorName
	 * @param date
	 * @return
	 */
	public Usage getUsage(String accumulatorName, Date date) {
		return getUsage(accumulatorName, date, units);
	}

	/**
	 * Gets the usage, in native units, since the last reading on this
	 * accumulator. Optionally re-sets the accumulator to the current meter
	 * value.
	 * 
	 * @param accumulatorName
	 * @param date
	 * @param reset
	 * @return
	 */
	public Usage getUsage(String accumulatorName, Date date, boolean reset) {
		return getUsage(accumulatorName, date, units, reset);
	}

	/**
	 * Gets the usage since the last reading on this accumulator. This will be
	 * the difference between the current meter value and the value that is
	 * recorded in the specified accumulator. Re-sets the accumulator to the
	 * current meter value.
	 * 
	 * @param accumulatorName
	 * @param date
	 * @return
	 */
	public Usage getUsage(String accumulatorName, Date date, VolumeUnits units) {
		return getUsage(accumulatorName, date, units, true);
	}

	/**
	 * Gets the usage since the last reading on this accumulator. Optionally
	 * re-sets the accumulator to the current meter value.
	 * 
	 * @param accumulatorName
	 * @param date
	 * @param reset
	 * @return
	 */
	public Usage getUsage(String accumulatorName, Date date, VolumeUnits volumeUnits, boolean reset) {
		Reading oldReading = getPreviousReading(accumulatorName);
		Reading newReading = new Reading(date, getValue(volumeUnits), volumeUnits);
		if (reset) readings.put(accumulatorName, newReading);
		Usage usage = new Usage(oldReading, newReading, volumeUnits);
		return usage;
	}
	
	
	/**
	 * Gets the usage between two accumulators, using one as the end and the other as the start.
	 * If either or both of the accumulators do not exist they are created, just as in the normal
	 * getUsage case. If the end accumulator reading date is actually earlier than the start
	 * accumulator reading date, returns null.
	 * @param startAccumulator
	 * @param endAccumulator
	 * @param date
	 * @return
	 */
	public Usage getUsage(String startAccumulator, String endAccumulator, Date date){
		return getUsage(startAccumulator, endAccumulator, date, this.units);
	}
	
	/**
	 * Gets the usage between two accumulators, using one as the end and the other as the start.
	 * If either or both of the accumulators do not exist they are created, just as in the normal
	 * getUsage case. If the end accumulator reading date is actually earlier than the start
	 * accumulator reading date, returns null.
	 * @param startAccumulator
	 * @param endAccumulator
	 * @param date
	 * @param volumeUnits
	 * @return
	 */
	public Usage getUsage(String startAccumulator, String endAccumulator, Date date, VolumeUnits volumeUnits){
		// TODO Two ways to do this- debug to ensure results are the same
		return new Usage(getPreviousReading(startAccumulator), getPreviousReading(endAccumulator), volumeUnits);
//		return getUsage(startAccumulator, date,volumeUnits, false).subtract(getUsage(endAccumulator, date, volumeUnits, false));

	}
	
	


	/**
	 * Assumes the amount being added is in the same units as the Tap meter.
	 * 
	 * @param amountToAddToMeter
	 * @return The new value of the meter.
	 */
	public double accumulate(double amountToAddToMeter) {
		return accumulate(amountToAddToMeter, units);
	}

	/**
	 * Adds an amount to the meter. The amount can be specified in units other
	 * than the meter's native units. The return value is the new value of the
	 * meter converted from native units to the units specified.
	 * 
	 * @param amountToAddToMeter
	 * @param volumeUnits
	 * @return
	 */
	public double accumulate(double amountToAddToMeter, VolumeUnits volumeUnits) {
		meter += UnitsAndConverters.convertVolume(amountToAddToMeter,
				volumeUnits, units);
		return UnitsAndConverters.convertVolume(meter, units, volumeUnits);
	}

	/**
	 * Convenience method for determining what units this meter uses natively.
	 * 
	 * @return
	 */
	public VolumeUnits getUnits() {
		return units;
	}

	/**
	 * Returns the value of the meter, in native units.
	 * 
	 * @return
	 */
	public double getValue() {
		return meter;
	}

	/**
	 * Drops a previous reading; the accumulator is lost.
	 * @param accumulatorName
	 */
	public void dropAccumulator(String accumulatorName){
		readings.remove(accumulatorName);
	}
	
	/**
	 * Returns the value of the meter in the specified units.
	 * 
	 * @param volumeUnits
	 * @return
	 */
	public double getValue(VolumeUnits volumeUnits) {
		return UnitsAndConverters.convertVolume(meter, units, volumeUnits);
	}

}
