package cnh.crb.parameter;

import java.util.HashMap;
import java.util.Map;

import repast.simphony.distribution.RSDistribution;
import repast.simphony.distribution.RSDistributionFactory;
import cern.jet.random.engine.RandomEngine;

/**
 * RSDistribution implementation that delegates to a wrapped distribution.
 * 
 * @author Nick Collier
 */
public class DistributionParameter implements RSDistribution {

	public static final String TYPE_ID = "distribution";

	private RSDistribution dist;
	private String type, name;
	private Map<String, String> params = new HashMap<>();

	/**
	 * Creates a DistributionParameter the specified name, from the specified
	 * definition string. The string must specify the distribution type and
	 * optionally the input parameters for that distribution.
	 * 
	 * @param name
	 * @param definition
	 */
	public DistributionParameter(String name, String type, Map<String, String> params) {
		this.name = name;
		this.type = type;
		this.params.putAll(params);
	}
	
	/**
	 * Initialize this DistributionParameter by creating the 
	 * distribution that it wraps using the specified RandomEngine.
	 * 
	 * @param generator the RandomEngine to use to generate the random 
	 * numbers
	 */
	public void initialize(RandomEngine generator) {
		Map<String, String> map = new HashMap<String, String>();
		map.putAll(this.params);
		map.put("name", name);
		dist = RSDistributionFactory.createDistribution(type, map, generator);
	}
	
	/**
	 * Initialize this DistributionParameter by creating the 
	 * distribution that it wraps using the specified RandomEngine
	 * and parameters. The params will override those contained
	 * in this distribution.
	 * 
	 * @param generator the RandomEngine to use to generate the random 
	 * numbers
	 */
	public void initialize(RandomEngine generator, Map<String, String> params) {
		Map<String, String> map = new HashMap<String, String>();
		map.putAll(this.params);
		map.putAll(params);
		map.put("name", name);
		dist = RSDistributionFactory.createDistribution(type, map, generator);
	}

	/**
	 * Updates the name of the distribution that this DistributionParameter
	 * delegates to.
	 * 
	 * @param name
	 */
	public void updateName(String name) {
		this.name = name;
	}
	
	/**
	 * Gets the type of distribution.
	 * 
	 * @return the type of distribution.
	 */
	public String getType() {
		return type;
	}
	
	/**
	 * Gets the map of the parameters that are used to setup the distribution 
	 * (mean, std, etc.). The key is the name of the parameter and the value is
	 * the value.
	 * 
	 * @return the map of the parameters that are used to setup the distribution 
	 * (mean, std, etc.).
	 */
	public Map<String, String> getDistributionParameters() {
		return new HashMap<String, String>(params);
	}

	

	/*
	 * (non-Javadoc)
	 * 
	 * @see repast.simphony.distribution.RSDistribution#getNewRandom()
	 */
	@Override
	public double nextDouble() {
		return dist.nextDouble();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see repast.simphony.distribution.RSDistribution#getName()
	 */
	@Override
	public String getName() {
		return dist == null ? name : dist.getName();
	}

	public String toString() {
		StringBuilder buf = new StringBuilder(type);
		if (params.size() > 0) {
			for (Map.Entry<String, String> entry : params.entrySet()) {
				buf.append(" ");
				buf.append(entry.getKey());
				buf.append(" ");
				buf.append(entry.getValue());
			}
		}
		return buf.toString();
	}

}
