package cnh.crb.regional.tucson;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import cnh.crb.econ.CostCalculator;
import cnh.crb.econ.CostCalculatorFactory;
import cnh.crb.econ.SimpleCostCalculator;

public class TucsonHouseholdSewerChargeCalculatorLoader {

	private static final QName ENTRY = new QName("entry");
	private static final QName YEAR = new QName("year");
	private static final QName MONTH = new QName("month");
	private static final QName COST = new QName("cost");
	private static final QName THRESHOLD = new QName("threshold");

	public static class TucsonHSCCCostCalculatorFactory {

		private TucsonHouseholdSewerChargeCalculator calc = new TucsonHouseholdSewerChargeCalculator();
		
		/**
		 * Adds an entry to the cost calculator. The date is the start of the
		 * period for which the value will be returned. The non-inclusive end of the period
		 * is the date of the nearest entry that is greater than this one. 
		 * 
		 * @param date
		 * @param value
		 */
		public void addEntry(Date date, double value) {
			this.addEntry(date, value, Double.NEGATIVE_INFINITY);
		}
		
		/**
		 * Adds an entry to the cost calculator. The date is the start of the
		 * period for which the value will be returned. The non-inclusive end of the period
		 * is the date of the nearest entry that is greater than this one. 
		 * 
		 * @param date
		 * @param value
		 * @param threshold the amount that must be exceeded before any cost 
		 * is incurred.
		 */
		public void addEntry(Date date, double value, double threshold) {
			calc.valueMap.put(date, new TucsonHouseholdSewerChargeCalculator.CostData(value, threshold));
		}
		
		/**
		 * Creates a calculator containing all the entries
		 * added between calls to createCalculator. Any
		 * calls to addEntry subsequent to calling this will
		 * create a new calculator. 
		 * 
		 * @return a calculator containing all the entries
		 * added between calls to createCalculator.
		 */
		public CostCalculator createCalculator() {
			CostCalculator tmp = calc;
			calc = new TucsonHouseholdSewerChargeCalculator();
			return tmp;
		}
	}
	
	/**
	 * Loads a simple cost table file and returns the SimpleCostCalculator created
	 * from it.
	 * 
	 * @param file
	 * @return
	 * @throws IOException
	 * @throws XMLStreamException
	 */
	public CostCalculator load(String file) throws IOException, XMLStreamException {

		Calendar cal = GregorianCalendar.getInstance();
		cal.set(Calendar.MILLISECOND, 0);
		XMLInputFactory factory = XMLInputFactory.newInstance();
		XMLEventReader reader = null;
		TucsonHSCCCostCalculatorFactory costFactory = new TucsonHSCCCostCalculatorFactory();
		try {
			reader = factory.createXMLEventReader(new BufferedReader(new FileReader(file)));

			while (reader.hasNext()) {
				XMLEvent evt = reader.nextEvent();
				if (evt.isStartElement()) {
					StartElement elm = evt.asStartElement();
					QName elmName = elm.getName();
					if (elmName.equals(ENTRY)) {
						String sYear = getAttribute(elm, YEAR);
						String sMonth = getAttribute(elm, MONTH);
						String sCost = getAttribute(elm, COST);
						String sThreshold = null;
						if (elm.getAttributeByName(THRESHOLD) != null) {
							sThreshold = getAttribute(elm, THRESHOLD);
						}

						try {
							int year = Integer.parseInt(sYear);
							int month = Integer.parseInt(sMonth);
							cal.set(year, month, 1, 0, 0, 0);
							double cost = Double.parseDouble(sCost);
							if (sThreshold != null) {
								costFactory.addEntry(cal.getTime(), cost, Double.parseDouble(sThreshold));
							} else {
								costFactory.addEntry(cal.getTime(), cost);
							}
						} catch (NumberFormatException ex) {
							throw new IOException("Invalid attribute. " + ex.getMessage());
						}
					}

				}
			}
		} finally {
			if (reader != null)
				try {
					reader.close();
				} catch (Exception ex) {
				}
		}

		return costFactory.createCalculator();
	}

	private String getAttribute(StartElement elm, QName name) throws IOException {
		Attribute attrib = elm.getAttributeByName(name);
		if (attrib == null)
			throw new IOException("Invalid simple rate table. entry element is missing required "
					+ name.getLocalPart() + " attribute");
		return attrib.getValue();
	}

	
}
