package cnh.crb.regional.tucson;

import java.util.Calendar;
import java.util.GregorianCalendar;

import cnh.crb.agent.ResidentialWaterManager;
import cnh.crb.econ.BillableEntity;
import cnh.crb.econ.HouseholdBillCalculator;
import cnh.crb.household.Household;
import cnh.crb.hydro.FlowNode;
import cnh.crb.hydro.Reservoir;
import cnh.crb.hydro.Tap;
import cnh.crb.hydro.UnitsAndConverters.VolumeUnits;
import cnh.crb.util.Utilities;

public class TucsonWaterDept extends ResidentialWaterManager {

	
	/**
	 * From time to time Tucson Water will need the 'winter usage'
	 * for a given household. This is the usage from Nov 1 to May 1.
	 * However, the method to request this information from a 'Tap'
	 * may return an invalid result- for example, if the Tap was
	 * initialized on Dec 1, so the record is incomplete. This method
	 * can determine whether a given record is a full, valid 'winter
	 * usage' record or not.
	 * 
	 * Returns valid iff: usage > 0, days > 0, and the number of days
	 * is the correct number (181 or 182, depending on whether it is
	 * a leap year).
	 * 
	 * @param winterUsage
	 * @param year
	 * @return
	 */
	public static boolean winterUsageIsValid(Tap.Usage winterUsage, int year){
        boolean usageIsValid = winterUsage.getDays() > 0 && winterUsage.getUsage() > 0;
		
		if(usageIsValid){
			GregorianCalendar winterStart = new GregorianCalendar();
			winterStart.set(year -1, 10, 1);
			GregorianCalendar winterEnd   = new GregorianCalendar();
			winterEnd.set(year, 4, 1);
			usageIsValid = (winterUsage.getDays() == Utilities.daysBetweenDates(winterStart, winterEnd)); 
		}
		return usageIsValid;
	}
	
	/**
	 * From time to time Tucson Water will need a 'sewer usage'
	 * for a given household. This is the usage from Dec 1 to Mar 1.
	 * However, the method to request this information from a 'Tap'
	 * may return an invalid result- for example, if the Tap was
	 * initialized on Dec 20, so the record is incomplete. This method
	 * can determine whether a given record is a full, valid 'winter
	 * usage' record or not.
	 * 
	 * Returns valid iff: usage > 0, days > 0, and the number of days
	 * is the correct number (90 or 91, depending on whether it is
	 * a leap year).
	 * 
	 * @param sewerUsage
	 * @param year
	 * @return
	 */
	public static boolean sewerUsageIsValid(Tap.Usage sewerUsage, int year){
        boolean usageIsValid = sewerUsage.getDays() > 0 && sewerUsage.getUsage() > 0;
		
		if(usageIsValid){
			GregorianCalendar winterStart = new GregorianCalendar(year -1, 10, 1);
			GregorianCalendar winterEnd   = new GregorianCalendar(year, 3, 1);
			usageIsValid = (sewerUsage.getDays() == Utilities.daysBetweenDates(winterStart, winterEnd)); 
		}
		return usageIsValid;
	}
	
	TucsonHouseholdSummerSurchargeCalculator summerSurchargeCalculator;
	
	// Constructor
	
	public TucsonWaterDept(Reservoir reservoir, HouseholdBillCalculator billCalculator, 
			Calendar initialDate, String managerID) {
		super(reservoir, billCalculator, initialDate, managerID);
	}
	
	public void setSummerSurchargeCalculator(TucsonHouseholdSummerSurchargeCalculator surchargeCalculator){
		summerSurchargeCalculator = surchargeCalculator;
	}
	
	public void billCustomers(Calendar billingDate) {
		int month = billingDate.get(Calendar.MONTH);
		if(billingDate.get(Calendar.DAY_OF_MONTH) == 1){
			switch (month){
			  case Calendar.MARCH:
				  markSewerEnd(billingDate);
				  break;
			  case Calendar.MAY:
				  markWinterEnd(billingDate);
				  break;
			  case Calendar.NOVEMBER:
				  markWinterStart(billingDate);
				  break;
			  case Calendar.DECEMBER:
				  markSewerStart(billingDate);
				  break;
			  default: // Do nothing
		    }
		}
		super.billCustomers(billingDate);
	}
	
	
	// Mark all residents' 'Winter Start' on Nov 1
	public void markWinterStart(Calendar cal){
		int year     = cal.get(Calendar.YEAR);
		String current  = "WINTER_START_" + year;
		String previous = "WINTER_START_" + (year - 2);
		for (FlowNode node : waterRecipients){
			if (node instanceof Household){
				BillableEntity household = (BillableEntity)node;
		        household.getMeteredWaterUse(current);
		        household.dropMeteredWaterUse(previous);
			}
		}
	}
	
	// Mark all residents' 'Sewer Start' on Dec 1
	public void markSewerStart(Calendar cal){
		int year = cal.get(Calendar.YEAR);
		String current  = "SEWER_START_" + year;
		String previous = "SEWER_START_" + (year - 2);
		for (FlowNode node : waterRecipients){
			if (node instanceof Household){
				BillableEntity household = (BillableEntity)node;
		        household.getMeteredWaterUse(current);
		        household.dropMeteredWaterUse(previous);
			}
		}
	}
	// Mark all residents' 'Sewer End' on Mar 1
	public void markSewerEnd(Calendar cal){
		int year = cal.get(Calendar.YEAR);
		String current  = "SEWER_END_" + year;
		String previous = "SEWER_END_" + (year - 2);
		for (FlowNode node : waterRecipients){
			if (node instanceof Household){
				BillableEntity household = (BillableEntity)node;
		        household.getMeteredWaterUse(current);
		        household.dropMeteredWaterUse(previous);
			}
		}
	}
	
	
	// Mark all residents' 'Winter End' on May 1
	// When doing this, also collect the average winter usage and pass that to the surcharge calculator;
	// going forward, it will use this as the default if a household's record is incomplete
	public void markWinterEnd(Calendar cal){
		int year = cal.get(Calendar.YEAR);
		String current      = "WINTER_END_"   + year;
		String startCurrent = "WINTER_START_" + (year - 1);
		String previous     = "WINTER_END_"   + (year - 2);
		int count    = 0;
	    double total = 0;
		
		for (FlowNode node : waterRecipients){
		    if (node instanceof Household){
		    	BillableEntity household = (BillableEntity)node;
		        Tap.Usage usage = household.getMeteredWaterUse(startCurrent, current);
		        // Determine if the usage is valid
		        boolean usageIsValid = false;
		        if(usageIsValid){
		        	total+= usage.getMonthlyAverage(VolumeUnits.CCF);
		        	count++;
		        }
		        household.dropMeteredWaterUse(previous);
		        
			}
		}
		if(count > 0) summerSurchargeCalculator.setGlobalDefault(total/(double)count);
	}
	
}
