/**
 * 
 */
package cnh.crb.stats;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;

import repast.simphony.data2.util.DataUtilities;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import cnh.crb.household.Household;

/**
 * Histograms various data from the Households.
 * 
 * @author nick
 */
public class HistogramDataRecorder {

	private static String PRICE_FACTOR_FILE = "price_factor_histogram.csv";
	private static String SOCIAL_FACTOR_FILE = "social_factor_histogram.csv";
	private static String USAGE_FILE = "usage_histogram.csv";
	private static String SERVICES = "service_counts_histogram.csv";
	private static String FLOW_REDUCTION = "low_flow_fixture_reduction_histogram.csv";

	private Histogram price, socialFactor, usage, services, lowFlowReduction;
	private BufferedWriter priceWriter, sfWriter, usageWriter, servicesWriter, lowFlowReductionWriter;

	public HistogramDataRecorder() throws IOException {
		double[] priceBins = { 0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0 };
		price = new Histogram(priceBins, false);

		double[] sfBins = { -1, -0.8, -0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6, 0.8, 1.0 };
		socialFactor = new Histogram(sfBins, false);

		double[] usageBins = { 0, 11, 16, 31, 45, Double.POSITIVE_INFINITY };
		usage = new Histogram(usageBins, true);
		// same bins for services
		services = new Histogram(usageBins, false);
		
		double[] flowBins = { 0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0 };
		lowFlowReduction = new Histogram(flowBins, false);

		initWriters();
	}

	private void initWriters() throws IOException {
		priceWriter = createWriter(PRICE_FACTOR_FILE, price);
		sfWriter = createWriter(SOCIAL_FACTOR_FILE, socialFactor);
		usageWriter = createWriter(USAGE_FILE, usage);
		servicesWriter = createWriter(SERVICES, services);
		lowFlowReductionWriter = createWriter(FLOW_REDUCTION, lowFlowReduction);
	}

	private BufferedWriter createWriter(String file, Histogram hist) throws IOException {
		Path path = Paths.get(file);
		BufferedWriter writer = null;
		if (RunEnvironment.getInstance().isBatch()) {
			if (path.toFile().exists()) {
				// open with append
				writer = Files.newBufferedWriter(path, Charset.forName("UTF-8"),
						StandardOpenOption.WRITE, StandardOpenOption.CREATE,
						StandardOpenOption.APPEND);
			} else {
				writer = Files.newBufferedWriter(path, Charset.forName("UTF-8"),
						StandardOpenOption.WRITE, StandardOpenOption.CREATE_NEW);
				writeHeader(writer, hist);

			}
		} else {
			DataUtilities.renameFileIfExists(path.toString());
			writer = Files.newBufferedWriter(path, Charset.forName("UTF-8"),
					StandardOpenOption.WRITE, StandardOpenOption.CREATE_NEW);
			writeHeader(writer, hist);
		}
		return writer;
	}

	private void writeHeader(BufferedWriter writer, Histogram hist) throws IOException {
		StringBuilder builder = new StringBuilder();
		builder.append("batch,run,tick");
		for (int i = 0; i < hist.binCount(); ++i) {
			builder.append(",");
			builder.append(hist.getBinTitle(i));
		}
		builder.append("\n");
		writer.write(builder.toString());
	}
	
	private void writeData(BufferedWriter writer, Histogram hist) throws IOException {
		StringBuilder builder = new StringBuilder();
		builder.append(RunState.getInstance().getRunInfo().getBatchNumber());
		builder.append(",");
		builder.append(RunState.getInstance().getRunInfo().getRunNumber());
		builder.append(",");
		builder.append(RunEnvironment.getInstance().getCurrentSchedule().getTickCount());
		for (int i = 0; i < hist.binCount(); ++i) {
			builder.append(",");
			builder.append(hist.getValue(i));
		}
		builder.append("\n");
		writer.write(builder.toString());
	}

	public void record(Household household) {
		price.addValue(household.getPriceFactor());
		socialFactor.addValue(household.getSocialFactors());
		double ccf = household.getBillMemory().getLastBillUsage();
		usage.addValue(ccf);
		services.addValue(ccf);
		lowFlowReduction.addValue(household.getLowFlowFixtureReduction());
	}

	public void writeAndReset() throws IOException {
		writeData(priceWriter, price);
		writeData(sfWriter, socialFactor);
		writeData(usageWriter, usage);
		writeData(servicesWriter, services);
		writeData(lowFlowReductionWriter, lowFlowReduction);
		reset();
	}
	
	public void close() throws IOException {
		priceWriter.close();
		sfWriter.close();
		usageWriter.close();
		servicesWriter.close();
		lowFlowReductionWriter.close();
	}

	private void reset() {
		price.reset();
		socialFactor.reset();
		usage.reset();
		services.reset();
		lowFlowReduction.reset();
	}
}
