package cnh.crb.wbm;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import repast.simphony.context.Context;
import cnh.crb.hydro.FlowRate;

/**
 * A WBMCell represents a grid cell in the WBM.
 * 
 * This class permits only one cell per ID to be created;
 * requesting the same ID a second time will give you the
 * first cell instead of a new one.
 * 
 * WBMCell objects know about Context objects, but are
 * not placed into Contexts. They use the context pointer
 * to pass to crop and farmer agents; farmers are placed
 * into contexts.
 * 
 * @author murphy
 *
 */
public class WBMCell {
	
	// Static class that will not permit duplicate IDs
	private static Map<String, WBMCell> instances = new HashMap<String, WBMCell>();
	
	public String                ID;
	public Point2D.Double        lowerLeft;
	public Point2D.Double        upperRight;
	public double                area = -1; // Units should be m^2
	
	public FlowRate              surfaceFlow   = null;
	public double                mmGroundwater = 0;
	
	public Map<Integer, WBMCrop> crops = new TreeMap<Integer, WBMCrop>();
	
	public Context               context;
	
	public static void reinitialize(){
		instances = new HashMap<String, WBMCell>();
	}
	
	public static WBMCell getInstance(String ID, Context context){
		if(instances.containsKey(ID)) return instances.get(ID);
		else{
			WBMCell cell = new WBMCell(ID, context);
			instances.put(ID,  cell);
			return cell;
		}
	}
	
	public boolean isInitialized(){ return area > 0; } // Zero values are invalid; indicates non-initialized state
	
	private WBMCell(String ID, Context context){
		this.ID = ID;
		this.context = context;
	}
	
	public boolean init(Point2D.Double lowerLeft, Point2D.Double upperRight, double area){
		if(area <= 0) return false;
		this.lowerLeft     = lowerLeft;
		this.upperRight    = upperRight;
		this.area          = area * 1000000;
		return true;
	}
	
	public void clearCrops(){
		for(WBMCrop crop: crops.values()) crop.active = false;
	}
	
	public void cullCrops(){
		List<Integer> toDelete = new ArrayList<Integer>();
		for(Map.Entry<Integer, WBMCrop> entry: crops.entrySet()) if(!entry.getValue().active) toDelete.add(entry.getKey());
		for(Integer i: toDelete){
			WBMCrop c = crops.remove(i);
			c.removeManager();
		}
	}
	
	public void setCrop(Integer id, double fraction, double moisture, double moistureFraction, double wbmDemand){
		WBMCrop crop;
		if(crops.containsKey(id)){
			crop = crops.get(id);
			crop.set(fraction, fraction * area, moisture, moistureFraction, wbmDemand);
		}
		else{
			crop = new WBMCrop(this.ID + "_" + id, id, fraction, fraction * area, moisture, moistureFraction, wbmDemand, context);
			crops.put(id, crop);
		}
		crop.active = true;
	}	
		
}
