/**
 * 
 */
package cnh.crb.wbm;

import java.awt.geom.Point2D;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.nio.charset.StandardCharsets;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.zeromq.ZMQ;

import cnh.crb.hydro.UnitsAndConverters;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.util.collections.Pair;
import simphony.util.messages.MessageCenter;

/**
 * Handles connection with the WBM water model.
 * 
 * @author nick
 */
public class WBMConnector {

	// Static values
	private   static final String        WBM_HEADER          = "lat,lon,manage_irr,flow,groundwater,deltaStorage,eliminated";
	public    static final int           DECIMAL_COORDINATES = 5;
	protected static       MessageCenter msg                 = MessageCenter.getMessageCenter(WBMConnector.class);
	private   static       int           exchangeCount       = 0;
	
	// Inner classes
	class IrrigationDemand{
		public double fraction;           // Fraction of cell area
		public double wbmDemandMM;        // Demand that the WBM calculates, in mm
		public double soilMoisture;       // WBM Soil Moisture info (mm)
		public double soilMoistureFrac;   // WBM Fraction of Soil Moisture (soilMoisture / field capacity)
	}

	class WBMReceiveDatum{
		public Point2D.Double                 requestedCoordinate;
		public String                         cellID;
		public Point2D.Double                 cellLowerLeft;
		public Point2D.Double                 cellUpperRight;
		public double                         cellArea;
		public double                         cellAnnualTemp;
		public double                         cellMonthlyTemp;
		public double                         cellDailyTemp;
		public double                         cellAnnualPrecip;
		public double                         cellMonthlyPrecip;
		public double                         cellDailyPrecip;
		public double                         surfaceDischarge;
		public double                         mmGroundwater;
		public boolean                        ABM_manages_Irrigation;
		public Map<Integer, IrrigationDemand> irrigationDemand = new TreeMap<Integer, IrrigationDemand>();
		
		public IrrigationDemand getIrrigationDemand(int id){
			if(irrigationDemand.containsKey(id)) return irrigationDemand.get(id);
			else{
				IrrigationDemand iDemand = new IrrigationDemand();
				irrigationDemand.put(id, iDemand);
				return iDemand;
			}
		}
	}
	
	
	private ZMQ.Context context;
	private ZMQ.Socket  req;
	private boolean     testWithZeroValues = false;
	
	public WBMConnector(String url) {
		context = ZMQ.context(1);
		req = context.socket(ZMQ.REQ);
		req.connect(url);
		testWithZeroValues = RunEnvironment.getInstance().getParameters()
				.getBoolean("ForceZeroTakenFromWBM");
		System.out.println("Setting test with zero values to " + testWithZeroValues);
	}
	
	public void init(List<Point2D.Double> data) {
		List<Pair<Point2D.Double, WBMSendDatum>> newData = new ArrayList<Pair<Point2D.Double, WBMSendDatum> >();
		for(Point2D.Double point: data){
			WBMSendDatum toSend = new WBMSendDatum();
			toSend.netSurfaceFlow = 0; // This is the default, but the point of this routine is to initialize
			newData.add(new Pair<Point2D.Double, WBMSendDatum>(point, toSend));
		}
		notifyWBMOfTaken(newData);
	}
	
	public void notifyWBMOfTaken(List<Pair<Point2D.Double, WBMSendDatum>> data){
		msg.debug("ABOUT TO NOTIFY...");
		String decs = ("##############").substring(0, DECIMAL_COORDINATES);
		DecimalFormat df = new DecimalFormat("#." + decs);
		df.setMaximumFractionDigits(DECIMAL_COORDINATES);
		
		DecimalFormat dfVal = new DecimalFormat("#.######");
		dfVal.setMaximumFractionDigits(6);
		
		StringBuilder WBMMessage = new StringBuilder(WBM_HEADER + "\n");
		StringBuilder printout = new StringBuilder(exchangeCount == 0 ? "EXCHANGE,EXCHANGENUM," + WBM_HEADER + "Crop,Amt_mm,AmtAF\n": "");
		int q = 0;
		for(int i = 0; i < data.size(); i++){
			StringBuilder line = new StringBuilder();
			Pair<Point2D.Double, WBMSendDatum> datum = data.get(i);
			Point2D.Double latlon = datum.getFirst();
		    line.append(df.format(latlon.getY()));
		    line.append(",");
		    line.append(df.format(latlon.getX()));
		    line.append(",");
			if(!testWithZeroValues){ 
			  WBMSendDatum toSend = datum.getSecond();
			  line.append((datum.getSecond().manageIrrigation ? "1" : "0") + ",");
			  line.append(dfVal.format(toSend.netSurfaceFlow) + ",");
			  line.append(dfVal.format(toSend.netBaseFlow) + ",");
			  line.append(dfVal.format(toSend.deltaStorage) + ",");
			  line.append(dfVal.format(toSend.discard));
			  String baseLine = line.toString();
			  Map<Integer, Double> toIrrigate = toSend.irrigation;
			  int count = 0;
			  for(Map.Entry<Integer, Double> irrigate: toIrrigate.entrySet()){
				  if(irrigate.getValue() > 0){
					  printout.append("EXCHANGE," + exchangeCount + "," + baseLine + "," + irrigate.getKey() + "," + irrigate.getValue() + "," + UnitsAndConverters.M3ToAF((irrigate.getValue()/1000) * toSend.irrigatedAreas.get(irrigate.getKey())) + "\n");
				  }					  
				  count++;
				  line.append("," + irrigate.getKey() + "," + irrigate.getValue());
			  }
			  if(count == 0) printout.append("EXCHANGE," + exchangeCount + "," + baseLine + ",-1,-1\n");
			}
			else line.append("0,0,0,0,0"); // For testing, always send zeros
			line.append("\n");
			
			WBMMessage.append(line.toString());
			q++;
		}
		System.out.println("CONNECTOR NOTIFYING (" + q + " with length " + WBMMessage.toString().length() + "): ");
		System.out.println(WBMMessage.toString());
		System.out.println("----- end");
		System.out.println(printout.toString());
		exchangeCount++;
		req.send(WBMMessage.toString());
	}
	
	public List<WBMReceiveDatum> receiveFromWBM() {
		String available = req.recvStr(StandardCharsets.UTF_8);
		System.out.println("RECEIVING: (" + available.length() + ")");
		System.out.println(available);
		List<WBMReceiveDatum > ret = new ArrayList<WBMReceiveDatum>();
		try{
			BufferedReader reader = new BufferedReader(new StringReader(available));
			String line = reader.readLine();  	// Skips header
			line = reader.readLine();
			while((line != null) && (line.length() > 0)){
				String[] vals = line.split(",");
				WBMReceiveDatum toReturn = new WBMReceiveDatum();
				toReturn.requestedCoordinate    = new Point2D.Double(Double.parseDouble(vals[2].trim()), Double.parseDouble(vals[1].trim()));
				toReturn.cellID                 = vals[3];
				toReturn.cellLowerLeft          = new Point2D.Double(Double.parseDouble(vals[5].trim()), Double.parseDouble(vals[4].trim()));
				toReturn.cellUpperRight         = new Point2D.Double(Double.parseDouble(vals[7].trim()), Double.parseDouble(vals[6].trim()));
				toReturn.cellArea               = Double.parseDouble(vals[8].trim());
				toReturn.cellAnnualTemp         = Double.parseDouble(vals[9].trim());
				toReturn.cellMonthlyTemp        = Double.parseDouble(vals[10].trim());
				toReturn.cellDailyTemp          = Double.parseDouble(vals[11].trim());
				toReturn.cellAnnualPrecip       = Double.parseDouble(vals[12].trim());
				toReturn.cellMonthlyPrecip      = Double.parseDouble(vals[13].trim());
				toReturn.cellDailyPrecip        = Double.parseDouble(vals[14].trim());
				toReturn.surfaceDischarge       = Double.parseDouble(vals[15].trim());
				toReturn.mmGroundwater          = Double.parseDouble(vals[16].trim());
				toReturn.ABM_manages_Irrigation = Boolean.parseBoolean(vals[17].trim());
				// Collect irrigation stats here
				int base = 18;
				while(base < vals.length){
					IrrigationDemand iDemand = toReturn.getIrrigationDemand(Integer.parseInt(vals[base]));
					iDemand.fraction = Double.parseDouble(vals[base + 1]);
					iDemand.wbmDemandMM = Double.parseDouble(vals[base + 2]);
					iDemand.soilMoisture = Double.parseDouble(vals[base + 3]);
					iDemand.soilMoistureFrac = Double.parseDouble(vals[base + 4]);
					base += 5;
				}
				ret.add(toReturn);
				line = reader.readLine();
			}
		} catch (IOException e) {
			msg.debug("Something happened while parsing line...");
		}
		
		return ret;
	}
	
	public void close() {
		msg.debug("Sending end message and closing...");
		req.send("END");
		req.close();
		context.term();
	}
}
