package cnh.crb.wbm;

import java.awt.geom.Point2D;
import java.text.DecimalFormat;
import java.util.Date;
import java.util.Map;

import org.apache.log4j.Logger;

import repast.simphony.util.ContextUtils;
import cnh.crb.climate.TemperatureStation;
import cnh.crb.hydro.FlowHydrograph;
import cnh.crb.hydro.FlowRate.FlowRateUnits;
import cnh.crb.hydro.FlowRateM3S;
import cnh.crb.hydro.NaturalWaterSource;
import cnh.crb.hydro.UnitsAndConverters;
import cnh.crb.hydro.UnitsAndConverters.VolumeUnits;
import cnh.crb.hydro.WaterSink;
import cnh.crb.wbm.WBMConnector.IrrigationDemand;
import cnh.crb.wbm.WBMConnector.WBMReceiveDatum;

public class WBMNaturalWaterSource extends NaturalWaterSource implements WaterSink, TemperatureStation{

	private static Logger LOGGER = Logger.getLogger(WBMNaturalWaterSource.class);
	
	public boolean manageIrrigation = true;
	
	double volumeReturnedToSurfaceFlow = 0; // Converted internally to m^3
	double volumeReturnedToBaseFlow    = 0;
	double volumeEliminated            = 0;
	double deltaStorage                = 0;
	double monthlyMeanTemperature      = 0;
	
	protected Point2D.Double latLon;
	FlowHydrograph amountFlowed = new FlowHydrograph();
	
	double defaultFlow; // Mainly used for testing
	
	// WBM Water Sources have additional components (beyond what Natural Water Sources have)
	WBMCell wbmCell;
	
	
	public WBMNaturalWaterSource(String id, Point2D.Double location){
		super(id, new FlowHydrograph()); // Empty flow hydrograph
		latLon = location;
	}	
	
	public Point2D.Double getLatLon(){ return latLon; }
	
	public void processWBMDatum(Date start, Date end, WBMReceiveDatum datum){
		// Assumes that 'available' is the total volume over all
		// days. Assumes units are m^3/second.
		FlowHydrograph toAdd = new FlowHydrograph(start, end, datum.surfaceDischarge, FlowRateUnits.M3S);
		this.hydrograph.add(toAdd);
		
		// Collect the rest of the information from this datum
		if(wbmCell == null) wbmCell = WBMCell.getInstance(datum.cellID, ContextUtils.getContext(this));
		if(!wbmCell.isInitialized()) wbmCell.init(datum.cellLowerLeft, datum.cellUpperRight, datum.cellArea);
		// Set the cell's surface flow and groundwater values
		wbmCell.surfaceFlow   = new FlowRateM3S(datum.surfaceDischarge);
		wbmCell.mmGroundwater = datum.mmGroundwater;
		// Collect information about the cell's crops
		wbmCell.clearCrops();
		Map<Integer, IrrigationDemand> irrigationDemands = datum.irrigationDemand;
		for(Map.Entry<Integer, IrrigationDemand> entry: irrigationDemands.entrySet()){
			manageIrrigation = true;
			IrrigationDemand demand = entry.getValue();
			wbmCell.setCrop(entry.getKey(), demand.fraction, demand.soilMoisture, demand.soilMoistureFrac, demand.wbmDemandMM);
		}
		wbmCell.cullCrops();
		
		monthlyMeanTemperature = datum.cellMonthlyTemp;

	}
		
	public void flow(Date end){
		if(useFullWaterFlow) amountFlowed.add(demands.hydrograph1);
		clearFlow();
	}
	
	public void setManageIrrigation(boolean toManageIrrigation){
		manageIrrigation = toManageIrrigation;
	}
	
	// TODO Make this more efficient
	public WBMSendDatum getWBMSend(Date start, Date end){
		WBMSendDatum ret = new WBMSendDatum();
		ret.manageIrrigation = manageIrrigation;
		double volumeTaken = amountFlowed.resetBounds(start,  end).getTotalFlowVolume(VolumeUnits.M3);
		volumeTaken -= volumeReturnedToSurfaceFlow;
		ret.netSurfaceFlow = (volumeTaken / (((end.getTime() - start.getTime()) / 1000))); 
//		ret.netSurfaceFlow = 100; // TESTING!
		ret.netBaseFlow = volumeReturnedToBaseFlow;
		ret.discard = volumeEliminated;
		
		// Discard = 0 for now
		// Evaporate = 0 for now
		if(manageIrrigation){
			if(wbmCell != null){
				for(Map.Entry<Integer, WBMCrop> crops : wbmCell.crops.entrySet()){
					WBMCrop crop = crops.getValue();
					ret.irrigation.put(crop.id, crop.mmApplied);
					ret.irrigatedAreas.put(crop.id, crop.area);
				}
			}
		}
		
		volumeReturnedToBaseFlow    = 0; // Reset for next run
		volumeReturnedToSurfaceFlow = 0;
		volumeEliminated            = 0;
		deltaStorage                = 0;
		return ret;
	}
	
	public void setDefaultFlow(double defaultFlow){
		this.defaultFlow = defaultFlow;
	}
	
	public double getDefaultFlow(){ return defaultFlow; }
	
	public void eliminateWater(double volume, VolumeUnits units, WATER_SINK_METHOD method){
		// What happens here depends on the sink method:
		switch(method){
		case DELTA_STORE:
			deltaStorage += UnitsAndConverters.convertVolume(volume, units, VolumeUnits.M3);
			break;
		case RETURN_TO_SURFACE:
			volumeReturnedToSurfaceFlow += UnitsAndConverters.convertVolume(volume, units, VolumeUnits.M3);
			break;
		case RETURN_TO_GROUNDWATER:
			double toAdd = UnitsAndConverters.convertVolume(volume, units, VolumeUnits.M3);
			if(wbmCell != null) toAdd = (toAdd / wbmCell.area) / 1000;
			volumeReturnedToBaseFlow += toAdd; // WBM Needs this in mm
			break;
		case ADD_TO_UNSUSTAINABLE:
		case EVAPORATE:
		default:
			volumeEliminated += UnitsAndConverters.convertVolume(volume, units, VolumeUnits.M3);
			break;
		}      
	}
	
	
	
	public double getVolumeReturned(){ return volumeReturnedToSurfaceFlow; }
	public double getMonthlyMeanTemperature(){ return monthlyMeanTemperature; }
	
	public void addWaterSinkAmounts(WaterSink other){
		if(other instanceof WBMNaturalWaterSource) this.volumeReturnedToSurfaceFlow += ((WBMNaturalWaterSource)other).getVolumeReturned();
	}

	public void notifyStorage(double storage){
		
	}

	public void setTemperature(double newTemp) {
	   monthlyMeanTemperature = newTemp;
	}

	public double getTemperature() {
		return monthlyMeanTemperature;
	}
	
	public String getTemperatureStationId(){
		Point2D.Double point = this.getLatLon();
		DecimalFormat df = new DecimalFormat("#.########");
		df.setMaximumFractionDigits(4);
		return df.format(point.x) + "_" + df.format(point.y);
		
	}
	
}
