package cnh.crb.wbm;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import repast.simphony.util.collections.Pair;
import simphony.util.messages.MessageCenter;
import cnh.crb.climate.TemperatureStation;
import cnh.crb.wbm.WBMConnector.WBMReceiveDatum;

public class WBMWaterSourceCollection{

	Map<Point2D.Double, WBMNaturalWaterSource> waterSources = new HashMap<Point2D.Double, WBMNaturalWaterSource>();
	WBMConnector wbmConnector;
	
	// When ABM first contacts WBM it uses an arbitrary list of water sources
	// WBM, however, will have a grid cell structure, of which ABM is ignorant.
	// The list of returned values from the WBM will have 'duplicate' values
	// for any two ABM water sources that lie within the same WBM grid
	// This map will store a correspondence: one of the sources will
	// be chosen as the canonical one, and the map will indicate
	// which of the others are to be replaced with the canonical one
	Map<WBMNaturalWaterSource, WBMNaturalWaterSource> duplication = new HashMap<WBMNaturalWaterSource, WBMNaturalWaterSource>();
	Map<WBMNaturalWaterSource, WBMNaturalWaterSource> duplicates = new HashMap<WBMNaturalWaterSource, WBMNaturalWaterSource>();
	boolean hasDuplication = false;
	
	static MessageCenter msg = MessageCenter.getMessageCenter(WBMWaterSourceCollection.class);
	
	public WBMWaterSourceCollection(String url){
		wbmConnector = new WBMConnector(url);
	}
	
	
	/**
	 * Call this to create the water sources. These can be treated as
	 * natural water sources. A call to 'notifySourcesOfAvailableWater' should
	 * be scheduled.
	 * 
	 * @param lat
	 * @param lon
	 * @return
	 */
	public WBMNaturalWaterSource createWaterSource(String name, double lat, double lon){
		Point2D.Double location = new Point2D.Double(lon, lat);
		if(waterSources.containsKey(location)) return waterSources.get(location);
		WBMNaturalWaterSource newWaterSource = new WBMNaturalWaterSource(name, location);
		waterSources.put(location, newWaterSource);
		return newWaterSource;
	}

	/**
	 * After the water sources are created, call this
	 * to initialize the WBM connection and inform the WBM
	 * of the list of locations
	 */
	public void init(){
		List<Pair<Point2D.Double, WBMSendDatum> > taken = new ArrayList<Pair<Point2D.Double, WBMSendDatum> >();
		for(int i = 0; i < waterSources.size(); i++){
		    WBMNaturalWaterSource waterSource = waterSources.get(i);
            WBMSendDatum toSend = new WBMSendDatum();
            toSend.netSurfaceFlow = -1.0;
			taken.add(new Pair<Point2D.Double, WBMSendDatum> (waterSource.getLatLon(), toSend));
		}
		wbmConnector.notifyWBMOfTaken(taken);
	}
	
	
	/**
	 * Contacts the WBM, gets the water available over the specified
	 * interval, and notifies the WBMNaturalWaterSources of the
	 * amount.
	 * 
	 * Note that start and end dates are not sent to the WBM;
	 * the calling program must know what time step the WBM is
	 * using and call this to match. So, if the WBM is operating
	 * on a daily time step and the current simulation date
	 * is 1 March 2014, the start and end dates should be
	 * March 1 and 2. If the WBM is on a 'monthly' time step,
	 * then the start and end dates would be March 1 and April 1.
	 * (Assuming that the WBM actually uses a calendar, and not
	 * just a 30-days-equals-a-month proxy).
	 */
	public void notifySourcesOfAvailableWater(Date start, Date end){
		System.out.println("Notifying...");
		// Receive amount available
		List<WBMReceiveDatum> wbmInfo = wbmConnector.receiveFromWBM();
		duplication.clear();
		hasDuplication = false;
		Map<String, WBMNaturalWaterSource> dupSet = new HashMap<String, WBMNaturalWaterSource>();
		
		// Loop through collection of water sources
		for(WBMReceiveDatum datum : wbmInfo ){
			WBMNaturalWaterSource source = waterSources.get(datum.requestedCoordinate);
			if(source == null){
				System.out.println("WBM Returned a lat/lon that does not match any initialize value: " + datum.requestedCoordinate);
			}
			else{
				source.processWBMDatum(start, end, datum);
				String gridCellId = datum.cellID;
				if(dupSet.containsKey(gridCellId)){
					WBMNaturalWaterSource earlier = dupSet.get(gridCellId); 
					if(source.manageIrrigation && (earlier.manageIrrigation == false)){ // Must prefer the one that manages irrigation
						Map<WBMNaturalWaterSource, WBMNaturalWaterSource> temp = new HashMap<WBMNaturalWaterSource, WBMNaturalWaterSource>();
						for(Entry<WBMNaturalWaterSource, WBMNaturalWaterSource> entry: duplication.entrySet()){
							WBMNaturalWaterSource chosen = entry.getValue();
							temp.put(entry.getKey(), chosen.equals(earlier) ? source : chosen);
						}
						duplication.clear();
						duplication.putAll(temp);
						duplication.put(source, source);
					}
					else{
						duplication.put(source, dupSet.get(gridCellId));
					}
					hasDuplication = true;
				}
				else{
					dupSet.put(gridCellId, source);
					duplication.put(source, source);
				}
			}
		}		
	}
	
	/**
	 * Notifies the WBM of the amount taken out of the volume reported
	 * available during the specified interval
	 * @param start
	 * @param end
	 */
	public void notifyWBM(Date start, Date end){
		List<Pair<Point2D.Double, WBMSendDatum> > data = new ArrayList<Pair<Point2D.Double, WBMSendDatum> >();
		for(WBMNaturalWaterSource waterSource: waterSources.values()){
			WBMSendDatum toSend = waterSource.getWBMSend(start, end);
			data.add(new Pair<Point2D.Double, WBMSendDatum> (waterSource.getLatLon(), toSend));
		}
		wbmConnector.notifyWBMOfTaken(data);
	}
	
	public void end(){
		wbmConnector.close();
	}
	
	public boolean hasDuplicates(){
		return hasDuplication;
	}
	
	public WBMNaturalWaterSource getDuplicate(WBMNaturalWaterSource original){
		return duplication.get(original);
	}
	
	public Map<WBMNaturalWaterSource, WBMNaturalWaterSource> removeDuplicates(){
		Map<WBMNaturalWaterSource, WBMNaturalWaterSource> dupMap = new HashMap<WBMNaturalWaterSource, WBMNaturalWaterSource>();
		Set<Point2D.Double> toDelete = new HashSet<Point2D.Double>();
		for(Entry<Point2D.Double, WBMNaturalWaterSource> sourceEntry: waterSources.entrySet()){
			WBMNaturalWaterSource source = sourceEntry.getValue();
			if(!duplication.get(source).equals(source)){
				toDelete.add(sourceEntry.getKey());
			}
		}
		for(Point2D.Double d: toDelete){
			WBMNaturalWaterSource toRemove = waterSources.get(d);
			TemperatureStation ts = (TemperatureStation)toRemove;
			System.out.println(((TemperatureStation)duplication.get(toRemove)).getTemperatureStationId() + "  has a duplicate; about to remove " + ((TemperatureStation)toRemove).getTemperatureStationId());
			dupMap.put(toRemove, duplication.get(toRemove));
			waterSources.remove(d);
		}
		return dupMap;
	}

}
