package repast.simphony;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringWriter;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import repast.simphony.parameter.Parameters;


/**
 * Writes the collection of properties to a file that should emulate
 * a batch run file, but is intended as a record of runs rather than
 * created in bulk to direct runs. (Ideally could be used as a UPF
 * file.)
 * @author murphy
 *
 */
public class PropertiesWriter {

	public static final int DO_NOT_WRITE                         = -1;     // If run number passed is -1, do not write
	public static final int USE_NEXT_LARGEST                     = -9999;  // Look in file, add one to largest run number recorded (1 if none exist)
	
	public static final int NO_EXECUTE_DUPLICATE_RUN_NUMBER      = -101;   // Did not write because the run number provided is a duplicate of one already in the file
	public static final int NO_EXECUTE_NO_RUN_NUMBER_PROVIDED    = -102;   // Did not write because the run number provided was -1 (or other invalid code)
	public static final int NO_EXECUTE_SCHEMA_DOES_NOT_MATCH     = -103;   // Did not write because the schema does not match the one already in the file
	public static final int ERROR_OCCURRED                       = -104;   // Some other error occurred during the write process
	
	public static final int NO_RUN_NUMBER_ESTABLISHED            = -10000; // Run number has not yet been established
	public static int lastRunNumberSaved                         = NO_RUN_NUMBER_ESTABLISHED;
	
	public static int write(String pathToFile, Parameters params){
		if(params.getSchema().contains("RunNumber")) return write(pathToFile, params.getInteger("RunNumber"), params);
		return NO_EXECUTE_NO_RUN_NUMBER_PROVIDED;
	}
	
	
	public static int write(String pathToFile, int runNumber, Parameters params){
		String timeStamp = ((new Date()).toString()).replace(" ", "_");
		int retVal = runNumber; // Default is to return the run number provided
		if(runNumber == -1) return NO_EXECUTE_NO_RUN_NUMBER_PROVIDED;
		File file = new File(pathToFile);
		if(!file.exists()){
			// All you have to do is write to the file here
			if(runNumber == USE_NEXT_LARGEST) runNumber = 1;
			retVal = runNumber;
			String line = stringToWrite(runNumber, timeStamp, params, null);
			try{
				Files.write(Paths.get(pathToFile), line.getBytes(), StandardOpenOption.CREATE);
			}
			catch(Exception E){
				System.err.println(E.getMessage());
				retVal = ERROR_OCCURRED;
			}
		}
		else{ // File already exists; must do some validation first
			if(!schemaMatches(pathToFile, params))       return NO_EXECUTE_SCHEMA_DOES_NOT_MATCH;
			if(containsRunNumber(pathToFile, runNumber)) return NO_EXECUTE_DUPLICATE_RUN_NUMBER;
			if(runNumber == USE_NEXT_LARGEST) runNumber = getMaxRunNumber(pathToFile) + 1;
			List<String> schemaList = getParameterList(pathToFile);
			String line = stringToWrite(runNumber, timeStamp, params, schemaList);
			try{
				Files.write(Paths.get(pathToFile), line.getBytes(), StandardOpenOption.APPEND);
			}
			catch(Exception E){
				System.err.println(E.getMessage());
				retVal = ERROR_OCCURRED;
			}
			retVal = runNumber;	
		}
		lastRunNumberSaved = retVal;
		return retVal;
	}
	
	public static String stringToWrite(int runNumber, String timeStamp, Parameters params, List<String> schemaList){
		StringWriter sw = new StringWriter();
		sw.write(runNumber + "\t" + "TimeStamp\t" + timeStamp);
		if(schemaList != null){
			for(String key: schemaList){
				if((key.compareTo("RunNumber") != 0) && 
				   (key.compareTo("TimeStamp") != 0)) sw.write("," + key + "\t" + params.getValueAsString(key));
			}
		}
		else{
			for(String paramKey: params.getSchema().parameterNames()){
				if((paramKey.compareTo("RunNumber") != 0) && 
				   (paramKey.compareTo("TimeStamp") != 0)) sw.write("," + paramKey + "\t" + params.getValueAsString(paramKey));
			}
		}
		sw.write(System.getProperty("line.separator"));
		return sw.toString();
	}
	
	// Returns true iff every parameter key in the params collection
	// is matched by an entry in the file, and vice versa. Only considers
	// the first line of the file; all others must match this. Does not
	// compare time stamp or run number.
	// Does not test if file exists; test first
	public static boolean schemaMatches(String pathToFile, Parameters params){
		boolean retVal = true;
		Set<String> paramSet = new HashSet<String>();
		for(String param: params.getSchema().parameterNames()) paramSet.add(param);
		// Add RunNumber and TimeStamp
		paramSet.add("RunNumber");
		paramSet.add("TimeStamp");
		BufferedReader bufferedReader = null;
		try {
			bufferedReader = new BufferedReader(new FileReader(pathToFile));
			String line = bufferedReader.readLine();
			Vector<String[]> entries = parseValues(line);
			for(int i = entries.size() - 1; i >= 0; i--){
				String entry = entries.get(i)[0];
				if(!paramSet.contains(entry)){
					retVal = false;
					break;
				}
				else{
					paramSet.remove(entry);
				}
			}
			bufferedReader.close(); // Should close at end...
			// There shouldn't be any leftovers, either
			if(paramSet.size() > 0) retVal = false;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return retVal;
	}
	
	// Does not test if file exists; test first!
	public static List<String> getParameterList(String pathToFile){
		List<String> retVal = new ArrayList<String>();
		BufferedReader bufferedReader = null;
		try {
			bufferedReader = new BufferedReader(new FileReader(pathToFile));
			String line = bufferedReader.readLine();
			Vector<String[]> entries = parseValues(line);
			for(String[] entry: entries) retVal.add(entry[0]);
			
			bufferedReader.close(); // Should close at end...
		} catch (IOException e) {
			e.printStackTrace();
		}
		return retVal;
	}
	
	public static Vector<String[]> parseValues(String line){
		Vector<String[]> ret = new Vector<String[]>();
		int firstSplit = line.indexOf("\t");
		String[] valPair = new String[2];
		valPair[0] = "RunNumber";
		valPair[1] = line.substring(0, firstSplit);
		ret.add(valPair);
		String restOfLine = line.substring(firstSplit+1);
		String[] pairs = restOfLine.split(",");
		for(String pair: pairs){
			valPair = pair.split("\t");
			ret.add(valPair);
		}
		return ret;
	}
	
	// Note: Does not validate if file exists; check first; returns -1 if no entries in file
	public static int getMaxRunNumber(String pathToFile){
	  BufferedReader bufferedReader = null;
	  int max = -1;
	  try {
		bufferedReader = new BufferedReader(new FileReader(pathToFile));
		String line;
		System.out.println("LOOPING TO GET MAX RUN NUMBER");
		while ((line = bufferedReader.readLine()) != null) {
			int newVal = Integer.parseInt(line.substring(0, line.indexOf("\t")));
			System.out.println("    NEW VAL = " + newVal);
			if(newVal > max) max = newVal;
		}
		bufferedReader.close(); // Should close at end...
	  } catch (IOException e) {
		e.printStackTrace();
	  }
	  System.out.println("Retuning " + max);
	  return max;
	}
	
	public static boolean containsRunNumber(String pathToFile, int candidateRunNumber){
	  BufferedReader bufferedReader = null;
	  boolean ret = false;
	  try {
		bufferedReader = new BufferedReader(new FileReader(pathToFile));
		String line;
		while (((line = bufferedReader.readLine()) != null)) {
			if(Integer.parseInt(line.substring(0, line.indexOf("\t"))) == candidateRunNumber){
				ret = true;
				break;
			}
		}
		bufferedReader.close();
	  } catch (IOException e) {
		e.printStackTrace();
	  }
	  return ret;
	}
	
	public static void main(String[] args){
		Vector<String[]> pairs = parseValues("1\tParam1\tValue1,Param2\tValue2");
		for(String[] pair: pairs){
			System.out.println("PARAM = '" + pair[0] + "'" + " VAL = '" + pair[1] + "'");
		}
	}
}
